// WL_AGENT.C

#include "wl_def.h"
#pragma hdrstop

/*
=============================================================================

                                LOCAL CONSTANTS

=============================================================================
*/

#define MAXMOUSETURN    10


#define MOVESCALE       150l
#define BACKMOVESCALE   100l
#define ANGLESCALE      20

/*
=============================================================================

                                GLOBAL VARIABLES

=============================================================================
*/



//
// player state info
//
int32_t         thrustspeed;

word            plux,pluy;          // player coordinates scaled to unsigned

short           anglefrac;

objtype        *LastAttacker;

extern	statetype s_rocket;
extern	statetype s_plasma1;
extern	statetype s_bfg1;
extern	statetype s_hellorb1;

/*
=============================================================================

                                                 LOCAL VARIABLES

=============================================================================
*/


void    T_Player (objtype *ob);
void    T_Attack (objtype *ob);
void	WarpPlayer (int facedir);

statetype   s_player = {false,0,0,(statefunc) T_Player,NULL,NULL};
statetype   s_attack = {false,0,0,(statefunc) T_Attack,NULL,NULL};

struct atkinf
{
    int8_t    tics,attack,frame;              // attack is 1 for gun, 2 for knife
} attackinfo[NUMWEAPONS][14] =
{
{ {8,2,2},{8,0,3},{8,-1,2} }, // wp_fists
{ {6,0,4},{6,2,4},{6,3,6},{6,-1,6} }, // wp_chainsaw
{ {8,1,1},{8,0,2},{8,0,3},{8,-1,4} }, // wp_pistol
{ {8,1,1},{8,0,2},{8,0,3},{16,0,4},{16,0,5},{16,0,4},{16,-1,3} }, // wp_shotgun
{ {8,1,1},{8,0,2},{16,0,3},{16,5,4},{16,0,5},{16,6,6},{16,0,8},{16,7,9},{16,-1,10} }, // wp_supershotgun
{ {8,0,0},{8,1,2},{8,4,3},{8,-1,1} }, // wp_chaingun
{ {8,0,1},{8,1,2},{8,0,3},{8,0,4},{8,0,5},{60,0,0},{8,-1,0} }, // wp_launcher
{ {8,1,1},{8,4,2},{40,0,3},{18,-1,3} }, // wp_plasmagun
{ {30,8,0},{12,1,2},{12,0,4},{12,0,6},{70,0,0},{8,-1,0} }, // wp_bfggun
{ {6,1,3},{6,0,4},{6,0,5},{6,0,6} /* fully down*/,{70,0,7},{6,0,8},{6,0,9},{6,0,10},{6,-1,11} } // wp_hellorb
};

//===========================================================================

boolean TryMove (objtype *ob);
void T_Player (objtype *ob);

void ClipMove (objtype *ob, int32_t xmove, int32_t ymove);

/*
=============================================================================

                                CONTROL STUFF

=============================================================================
*/

/*
======================
=
= CheckWeaponChange
=
= Keys 1-4 change weapons
=
======================
*/

int ChkAtkAmmo (int weapon)
{
    switch (weapon)
    { 
        case wp_chainsaw:
        case wp_fists:
        case wp_hellorb:
            return 1;
        case wp_pistol:
        case wp_chaingun:
            if (gamestate.ammo > 0)
                return 1;
            break;
        case wp_supershotgun:
            if (gamestate.ammo2 > 1)
                return 1;
            break;
        case wp_shotgun:
            if (gamestate.ammo2 > 0)
                return 1;
            break;
        case wp_launcher:
            if (gamestate.ammo3 > 0)
                return 1;
            break;
        case wp_bfggun:
            if (gamestate.ammo4 > 39)
                return 1;
            break;
        case wp_plasmagun:
            if (gamestate.ammo4 > 0)
                return 1;
            break;
    }

    return 0;
}
void CheckWeaponChange (void)
{
    int   kbuttons[9]={sc_1,sc_8,sc_2,sc_3,sc_3,sc_4,sc_5,sc_6,sc_7};
    int i,newWeapon = -1;

    //if (ChkAtkAmmo (i))            // must use knife with no ammo
    //    return;

    if ((!gamestate.ammo && gamestate.nextweapon==wp_pistol) ||
        (!gamestate.ammo && gamestate.nextweapon==wp_chaingun) ||
        (!gamestate.ammo2 && gamestate.nextweapon==wp_shotgun) ||
        (gamestate.ammo2<2 && gamestate.nextweapon==wp_supershotgun) ||
        (!gamestate.ammo3 && gamestate.nextweapon==wp_launcher) ||
        (!gamestate.ammo4 && gamestate.nextweapon==wp_plasmagun) ||
        (gamestate.ammo4<40 && gamestate.nextweapon==wp_bfggun)) {
        return;
    }

        for (i=0;i<9;i++) {
            if ((Keyboard[kbuttons[i]]) &&(i != gamestate.nextweapon))
            {
                if (gamestate.bestweapon & (1<<i) && ChkAtkAmmo(i))
                {
                    if (gamestate.weaponframe > 0 || gamestate.attackframe > 0) {
                        gamestate.nextweapon = (weapontype) i;
                    }
                    else {
                        gamestate.nextweapon = gamestate.chosenweapon = (weapontype) i;
                        gamestate.changingweapon = gamestate.goingdown = true;
                    }
                    //IN_ClearKeysDown();
                    //LastScan=0;
                    return;
                }
            } 
        }
}

/*
=======================
=
= ControlMovement
=
= Takes controlx,controly, and buttonstate[bt_strafe]
=
= Changes the player's angle and position
=
= There is an angle hack because when going 70 fps, the roundoff becomes
= significant
=
=======================
*/

void ControlMovement (objtype *ob)
{
    int32_t oldx,oldy;
    int     angle;
    int     angleunits;

    thrustspeed = 0;

    oldx = player->x;
    oldy = player->y;

    if(buttonstate[bt_strafeleft])
    {
        angle = ob->angle + ANGLES/4;
        if(angle >= ANGLES)
            angle -= ANGLES;
        if(buttonstate[bt_run])
            Thrust(angle, RUNMOVE * MOVESCALE * tics);
        else
            Thrust(angle, BASEMOVE * MOVESCALE * tics);
    }

    if(buttonstate[bt_straferight])
    {
        angle = ob->angle - ANGLES/4;
        if(angle < 0)
            angle += ANGLES;
        if(buttonstate[bt_run])
            Thrust(angle, RUNMOVE * MOVESCALE * tics );
        else
            Thrust(angle, BASEMOVE * MOVESCALE * tics);
    }

    //
    // side to side move
    //
    if (buttonstate[bt_strafe])
    {
        //
        // strafing
        //
        //
        if (controlx > 0)
        {
            angle = ob->angle - ANGLES/4;
            if (angle < 0)
                angle += ANGLES;
            Thrust (angle,controlx*MOVESCALE);      // move to left
        }
        else if (controlx < 0)
        {
            angle = ob->angle + ANGLES/4;
            if (angle >= ANGLES)
                angle -= ANGLES;
            Thrust (angle,-controlx*MOVESCALE);     // move to right
        }
    }
    else
    {
        //
        // not strafing
        //
        anglefrac += controlx;
        angleunits = anglefrac/ANGLESCALE;
        anglefrac -= angleunits*ANGLESCALE;
        ob->angle -= angleunits;

        if (ob->angle >= ANGLES)
            ob->angle -= ANGLES;
        if (ob->angle < 0)
            ob->angle += ANGLES;

    }

    //
    // forward/backwards move
    //
    if (controly < 0)
    {
        Thrust (ob->angle,-controly*MOVESCALE); // move forwards
    }
    else if (controly > 0)
    {
        angle = ob->angle + ANGLES/2;
        if (angle >= ANGLES)
            angle -= ANGLES;
        Thrust (angle,controly*BACKMOVESCALE);          // move backwards
    }

    if (gamestate.victoryflag)              // watching the BJ actor
        return;
}

/*
=============================================================================

                            STATUS WINDOW STUFF

=============================================================================
*/


/*
==================
=
= StatusDrawPic
=
==================
*/

void StatusDrawPic (unsigned x, unsigned y, unsigned picnum)
{
    LatchDrawPicScaledCoord ((screenWidth-scaleFactor*320)/16 + scaleFactor*x,
        screenHeight-scaleFactor*(STATUSLINES-y),picnum);
}

void StatusDrawFace(unsigned picnum)
{
    byte *source = grsegs[STATUSBARPIC];
    if(viewsize < 21 && ingame)
        VL_MemToScreenScaledCoord(source, 320,32, 144, 1, 144*scaleFactor, 169*scaleFactor,32,31);

    StatusDrawPic(144, 1, picnum);

#ifdef _arch_dreamcast
    DC_StatusDrawLCD(picnum);
#endif
}


/*
==================
=
= DrawFace
=
==================
*/

short ouchcount = 0;
short evilcount = 0;

void EvilFace (void)
{
    if (gamestate.health>80)
        StatusDrawFace (EVIL0FACE);
    if (gamestate.health>60 && gamestate.health<=80)
        StatusDrawFace (EVIL1FACE);
    if (gamestate.health>40 && gamestate.health<=60)
        StatusDrawFace (EVIL2FACE);
    if (gamestate.health>20 && gamestate.health<=40)
        StatusDrawFace (EVIL3FACE);
    if (gamestate.health>0 && gamestate.health<=20)
        StatusDrawFace (EVIL4FACE);
}

void DrawOuchFace() {
    if (viewsize != 22) {
        if (gamestate.health >80)
            StatusDrawFace (BJOUCH1PIC);
        else if (gamestate.health>60 && gamestate.health<81)
            StatusDrawFace (BJOUCH2PIC);
        else if (gamestate.health>40 && gamestate.health<61)
            StatusDrawFace (BJOUCH3PIC);
        else if (gamestate.health>20 && gamestate.health<41)
            StatusDrawFace (BJOUCH4PIC);
        else if (gamestate.health<21)
            StatusDrawFace (BJOUCH5PIC);
    }
}

void DrawFace (void)
{   
    if (evilcount > 0) { // weapon pickup?
        EvilFace();
    }
    else if (ouchcount > 0) {
        DrawOuchFace();
    }
    else if (gamestate.health)
    {
        if (gamestate.godmode || gamestate.bonusgodmode)
            StatusDrawFace(GODFACEPIC);
        else if (gamestate.health>=100)
            StatusDrawFace (FACE1APIC+gamestate.faceframe);
        else
            StatusDrawFace (FACE1APIC+3*((100-gamestate.health)/20)+gamestate.faceframe);
    }
    else {
        StatusDrawFace(FACE8APIC);
    }
}

byte nukagecount;
byte lavacount;
/*
========================
=
= DamageableFloors
=
= Every period, the floors give damage to the player
= unless they wear a biosuit
========================
*/
void DamageableFloors() {
    nukagecount+=tics;
    lavacount+=tics;
    if (nukagecount >= 70)
    {
        nukagecount=0;
        if ((MAPSPOT(player->tilex,player->tiley,2)&0xff)==NUKAGETILE && !gamestate.biosuit) // nukage
                TakeDamage ((gamestate.difficulty+1)*4,NULL);
    }
    if (lavacount >= 35)
    {
        lavacount=0;
        if ((MAPSPOT(player->tilex,player->tiley,2)&0xff)==LAVATILE && !gamestate.biosuit) // lava
                TakeDamage ((gamestate.difficulty+1)*4,NULL);
    }
}
/*
===============
=
= UpdateFace
=
= Calls draw face if time to change
=
===============
*/

int facecount = 0;
int facetimes = 0;

void UpdateFace (void)
{
    DamageableFloors();
    if (ouchcount > 0)
        ouchcount--;

    if (evilcount > 0)
        evilcount--;

    // don't make demo depend on sound playback
    if(demoplayback || demorecord)
    {
        if(facetimes > 0)
        {
            facetimes--;
            return;
        }
    }
    else if(SD_SoundPlaying() == DSWPNUP)
        return;

    facecount += tics;
    if (facecount > US_RndT())
    {
        gamestate.faceframe = (US_RndT()>>6);
        if (gamestate.faceframe==3)
            gamestate.faceframe = 1;

        facecount = 0;
    }
}

/*
===============
=
= LatchNumber
=
= right justifies and pads with blanks
=
===============
*/

static void LatchNumber (int x, int y, unsigned width, int32_t number)
{
    unsigned length,c;
    char    str[20];

    ltoa (number,str,10);

    length = (unsigned) strlen (str);

    while (length<width)
    {
        StatusDrawPic (x,y,N_BLANKPIC);
        x+=14;
        width--;
    }

    c = length <= width ? 0 : length-width;

    while (c<length)
    {
        StatusDrawPic (x,y,str[c]-'0'+ N_0PIC);
        if (str[c]-'0'+N_0PIC==N_1PIC)
        x+=12;
        else
        x+=14;
        c++;
    }
}

static void LatchSNumber (int x, int y, unsigned width, int32_t number)
{
    // TODO: Add the S_numbers
    unsigned length,c;
    char    str[20];

    ltoa (number,str,10);

    length = (unsigned) strlen (str);

    while (length<width)
    {
        StatusDrawPic (x,y,A_BLANKPIC);
        x+=4;
        width--;
    }

    c = length <= width ? 0 : length-width;

    while (c<length)
    {
        StatusDrawPic (x,y,str[c]-'0'+ A_0PIC);
        x+=4;
        c++;
    }
}

void FullScreenDraw (void)
{
    LatchDrawPic (46,181,N_PRCNTPIC);
    LatchDrawPic (66,178,S_HEALTHPIC);
    LatchDrawPic (300,181,N_PRCNTPIC);

    if (gamestate.armor<=100)
        LatchDrawPic (222,179,S_ARMOR100PIC);
    else if (gamestate.armor>100)
        LatchDrawPic (222,179,S_ARMOR200PIC);

    switch (gamestate.nextweapon)
    {
        case wp_pistol:
        case wp_chaingun:
            LatchDrawPic (50,14,S_AMMO1PIC);
            break;
        case wp_shotgun:
        case wp_supershotgun:
            LatchDrawPic (50,16,S_AMMO2PIC);
            break;
        case wp_launcher:
            LatchDrawPic (50,5,S_AMMO3PIC);
            break;
        case wp_plasmagun:
        case wp_bfggun:
            LatchDrawPic (50,13,S_AMMO4PIC);
            break;
    }

    if (gamestate.keys & 1)
        StatusDrawPic (272,6,KEY1PIC);
    if (gamestate.keys & 2)
        StatusDrawPic (286,6,KEY2PIC);
    if (gamestate.keys & 4)
        StatusDrawPic (300,6,KEY3PIC);

}

/*
===============
=
= DrawHealth
=
===============
*/

void DrawHealth (void)
{
    byte *source = grsegs[STATUSBARPIC];
    if(viewsize >= 21 && ingame) return;
    VL_MemToScreenScaledCoord(source, 320,32, 49, 3, 49*scaleFactor, 171*scaleFactor,42,16);
    LatchNumber (49,3,3,gamestate.health);
    StatusDrawPic (90,3,N_PRCNTPIC);
}

/*
===============
=
= TakeDamage
=
===============
*/
void TakeDamage (int points,objtype *attacker)
{
    LastAttacker = attacker;

    if (gamestate.victoryflag)
        return;
    if (gamestate.godmode || gamestate.bonusgodmode)
        return;

    if (gamestate.difficulty==gd_baby)
        points>>=2;

    SD_PlaySound (DSPLPAIN);
    
    if (gamestate.armor > 0)
    {
        if ((gamestate.health -= points/3) >= 0)
        {
            gamestate.health -= ((int)points/3);
            if (gamestate.health > 200) {
                gamestate.health = 200;
            }
        }

        gamestate.armor -= points/2+2;
    }
    else {
        gamestate.health -= points;
    }

    if (gamestate.armor<=0)
    gamestate.armor=0;

    if (gamestate.health<=0) {
        gamestate.health = 0;
        playstate = ex_died;
        killerobj = attacker;
    }

    if (!gamestate.godmode && !gamestate.bonusgodmode)
        StartDamageFlash (points);

    if (points>=20 && gamestate.health!=0 && (!gamestate.godmode && !gamestate.bonusgodmode))
    {
        ouchcount = 70; // 1 second
    }
}

/*
===============
=
= HealSelf
=
===============
*/

void HealSelf (int points)
{
    gamestate.health += points;
    if (gamestate.health>100)
        gamestate.health = 100;
}

void HealBonus (int points)
{
    gamestate.health += points;
    if (gamestate.health > 200)
        gamestate.health = 200;
}

void DrawArmor (void)
{
    byte *source = grsegs[STATUSBARPIC];
    if (viewsize >=21 && ingame) return;
    VL_MemToScreenScaledCoord(source,320,32,179,3,179*scaleFactor,171*scaleFactor,42,16);
    LatchNumber (179,3,3,gamestate.armor);
    StatusDrawPic (221,3,N_PRCNTPIC);
}

void GiveArmor (int points)
{
    gamestate.armor += points;
    if (gamestate.armor>100)
    gamestate.armor=100;
}

void ArmorBonus (int points)
{
    gamestate.armor += points;
    if (gamestate.armor>200)
    gamestate.armor=200;
}

//===========================================================================


/*
===============
=
= DrawLevel
=
===============
*/

void DrawLevel (void)
{
    if(viewsize == 21 && ingame) return;
#ifdef SPEAR
    if (gamestate.mapon == 20)
        LatchNumber (2,16,2,18);
    else
#endif
        LatchNumber (2,16,2,gamestate.mapon+1);
}

//===========================================================================


/*
===============
=
= DrawLives
=
===============
*/

void DrawLives (void)
{
    byte *source = grsegs[STATUSBARPIC];
    if(viewsize >= 21 && ingame) return;
    
    if (viewsize==20)
    VL_MemToScreenScaledCoord(source, 320,32, 144, 25, 144*scaleFactor, 193*scaleFactor,4,6);
    LatchSNumber (144,25,1,gamestate.lives);
}


/*
===============
=
= GiveExtraMan
=
===============
*/

void GiveExtraMan (void)
{
    if (gamestate.lives<9)
        gamestate.lives++;
    SD_PlaySound (DSGETPOW);
}

//===========================================================================

/*
===============
=
= DrawScore
=
===============
*/

void DrawScore (void)
{
    //if(viewsize == 21 && ingame) return;
    //LatchNumber (6,16,6,gamestate.score);
}

/*
===============
=
= GivePoints
=
===============
*/

void GivePoints (int32_t points)
{
    gamestate.score += points;
    while (gamestate.score >= gamestate.nextextra)
    {
        gamestate.nextextra += EXTRAPOINTS;
        GiveExtraMan ();
    }
}

//===========================================================================

/*
==================
=
= DrawWeapon
=
==================
*/

void DrawWeapon (void)
{
    if(viewsize >= 21 && ingame) return;
    if (gamestate.bestweapon & 4)
        StatusDrawPic (111,4,A_2PIC);
    else
        StatusDrawPic (111,4,S_2PIC);

    if (gamestate.bestweapon & 8 || 
        gamestate.bestweapon & 16)
        StatusDrawPic (123,4,A_3PIC);
    else
        StatusDrawPic (123,4,S_3PIC);

    if (gamestate.bestweapon & 32)
        StatusDrawPic (135,4,A_4PIC);
    else
        StatusDrawPic (135,4,S_4PIC);

    if (gamestate.bestweapon & 64)
        StatusDrawPic (111,14,A_5PIC);
    else
        StatusDrawPic (111,14,S_5PIC);

    if (gamestate.bestweapon & 128)
        StatusDrawPic (123,14,A_6PIC);
    else
        StatusDrawPic (123,14,S_6PIC);

    if (gamestate.bestweapon & 256)
        StatusDrawPic (135,14,A_7PIC);
    else
        StatusDrawPic (135,14,S_7PIC);
}


/*
==================
=
= DrawKeys
=
==================
*/

void DrawKeys (void)
{
   byte *source = grsegs[STATUSBARPIC];
    if (gamestate.keys & 1)
        StatusDrawPic (239,4,KEY1PIC);
    else
        VL_MemToScreenScaledCoord(source, 320,32, 239, 4, 239*scaleFactor, 172*scaleFactor,7,5);

    if (gamestate.keys & 2)
        StatusDrawPic (239,14,KEY2PIC);
    else
        VL_MemToScreenScaledCoord(source, 320,32, 239, 14, 239*scaleFactor, 182*scaleFactor,7,5);

    if (gamestate.keys & 4)
        StatusDrawPic (239,24,KEY3PIC);
    else
        VL_MemToScreenScaledCoord(source, 320,32, 239, 24, 239*scaleFactor, 192*scaleFactor,7,5);
}

/*
==================
=
= GiveWeapon
=
==================
*/

void GiveWeapon (int weapon, int giveammo)
{
    switch (weapon)
    {
        case wp_shotgun:
            GiveAmmo(giveammo,ammo2);
            break;
        case wp_supershotgun:
            GiveAmmo(giveammo,ammo2);
            break;
        case wp_chaingun:
            GiveAmmo(giveammo,ammo1);
            break;
        case wp_launcher:
            GiveAmmo(giveammo,ammo3);
            break;
        case wp_plasmagun:
            GiveAmmo(giveammo,ammo4);
            break;
        case wp_bfggun:
            GiveAmmo(giveammo,ammo4);
            break;
        }

    if (gamestate.bestweapon & (1<<weapon))
        return;
    if (gamestate.attackframe > 0 || gamestate.weaponframe > 0) {
        gamestate.nextweapon = (weapontype) weapon; 
    } else {
        gamestate.nextweapon = gamestate.chosenweapon = (weapontype) weapon; 
        gamestate.goingdown = gamestate.changingweapon = true;
    }
    short temp = (short) gamestate.bestweapon;
    temp |= (1<<((short) weapon));
    gamestate.bestweapon = (weapontype) temp;
}

//===========================================================================

/*
===============
=
= DrawAmmo
=
===============
*/

void DrawAmmo (void)
{
    if(viewsize >= 21 && ingame) return;

    byte *source = grsegs[STATUSBARPIC];
    VL_MemToScreenScaledCoord(source, 320,32, 2, 3, 2*scaleFactor, 171*scaleFactor,42,16);
    VL_MemToScreenScaledCoord(source,320,32,276,5,276*scaleFactor,173*scaleFactor,12,24);
    VL_MemToScreenScaledCoord(source,320,32,302,5,302*scaleFactor,173*scaleFactor,12,24);

    switch (gamestate.nextweapon)
    {
        case wp_fists:
        case wp_chainsaw:
            // Nothing is drawn
            break;
        case wp_pistol:
        case wp_chaingun:
            LatchNumber (2,3,3,gamestate.ammo);
            break;
        case wp_shotgun:
        case wp_supershotgun:
            LatchNumber (2,3,3,gamestate.ammo2);
            break;
        case wp_launcher:
            LatchNumber (2,3,3,gamestate.ammo3);
            break;
        case wp_plasmagun:
        case wp_bfggun:
            LatchNumber (2,3,3,gamestate.ammo4);
            break;
    }
    
    LatchSNumber (276,5,3,gamestate.ammo);
    LatchSNumber (276,11,3,gamestate.ammo2);
    LatchSNumber (276,17,3,gamestate.ammo3);
    LatchSNumber (276,23,3,gamestate.ammo4);
    LatchSNumber (302,5,3,gamestate.ammomax);
    LatchSNumber (302,11,3,gamestate.ammo2max);
    LatchSNumber (302,17,3,gamestate.ammo3max);
    LatchSNumber (302,23,3,gamestate.ammo4max);
}

/*
===============
=
= GiveAmmo
=
===============
*/

void GiveAmmo (int ammo, int ammotype)
{
    switch (ammotype)
    {
        case ammo1:
        {
            gamestate.ammo += ammo;
            if (gamestate.ammo > gamestate.ammomax)
            gamestate.ammo = gamestate.ammomax;
        }
        break;
        case ammo2:
        {
            gamestate.ammo2 += ammo;
            if (gamestate.ammo2 > gamestate.ammo2max)
            gamestate.ammo2 = gamestate.ammo2max;
        }
        break;
        case ammo3:
        {
            gamestate.ammo3 += ammo;
            if (gamestate.ammo3 > gamestate.ammo3max)
            gamestate.ammo3 = gamestate.ammo3max;
        }
        break;
        case ammo4:
        {
            gamestate.ammo4 += ammo;
            if (gamestate.ammo4 > gamestate.ammo4max)
            gamestate.ammo4 = gamestate.ammo4max;
        }
        break;
    }
}

//===========================================================================

void GiveBackPack (void)
{
    if (gamestate.ammomax < 400)
        gamestate.ammomax += MAXAMMO;
    else
        gamestate.ammomax = 400;

    if (gamestate.ammo2max < 100)
        gamestate.ammo2max += MAXAMMO2;
    else
        gamestate.ammo2max = 100;

    if (gamestate.ammo3max < 100)
        gamestate.ammo3max += MAXAMMO3;
    else
        gamestate.ammo3max = 100;

    if (gamestate.ammo4max < 600)
        gamestate.ammo4max += MAXAMMO4;
    else
        gamestate.ammo4max = 600;

    GiveAmmo (10,ammo1);
    GiveAmmo (4,ammo2);
    GiveAmmo (1,ammo3);
    GiveAmmo (20,ammo4);
}

/*
==================
=
= GiveKey
=
==================
*/

void GiveKey (int key)
{
    gamestate.keys |= (1<<key);
}


void TimedBonus (int bonustype, int time)
{
    switch (bonustype)
    {
        case p_godmode:
            gamestate.bonusgodmode = 1;
            gamestate.godmodecount+=time;
            break;
        case p_berserk:
            gamestate.berserk=true;
            gamestate.berserkcount++;
            break;
        case p_biosuit:
            gamestate.biosuit = 1;
            gamestate.biosuitcount=time;
            break;
        case p_invisibility:
            gamestate.invisibility = 1;
            gamestate.invisibilitycount+=time;
            break;
        case p_lightamp:
            gamestate.lightamp = 1;
            gamestate.lightampcount+=time;
            break;
    }

}

/*
=============================================================================

                                MOVEMENT

=============================================================================
*/


/*
===================
=
= GetBonus
=
===================
*/
void GetBonus (statobj_t *check)
{
    boolean secrettile = false;
    switch (check->itemnumber)
    {
        case bo_key1:
            GetMessage ("You found the yellow key.");
            GiveKey (check->itemnumber - bo_key1);
            SD_PlaySound (DSITEMUP);
            break;
        case bo_key2:
            GetMessage ("You found the blue key.");
            GiveKey (check->itemnumber - bo_key1);
            SD_PlaySound (DSITEMUP);
            break;
        case bo_key3:
            GetMessage ("You found the red key.");
            GiveKey (check->itemnumber - bo_key1);
            SD_PlaySound (DSITEMUP);
            break;
        case bo_ammo1:
            if (gamestate.ammo==gamestate.ammomax)
            return;

            SD_PlaySound (DSITEMUP);
            GiveAmmo (10,ammo1);
            GetMessage("Picked up a clip.");
            break;
        case bo_ammoused:
            if (gamestate.ammo==gamestate.ammomax)
            return;

            SD_PlaySound (DSITEMUP);
            GiveAmmo (5,ammo1);
            GetMessage ("Picked up a clip.");
            break;
        case bo_ammo1box:
            if (gamestate.ammo==gamestate.ammomax)
            return;

            SD_PlaySound (DSITEMUP);
            GiveAmmo (25,ammo1);
            GetMessage ("Picked up a box of ammo.");
            break;
        case bo_ammo2:
            if (gamestate.ammo2==gamestate.ammo2max)
            return;

            SD_PlaySound (DSITEMUP);
            GiveAmmo (4,ammo2);
            GetMessage ("Picked up some shells.");
            break;
        case bo_ammo2box:
            if (gamestate.ammo2==gamestate.ammo2max)
            return;

            SD_PlaySound (DSITEMUP);
            GiveAmmo (16,ammo2);
            GetMessage("Picked up a box of shells.");
            break;
        case bo_ammo3:
            if (gamestate.ammo3==gamestate.ammo3max)
            return;

            SD_PlaySound (DSITEMUP);
            GiveAmmo (1,ammo3);
            GetMessage ("Picked up a rocket.");
            break;
        case bo_ammo3box:
            if (gamestate.ammo3==gamestate.ammo3max)
            return;

            SD_PlaySound (DSITEMUP);
            GiveAmmo (4,ammo3);
            GetMessage ("Picked up a box of rockets.");
            break;
        case bo_ammo4:
            if (gamestate.ammo4==gamestate.ammo4max)
            return;

            SD_PlaySound (DSITEMUP);
            GiveAmmo (30,ammo4);
            GetMessage("Picked up a cell.");
            break;
        case bo_ammo4box:
            if (gamestate.ammo4==gamestate.ammo4max)
            return;

            SD_PlaySound (DSITEMUP);
            GiveAmmo (100,ammo4);
            GetMessage ("Picked up a cell pack.");
            break;
        case bo_shotgun:
            if (gamestate.chosenweapon == wp_shotgun)
            GiveAmmo (4,ammo2);
            else
            GiveWeapon (wp_shotgun, 4);
            SD_PlaySound (DSWPNUP);
            GetMessage ("Picked up a shotgun.");
            break;
        case bo_supershotgun:
            GiveWeapon (wp_supershotgun, 8);
            evilcount = 40;
            SD_PlaySound (DSWPNUP);
            GetMessage("You found a Super Shotgun!");
            break;
        case bo_chaingun:
            GiveWeapon (wp_chaingun, 20);
            evilcount = 40;
            SD_PlaySound (DSWPNUP);
            GetMessage("Picked up a chaingun.");
            break;
        case bo_launcher:
            GiveWeapon (wp_launcher, 2);
            evilcount = 40;
            GetMessage("You found the Rocket Launcher!");
            SD_PlaySound (DSWPNUP);
            break;
        case bo_plasmagun:
            GiveWeapon (wp_plasmagun, 20);
            GetMessage ("Picked up the plasma gun.");
            SD_PlaySound (DSWPNUP);
            break;
        case bo_bfggun:
            GiveWeapon (wp_bfggun, 40);
            evilcount = 40;
            GetMessage("You found the BFG 9000! Oh yes!");
            SD_PlaySound (DSWPNUP);
            break;
        case bo_chainsaw:
            SD_PlaySound (DSWPNUP);
            GetMessage ("You found the chainsaw! Go find some meat!");
            GiveWeapon (wp_chainsaw, 0);
            
            evilcount = 40;
            break;
        case bo_megaarmor:
            SD_PlaySound (DSITEMUP);
            ArmorBonus (200);
            GetMessage ("Picked up the MegaArmor!");
            break;
        case bo_armor:
            if (gamestate.armor>=100)
            return;

            SD_PlaySound (DSITEMUP);
            GiveArmor (100);
            GetMessage ("Picked up the armor.");
            break;
        case bo_armor1:
            SD_PlaySound (DSITEMUP);
            ArmorBonus (1);
            GetMessage ("Picked up an armor bonus.");
            gamestate.treasurecount++;
            break;
        case bo_health1:
            SD_PlaySound (DSITEMUP);
            HealBonus (1);
            GetMessage ("Picked up a health bonus.");
            gamestate.treasurecount++;
            break;
        case bo_medpack:
            if (gamestate.health>=100)
            return;

            SD_PlaySound (DSITEMUP);
            GetMessage("Picked up a stimpack.");
            HealSelf (10);
            break;
        case bo_healthkit:
            if (gamestate.health>=100)
            return;

            SD_PlaySound (DSITEMUP);
            GetMessage("Picked up a health pack.");
            HealSelf (25);
            break;
        case bo_berserk:
            gamestate.treasurecount++;
            SD_PlaySound (DSITEMUP);
            if (gamestate.attackframe > 0 || gamestate.weaponframe > 0) {
                gamestate.nextweapon = wp_fists;
            } else {
                gamestate.nextweapon = gamestate.chosenweapon = wp_fists;
                gamestate.goingdown = gamestate.changingweapon = true;
            }
            if (gamestate.health < 100) {
                gamestate.health = 100;
            }
            TimedBonus (p_berserk,BONUS_TIME);
            GivePoints(500);
            GetMessage("Berserk!");
            break;
        case bo_backpack:
            SD_PlaySound (DSITEMUP);
            GiveBackPack ();
            GetMessage ("You found a backpack.");
            gamestate.treasurecount++;
            break;
        case bo_godmode:
            SD_PlaySound (DSGETPOW);
            if (!gamestate.godmode)
                TimedBonus (p_godmode,BONUS_TIME);
            gamestate.treasurecount++;
            gamestate.health = 100;
            gamestate.armor = 0;
            GetMessage("Godmode!");
            GivePoints(2000);
            break;
        case bo_megasphere:
            SD_PlaySound (DSITEMUP);
            HealBonus (199);
            ArmorBonus (200);
            gamestate.treasurecount++;
            GetMessage("MEGASPHERE!");
            break;
        case bo_goggles:
            gamestate.treasurecount++;
            TimedBonus (p_lightamp,BONUS_TIME);
            GetMessage("Light Amplification Visor.");
            GivePoints(500);
            break;
        case bo_cloak:
            SD_PlaySound (DSGETPOW);
            gamestate.treasurecount++;
            GetMessage("Partial Invisibility.");
            TimedBonus (p_invisibility,BONUS_TIME*2);
            GivePoints(1000);
            break;
        case bo_soul:
            HealBonus (100);
            gamestate.treasurecount++;
            GetMessage("Soulsphere!");
            GivePoints(500);
            break;
        case bo_biosuit:
            SD_PlaySound (DSITEMUP);
            TimedBonus (p_biosuit,BONUS_TIME*2);
            GivePoints(100);
            GetMessage("Biosuit");
            break;
        case bo_hellorb:
            spearflag=true;
            spearx = player->x;
            speary = player->y;
            spearangle=player->angle;
            playstate = ex_completed;
            break;
        case bo_secret:
            gamestate.secretcount++;
            GetMessage("You found a secret area");
            secrettile = true;
            break;
    }

    if (!secrettile)
        StartBonusFlash ();
    check->shapenum = -1;                   // remove from list
}

/*
===================
=
= TryMove
=
= returns true if move ok
= debug: use pointers to optimize
===================
*/

boolean TryMove (objtype *ob)
{
    int         xl,yl,xh,yh,x,y;
    objtype    *check;
    int32_t     deltax,deltay;

    xl = (ob->x-PLAYERSIZE) >>TILESHIFT;
    yl = (ob->y-PLAYERSIZE) >>TILESHIFT;

    xh = (ob->x+PLAYERSIZE) >>TILESHIFT;
    yh = (ob->y+PLAYERSIZE) >>TILESHIFT;

#define PUSHWALLMINDIST PLAYERSIZE

    //
    // check for solid walls
    //
    for (y=yl;y<=yh;y++)
    {
        for (x=xl;x<=xh;x++)
        {
            check = actorat[x][y];
            if (check && !ISPOINTER(check))
            {
                if(tilemap[x][y]==64 && x==pwallx && y==pwally)   // back of moving pushwall?
                {
                    switch(pwalldir)
                    {
                        case di_north:
                            if(ob->y-PUSHWALLMINDIST<=(pwally<<TILESHIFT)+((63-pwallpos)<<10))
                                return false;
                            break;
                        case di_west:
                            if(ob->x-PUSHWALLMINDIST<=(pwallx<<TILESHIFT)+((63-pwallpos)<<10))
                                return false;
                            break;
                        case di_east:
                            if(ob->x+PUSHWALLMINDIST>=(pwallx<<TILESHIFT)+(pwallpos<<10))
                                return false;
                            break;
                        case di_south:
                            if(ob->y+PUSHWALLMINDIST>=(pwally<<TILESHIFT)+(pwallpos<<10))
                                return false;
                            break;
                    }
                }
                else return false;
            }
        }
    }

    //
    // check for actors
    //
    if (yl>0)
        yl--;
    if (yh<MAPSIZE-1)
        yh++;
    if (xl>0)
        xl--;
    if (xh<MAPSIZE-1)
        xh++;

    for (y=yl;y<=yh;y++)
    {
        for (x=xl;x<=xh;x++)
        {
            check = actorat[x][y];
            if (ISPOINTER(check) && check != player && (check->flags & FL_SHOOTABLE) )
            {
                deltax = ob->x - check->x;
                if (deltax < -MINACTORDIST || deltax > MINACTORDIST)
                    continue;
                deltay = ob->y - check->y;
                if (deltay < -MINACTORDIST || deltay > MINACTORDIST)
                    continue;

                return false;
            }
        }
    }

    return true;
}


/*
===================
=
= ClipMove
=
===================
*/

void ClipMove (objtype *ob, int32_t xmove, int32_t ymove)
{
    int32_t    basex,basey;

    basex = ob->x;
    basey = ob->y;

    ob->x = basex+xmove;
    ob->y = basey+ymove;
    if (TryMove (ob))
        return;

#ifndef REMDEBUG
    if (noclip && ob->x > 2*TILEGLOBAL && ob->y > 2*TILEGLOBAL
        && ob->x < (((int32_t)(mapwidth-1))<<TILESHIFT)
        && ob->y < (((int32_t)(mapheight-1))<<TILESHIFT) )
        return;         // walk through walls
#endif

 //   if (!SD_SoundPlaying())
//        SD_PlaySound (HITWALLSND);

    ob->x = basex+xmove;
    ob->y = basey;
    if (TryMove (ob))
        return;

    ob->x = basex;
    ob->y = basey+ymove;
    if (TryMove (ob))
        return;

    ob->x = basex;
    ob->y = basey;
}

//==========================================================================

/*
===================
=
= VictoryTile
=
===================
*/

void VictoryTile (void)
{
    gamestate.victoryflag = true;
}

/*
===================
=
= Thrust
=
===================
*/

// For player movement in demos exactly as in the original Wolf3D v1.4 source code
static fixed FixedByFracOrig(fixed a, fixed b)
{
    int sign = 0;
    if(b == 65536) b = 65535;
    else if(b == -65536) b = 65535, sign = 1;
    else if(b < 0) b = (-b), sign = 1;

    if(a < 0)
    {
        a = -a;
        sign = !sign;
    }
    fixed res = (fixed)(((int64_t) a * b) >> 16);
    if(sign)
        res = -res;
    return res;
}

void WeaponBobbing() {
    if (gamestate.bobdir != 2) {
        gamestate.bobber++;
    }

    if (gamestate.bobdir == 2) {
        gamestate.bobber--;
    }

    if (gamestate.bobber == (15*(screenHeight/200))) {
        gamestate.bobdir = 2;
    }

    if (gamestate.bobber == 0) {
        gamestate.bobdir = 1;
    }

    if (gamestate.bobdir2 != 2) {
        gamestate.bobber2++;
    }

    if (gamestate.bobdir2 == 2) {
        gamestate.bobber2--;
    }

    if (gamestate.bobber2 == (30*(screenHeight/200))) {
        gamestate.bobdir2 = 2;
    }

    if (gamestate.bobber2 == 0) {
        gamestate.bobdir2 = 1;
    }
}

void Thrust (int angle, int32_t speed)
{
    int32_t xmove,ymove;
    unsigned offset;

    WeaponBobbing();

    thrustspeed += speed;
    //
    // moving bounds speed
    //
    if (speed >= MINDIST*2)
        speed = MINDIST*2-1;

    xmove = DEMOCHOOSE_ORIG_SDL(
                FixedByFracOrig(speed, costable[angle]),
                FixedMul(speed,costable[angle]));
    ymove = DEMOCHOOSE_ORIG_SDL(
                -FixedByFracOrig(speed, sintable[angle]),
                -FixedMul(speed,sintable[angle]));

    ClipMove(player,xmove,ymove);

    player->tilex = (short)(player->x >> TILESHIFT);                // scale to tile values
    player->tiley = (short)(player->y >> TILESHIFT);

    offset = (player->tiley<<mapshift)+player->tilex;
    player->areanumber = *(mapsegs[0] + offset) -AREATILE;

    switch (*(mapsegs[1] + offset)) {
        case EXITTILE:
            playstate = ex_victorious;
            break;
        case ENDLEVELTILE:
            playstate = ex_completed;
            break;
        case ALTELEVATORTILE:
            playstate = ex_secretlevel;
            break;
        case WARPEASTTILE:
            WarpPlayer(EAST);
            break;
        case WARPWESTTILE:
            WarpPlayer(WEST);
            break;
    }
}


/*
=============================================================================

                                ACTIONS

=============================================================================
*/


/*
===============
=
= Cmd_Fire
=
===============
*/

void Cmd_Fire (void)
{
    if(gamestate.changingweapon == true)
        return;

    buttonheld[bt_attack] = true;

    gamestate.weaponframe = 0;

    player->state = &s_attack;

    gamestate.attackframe = 0;
    gamestate.attackcount =
        attackinfo[gamestate.weapon][gamestate.attackframe].tics;
    gamestate.weaponframe =
        attackinfo[gamestate.weapon][gamestate.attackframe].frame;
}

//===========================================================================

/*
===============
=
= Cmd_Use
=
===============
*/

void Cmd_Use (void)
{
    int     checkx,checky,doornum,dir;

    //
    // find which cardinal direction the player is facing
    //
    if (player->angle < ANGLES/8 || player->angle > 7*ANGLES/8)
    {
        checkx = player->tilex + 1;
        checky = player->tiley;
        dir = di_east;
    }
    else if (player->angle < 3*ANGLES/8)
    {
        checkx = player->tilex;
        checky = player->tiley-1;
        dir = di_north;
    }
    else if (player->angle < 5*ANGLES/8)
    {
        checkx = player->tilex - 1;
        checky = player->tiley;
        dir = di_west;
    }
    else
    {
        checkx = player->tilex;
        checky = player->tiley + 1;
        dir = di_south;
    }

    doornum = tilemap[checkx][checky];
    if (MAPSPOT(checkx,checky,1) == PUSHABLETILE)
    {
        //
        // pushable wall
        //

        PushWall (checkx,checky,dir);
        return;
    }
    else if (!buttonheld[bt_use] && doornum & BIT_DOOR) {
        buttonheld[bt_use] = true;
        OperateDoor (doornum & ~BIT_DOOR);
    }
    else if (!buttonheld[bt_use] && MAPSPOT(checkx,checky,2) >= SWITCHTILE && MAPSPOT(checkx,checky,2) <= SWITCHTILE+17) {
        buttonheld[bt_use] = true;
        if (MAPSPOT(checkx,checky,2) >= SWITCHTILE && MAPSPOT(checkx,checky,2) <= SWITCHTILE+17) {
            switch (MAPSPOT(checkx,checky,2)) { // Flip the switch
                case SWITCHTILE:
                case SWITCHTILE+4:
                    MAPSPOT(checkx,checky,2)++;
                    SD_PlaySound (DSSWTCHX);
                    break;
                case SWITCHTILE+2:
                case SWITCHTILE+6:
                case SWITCHTILE+10:
                case SWITCHTILE+12:
                case SWITCHTILE+14:
                case SWITCHTILE+16:
                    MAPSPOT(checkx,checky,2)++;
                    SD_PlaySound (DSSWTCHN);
                    break;
            }

            switchmap[checkx][checky] = MAPSPOT(checkx,checky,2);
            switch (MAPSPOT(checkx,checky,1))  // activate/deactivate item
            {
                case ENDLEVELTILE:
                    playstate = ex_completed;
                    break;
                case ALTELEVATORTILE:
                    playstate = ex_secretlevel;
                    break;
                case 700: // Pushwall North
                   PushWall (pwx,pwy,di_north);
                   break;
                case 701: // Pushwall South
                   PushWall (pwx,pwy,di_south);
                   break;
                case 702: // Pushwall East
                   PushWall (pwx,pwy,di_east);
                   break;
                case 703: // Pushwall West
                   PushWall (pwx,pwy,di_west);
                   break;
                case 708: // Pushwall2 North
                   PushWall (pw2x,pw2y,di_north);
                   break;
                case 709: // Pushwall2 South
                   PushWall (pw2x,pw2y,di_south);
                   break;
                case 710: // Pushwall2 East
                   PushWall (pw2x,pw2y,di_east);
                   break;
                case 711: // Pushwall2 West
                   PushWall (pw2x,pw2y,di_west);
                   break;
                case 716: // Pushwall3 North
                   PushWall (pw3x,pw3y,di_north);
                   break;
                case 717: // Pushwall3 South
                   PushWall (pw3x,pw3y,di_south);
                   break;
                case 718: // Pushwall3 East
                   PushWall (pw3x,pw3y,di_east);
                   break;
                case 719: // Pushwall3 West
                   PushWall (pw3x,pw3y,di_west);
                   break;
                case 724: // Pushwall4 North
                   PushWall (pw4x,pw4y,di_north);
                   break;
                case 725: // Pushwall4 South
                   PushWall (pw4x,pw4y,di_south);
                   break;
                case 726: // Pushwall4 East
                   PushWall (pw4x,pw4y,di_east);
                   break;
                case 727: // Pushwall4 West
                   PushWall (pw4x,pw4y,di_west);
                   break;
             }
        }
    }
}

/*
=============================================================================

                                PLAYER CONTROL

=============================================================================
*/



/*
===============
=
= SpawnPlayer
=
===============
*/

void SpawnPlayer (int tilex, int tiley, int dir)
{
    player->obclass = playerobj;
    player->active = ac_yes;
    player->tilex = tilex;
    player->tiley = tiley;
    player->areanumber = (byte) *(mapsegs[0]+(player->tiley<<mapshift)+player->tilex);
    player->x = ((int32_t)tilex<<TILESHIFT)+TILEGLOBAL/2;
    player->y = ((int32_t)tiley<<TILESHIFT)+TILEGLOBAL/2;
    player->state = &s_player;
    player->angle = (1-dir)*90;
    if (player->angle<0)
        player->angle += ANGLES;
    player->flags = FL_NEVERMARK;
    Thrust (0,0);                           // set some variables

    InitAreas ();
}


//===========================================================================
void RocketLauncherAttack (objtype *ob)
{
    if (gamestate.ammo3<1)
        return;
    
    madenoise = true;
    SD_PlaySound (DSRLAUNC);

    GetNewActor ();

    newobj->state = &s_rocket;
    newobj->ticcount = 1;

    newobj->x = ob->x;
    newobj->y = ob->y;
    newobj->tilex = (short)(newobj->x >> TILESHIFT);
    newobj->tiley = (short)(newobj->y >> TILESHIFT);

    newobj->obclass = rocketobj;
    newobj->dir = nodir;
    newobj->angle = ob->angle;
    newobj->speed = 0x4200l;
    newobj->flags = FL_NONMARK | FL_BONUS | FL_NEVERMARK;
    newobj->active = ac_yes;
}
void PlasmaGunAttack (objtype *ob)
{
    if (gamestate.ammo4<1)
        return;
    
    madenoise = true;
    SD_PlaySound (DSPLASMA);

    GetNewActor ();

    newobj->state = &s_plasma1;
    newobj->ticcount = 1;

    newobj->x = ob->x;
    newobj->y = ob->y;
    newobj->tilex = (short)(newobj->x >> TILESHIFT);
    newobj->tiley = (short)(newobj->y >> TILESHIFT);

    newobj->obclass = plasmaobj;
    newobj->dir = nodir;
    newobj->angle = ob->angle;
    newobj->speed = 0x4200l;
    newobj->flags = FL_NONMARK | FL_BONUS | FL_NEVERMARK;
    newobj->active = ac_yes;
}
void BFGGunAttack (objtype *ob)
{
    if (gamestate.ammo4<40)
        return;
    
    madenoise = true;
    GetNewActor ();

    newobj->state = &s_bfg1;
    newobj->ticcount = 1;

    newobj->x = ob->x;
    newobj->y = ob->y;
    newobj->tilex = (short)(newobj->x >> TILESHIFT);
    newobj->tiley = (short)(newobj->y >> TILESHIFT);

    newobj->obclass = bfgobj;
    newobj->dir = nodir;
    newobj->angle = ob->angle;
    newobj->speed = 0x4200l;
    newobj->flags = FL_NONMARK | FL_BONUS | FL_NEVERMARK;
    newobj->active = ac_yes;
}

void HellOrbAttack (objtype *ob)
{
    SD_PlaySound (DSBOSPIT);

    GetNewActor ();

    newobj->state = &s_hellorb1;
    newobj->ticcount = 1;
    newobj->x = ob->x;
    newobj->y = ob->y;
    newobj->tilex = newobj->x >> TILESHIFT;
    newobj->tiley = newobj->y >> TILESHIFT;

    newobj->obclass = hellorbobj;
    newobj->dir = nodir;
    newobj->angle = ob->angle;
    newobj->speed = 0x4200l;
    newobj->flags = FL_NONMARK | FL_BONUS | FL_NEVERMARK;
    newobj->active = ac_yes;
}
/*
===============
=
= T_KnifeAttack
=
= Update player hands, and try to do damage when the proper frame is reached
=
===============
*/
int played = 0;
void    KnifeAttack (objtype *ob)
{
    objtype *check,*closest;
    int32_t  dist;
    int damage;

    // actually fire
    dist = 0x7fffffff;
    closest = NULL;
    for (check=ob->next; check; check=check->next)
    {
        if ( (check->flags & FL_SHOOTABLE) && (check->flags & FL_VISABLE)
            && abs(check->viewx-centerx) < shootdelta)
        {
            if (check->transx < dist)
            {
                dist = check->transx;
                closest = check;
            }
        }
    }

    if (!closest || dist > 0x18000l)
    {
        // missed
        if (gamestate.nextweapon==wp_chainsaw) {
            SD_PlaySound (DSSAWFUL, false);
        }
        return;
    }
    else
        {
        if (gamestate.nextweapon==wp_chainsaw)
            SD_PlaySound (DSSAWHIT, false);
        else if (gamestate.nextweapon==wp_fists)
            SD_PlaySound (DSPUNCH);
        }
    damage = US_RndT() >>4;
    // hit something
    if (gamestate.berserk)
        DamageActor (closest,damage*10);
    else
        DamageActor (closest,damage);
}

#define COLSIZE 0x4000l
#define STEPDIST 0x800l

void GunAttack (objtype *ob)
{
    switch (gamestate.nextweapon)
    {
        
        case wp_pistol:
        case wp_chaingun:
                SD_PlaySound (DSPISTOL);
                break;
        case wp_shotgun:
                SD_PlaySound (DSSHOTGN);
                break;
        case wp_supershotgun:
                SD_PlaySound (DSDSHTGN);
                break;
    }

    madenoise = true;

    int32_t         bulletx,bullety;
    short           bangle,damage,accuracy,numshots;
    boolean         hit;
    objtype         *check;

//
// AlumiuN's new weaponry code - trace bullets
//


    switch (gamestate.weapon)
    {
        case wp_pistol: accuracy = 1; numshots = 1; break;
        case wp_shotgun: accuracy = 3; numshots = 7; break;
        case wp_supershotgun: accuracy = 3; numshots = 20; break;
        case wp_chaingun: accuracy = 3; numshots = 1; break;
        case wp_bfggun: accuracy = 22; numshots = 40; break;
    }
    for (int i = 0; i<numshots; i++)
    {
        bulletx = player->x;
        bullety = player->y;
        bangle = player->angle;

        if (accuracy)
        {
            if (US_RndT() > 127)
                bangle -= US_RndT()%(accuracy + 1);
            else
                bangle += US_RndT()%(accuracy + 1);
        }

        hit = false;    // Assume the worst! Oh, and reset the variable. :)

        while(1)
        {
            //bulletx += FixedMul(STEPDIST,costable[bangle]);
            //bullety -= FixedMul(STEPDIST,sintable[bangle]);
            
            bulletx += (int32_t) (STEPDIST * cos(DEG2RAD(bangle)));
            bullety -= (int32_t) (STEPDIST * sin(DEG2RAD(bangle))); 

            //
            // check for solid walls
            //

            // ERROR: Bugs out at weird spots. Long dist... perhaps number might be too great for variable
            // POSSIBLE FIX: changed int32_t to long (no need for negative numbers)
            // check is wrong!
            check = actorat[bulletx >> TILESHIFT][bullety >> TILESHIFT];
            uintptr_t c = (uintptr_t)check;
            if (check && !ISPOINTER(check) && c != BIT_WALL)
            {
                if (c < BIT_DOOR)      // Hit a wall 
                {
                    if (gamestate.weapon != wp_bfggun)
                        T_BulletPuff (bulletx,bullety,ob);
                    hit = true;
                }
                else if (doorposition[c - BIT_DOOR] < 0xdddf/4)
                {
                     if (gamestate.weapon != wp_bfggun)
                        T_BulletPuff (bulletx,bullety,ob);
                    hit = true;
                }
            }

            if (hit)
                break;

            check = objlist;
            while (check)
            {
                if (check->flags & FL_SHOOTABLE)
                {
                    if (LABS(bulletx - check->x) < COLSIZE && LABS(bullety - check->y) < COLSIZE)
                    {
                        hit = true;   
                        switch (gamestate.weapon)
                        {
                            case wp_pistol:
                            case wp_shotgun:
                            case wp_supershotgun:
                            case wp_chaingun:
                                damage = 5 + US_RndT()%10;
                                if (check->obclass == barrelobj) {
                                    T_BulletPuff (bulletx,bullety,ob);
                                }
                                else {
                                    T_BloodSplatter (bulletx,bullety,ob);
                                }
                                break;
                            case wp_bfggun:
                                damage = 49 + US_RndT()%38;
                                T_BFGSplash (bulletx,bullety,ob);
                                break;
                        }

                        DamageActor (check, damage);
                    }
                }

                check = check->next;
            }

            if (hit) {
                break;
            }
        }
    }
}

//===========================================================================

/*
===============
=
= VictorySpin
=
===============
*/

void VictorySpin (void)
{
    int32_t    desty;

    if (player->angle > 270)
    {
        player->angle -= (short)(tics * 3);
        if (player->angle < 270)
            player->angle = 270;
    }
    else if (player->angle < 270)
    {
        player->angle += (short)(tics * 3);
        if (player->angle > 270)
            player->angle = 270;
    }

    desty = (((int32_t)player->tiley-5)<<TILESHIFT)-0x3000;

    if (player->y > desty)
    {
        player->y -= tics*4096;
        if (player->y < desty)
            player->y = desty;
    }
}


//===========================================================================

/*
===============
=
= T_Attack
=
===============
*/

void    T_Attack (objtype *ob)
{
    struct  atkinf  *cur;
    short i;

    UpdateFace ();
    
    if ( buttonstate[bt_use] && !buttonheld[bt_use] )
        buttonstate[bt_use] = false;

    if ( buttonstate[bt_attack] && !buttonheld[bt_attack])
        buttonstate[bt_attack] = false;

    ControlMovement (ob);

    plux = (word) (player->x >> UNSIGNEDSHIFT);                     // scale to fit in unsigned
    pluy = (word) (player->y >> UNSIGNEDSHIFT);
    player->tilex = (short)(player->x >> TILESHIFT);                // scale to tile values
    player->tiley = (short)(player->y >> TILESHIFT);

    //
    // change frame and fire
    //
    gamestate.attackcount -= (short) tics;
    while (gamestate.attackcount <= 0)
    {
        cur = &attackinfo[gamestate.weapon][gamestate.attackframe];
        switch (cur->attack)
        {
            case -1:
                ob->state = &s_player;
                if (gamestate.ammo2<2 && gamestate.nextweapon == wp_supershotgun) {
                    gamestate.nextweapon = gamestate.chosenweapon = wp_shotgun;
                    gamestate.changingweapon = gamestate.goingdown = true;
                }
                else if (gamestate.ammo4<40 && gamestate.nextweapon == wp_bfggun) {
                    gamestate.nextweapon = gamestate.chosenweapon = wp_plasmagun;
                    gamestate.changingweapon = gamestate.goingdown = true;
                }

                if (!ChkAtkAmmo (gamestate.weapon) && gamestate.chosenweapon != wp_fists) {
                    gamestate.nextweapon = wp_fists;
                    gamestate.changingweapon = gamestate.goingdown = true;
                    for ( i = wp_pistol; i < NUMWEAPONS; i++) {
                        if (ChkAtkAmmo (i) && gamestate.bestweapon & (1<<i))
                        {
                            gamestate.nextweapon = (weapontype) i;
                            gamestate.changingweapon = gamestate.goingdown = true;
                        }
                    }
                }
                else {
                    if (gamestate.nextweapon != gamestate.chosenweapon)
                    {
                        // What does this do?
                        // Seems to fix a broken nextweapon, never is the case :/
                        //if (ChkAtkAmmo(gamestate.chosenweapon)) {
                        //    gamestate.weapon = gamestate.nextweapon = gamestate.chosenweapon;
                        //}
                            // if nextweapon is different than current weapon, bob to next
                            //gamestate.chosenweapon = gamestate.weapon = gamestate.nextweapon; 
                        gamestate.goingdown = gamestate.changingweapon = true;
                            // this may be in the wrong spot
                        //}
                    }
                }

                gamestate.attackframe = gamestate.weaponframe = 0;
                return;

            case 4:
                if (!ChkAtkAmmo(gamestate.chosenweapon)) {
                    break;
                }

                if (buttonstate[bt_attack]) {
                    gamestate.attackframe -= 2;
                }
            case 1:
                if (!ChkAtkAmmo(wp_plasmagun) && gamestate.nextweapon == wp_plasmagun) {
                    gamestate.attackframe++;
                    break;
                }

                if (ChkAtkAmmo(gamestate.weapon))
                {
                    switch (gamestate.weapon)
                    {
                        case wp_pistol:
                        case wp_shotgun:
                        case wp_supershotgun:
                        case wp_chaingun:
                            GunAttack (ob);
                            break;
                        case wp_launcher:
                            RocketLauncherAttack (ob);
                            break;
                        case wp_plasmagun:
                            PlasmaGunAttack (ob);
                            break;
                        case wp_bfggun:
                            BFGGunAttack (ob);
                            break;
                        case wp_hellorb:
                            HellOrbAttack(ob);
                            break;
                    }

                    switch (gamestate.weapon)
                    {
                        case wp_pistol:
                        case wp_chaingun:
                            gamestate.ammo--;
                            break;
                        case wp_shotgun:
                            gamestate.ammo2--;
                            break;
                        case wp_supershotgun:
                            gamestate.ammo2-=2;
                            break;
                        case wp_launcher:
                            gamestate.ammo3--;
                            break;
                        case wp_plasmagun:
                            gamestate.ammo4--;
                            break;
                        case wp_bfggun:
                            gamestate.ammo4-=40;
                            break;
                    }

                    break;
                }
                else {
                    break;
                }

            case 2:
                KnifeAttack (ob);
                break;

            case 3:
                if (ChkAtkAmmo(gamestate.nextweapon) && buttonstate[bt_attack]) {
                    gamestate.attackframe -= 2;
                }

                break;
            case 5:
                SD_PlaySound (DSDBOPN);
                break;
            case 6:
                SD_PlaySound (DSDBLOAD);
                break;
            case 7:
                SD_PlaySound (DSDBCLS);
                break;
            case 8:
                SD_PlaySound (DSBFG);
                break;
        }

        gamestate.attackcount += cur->tics;
        gamestate.attackframe++;
        gamestate.weaponframe =
            attackinfo[gamestate.weapon][gamestate.attackframe].frame;
    }
}



//===========================================================================

/*
===============
=
= T_Player
=
===============
*/

void    T_Player (objtype *ob)
{
    if (gamestate.victoryflag)              // watching the BJ actor
    {
        VictorySpin ();
        return;
    }

    UpdateFace ();

    if ( buttonstate[bt_use] && !buttonheld[bt_use] )
        Cmd_Use ();

    if ( buttonstate[bt_attack] && !buttonheld[bt_attack])
        Cmd_Fire ();

    ControlMovement (ob);
    if (gamestate.victoryflag)              // watching the BJ actor
        return;

    plux = (word) (player->x >> UNSIGNEDSHIFT);                     // scale to fit in unsigned
    pluy = (word) (player->y >> UNSIGNEDSHIFT);
    player->tilex = (short)(player->x >> TILESHIFT);                // scale to tile values
    player->tiley = (short)(player->y >> TILESHIFT);
}
void WarpPlayer (int facedir)
{
    objtype  *check;

    buttonstate[bt_attack] = false;
    buttonheld[bt_attack] = false;
    gamestate.attackframe = 0;
    gamestate.weaponframe = 0;
    //DrawPlayerWeapon();
    ThreeDRefresh();
    switch (facedir)
    {
        case EAST:
            check = actorat[warpex][warpey];
            if (check > objlist && (check->flags & FL_SHOOTABLE))
            KillActor(check);
            //SpawnNewObj (warpex,warpey,&s_telebeam1);
            SpawnPlayer(warpex,warpey,facedir);
            break;
        case WEST:
            check = actorat[warpwx][warpwy];
            if (check > objlist && (check->flags & FL_SHOOTABLE))
            KillActor(check);
            //SpawnNewObj (warpwx,warpwy,&s_telebeam1);
            SpawnPlayer(warpwx,warpwy,facedir);
            break;
    }

    SD_PlaySound (DSTELEPT);
    //VW_WaitVBL(1*50);

    buttonstate[bt_attack] = false;
    buttonheld[bt_attack] = false;
    gamestate.attackframe = 0;
    gamestate.weaponframe = 0;
    //DrawPlayerWeapon();
    ThreeDRefresh();
}