#ifndef WL_DEF_H
#define WL_DEF_H

// Defines which version shall be built and configures supported extra features
#include "version.h"

#include <assert.h>
#include <fcntl.h>
#include <math.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#if defined(_arch_dreamcast)
#	include <string.h>
#	include "dc/dc_main.h"
#elif !defined(_WIN32)
#	include <stdint.h>
#	include <string.h>
#	include <stdarg.h>
#endif
#include <SDL.h>

#if !defined O_BINARY
#	define O_BINARY 0
#endif

#pragma pack(1)

#if defined(_arch_dreamcast)
#define YESBUTTONNAME "A"
#define NOBUTTONNAME  "B"
#elif defined(GP2X)
#define YESBUTTONNAME "Y"
#define NOBUTTONNAME  "B"
#else
#define YESBUTTONNAME "Y"
#define NOBUTTONNAME  "N"
#endif

#include "foreign.h"

#include "audiowl6.h"
#include "gfxv_wl6.h"

typedef uint8_t byte;
typedef uint16_t word;
typedef int32_t fixed;
typedef uint32_t longword;
typedef int8_t boolean;
typedef void * memptr;

typedef struct
{
    int x,y;
} Point;
typedef struct
{
    Point ul,lr;
} Rect;

void Quit(const char *errorStr, ...);

#include "id_pm.h"
#include "asmcomp.h"
#include "id_in.h"
#include "id_vl.h"
#include "id_vh.h"
#include "id_us.h"
#include "id_ca.h"

#include "wl_menu.h"

#define MAPSPOT(x,y,plane) (mapsegs[plane][((y)<<mapshift)+(x)])

#define SIGN(x)         ((x)>0?1:-1)
#define ABS(x)          ((int)(x)>0?(x):-(x))
#define LABS(x)         ((int32_t)(x)>0?(x):-(x))
#define DEG2RAD(x)     (((x)*M_PI)/180.0) 

#define abs(x) ABS(x)

/*
=============================================================================

                            GLOBAL CONSTANTS

=============================================================================
*/

extern bool allowwindow;

#define MAXTICS 10
#define DEMOTICS        4

// walls bit-shift
#define WALLSHIFT       7

#define BIT_WALL        (1<<WALLSHIFT)
#define BIT_DOOR        (1<<(WALLSHIFT+1))
#define BIT_ALLTILES    (1<<(WALLSHIFT+2))
#define LAST_WALLNUM    BIT_WALL-1
#define LAST_DOORNUM    BIT_DOOR-1

#define MAXACTORS       100000         // max number of nazis, etc / map
#define MAXSTATS        100000         // max number of lamps, bonus, etc
#define MAXDOORS        BIT_DOOR          // max number of sliding doors
#define MAXWALLTILES    BIT_WALL          // max number of wall tiles
#define MAXLOSTSOULS    15              // a pain elemental should only be able to spit out 10 (be fair)

//
// tile constants
//

#define ICONARROWS      1
#define PUSHABLETILE    9
#define EXITTILE        10          // at end of episode
#define ENDLEVELTILE	16          // end level switch
#define ALTELEVATORTILE 17          // secret level switch
#define SECRETFINDSPOT  107
#define AREATILE        107         // first of NUMAREAS floor tiles
#define NUMAREAS        37
#define AMBUSHTILE      106
#define WARPEASTTILE    11
#define WARPWESTTILE    13
#define SWITCHTILE	    2
#define PLX_MOSSY       130
#define PLX_CLOUDS      132
#define PLX_HLL_MTN     134
#define PLX_RED_MTN     136
#define PLX_HLL_COL     138
#define PLX_ORG_SKY     140
#define WATERTILE       144
#define NUKAGETILE      148
#define LAVATILE        151
#define BLOCKTILE	    18
extern byte pwx,pwy,pw2x,pw2y,pw3x,pw3y,pw4x,pw4y;

#define NUMBERCHARS     9


//----------------

#define EXTRAPOINTS     40000

#define PLAYERSPEED     3000
#define RUNSPEED        6000

#define SCREENSEG       0xa000

#define SCREENBWIDE     80

#define HEIGHTRATIO     0.50            // also defined in id_mm.c

#define BORDERCOLOR     3
#define FLASHCOLOR      5
#define FLASHTICS       4

#ifndef SPEAR
    #define LRpack      10       // # of levels to store in endgame
#else
    #define LRpack      20
#endif

#define PLAYERSIZE      MINDIST         // player radius
#define MINACTORDIST    0x10000l        // minimum dist from player center
                                        // to any actor center

#define NUMLATCHPICS    100

#define MAXAMMO		200
#define MAXAMMO2        50
#define MAXAMMO3	50
#define MAXAMMO4	300

#undef M_PI
#define PI              3.141592657
#define M_PI PI

#define GLOBAL1         (1l<<16)
#define TILEGLOBAL      GLOBAL1
#define PIXGLOBAL       (GLOBAL1/64)
#define TILESHIFT       16l
#define UNSIGNEDSHIFT   8

#define ANGLES          360             // must be divisable by 4
#define ANGLEQUAD       (ANGLES/4)
#define FINEANGLES      3600
#define ANG90           (FINEANGLES/4)
#define ANG180          (ANG90*2)
#define ANG270          (ANG90*3)
#define ANG360          (ANG90*4)
#define VANG90          (ANGLES/4)
#define VANG180         (VANG90*2)
#define VANG270         (VANG90*3)
#define VANG360         (VANG90*4)

#define MINDIST         (0x5800l)

#define mapshift        6
#define MAPSIZE         (1<<mapshift)
#define maparea         MAPSIZE*MAPSIZE

#define mapheight       MAPSIZE
#define mapwidth        MAPSIZE

#ifdef USE_HIRES

#define TEXTURESHIFT    7
#define TEXTURESIZE     (1<<TEXTURESHIFT)
#define TEXTUREFROMFIXEDSHIFT 2
#define TEXTUREMASK     (TEXTURESIZE*(TEXTURESIZE-1))

#define SPRITESCALEFACTOR 1

#else

#define TEXTURESHIFT    6
#define TEXTURESIZE     (1<<TEXTURESHIFT)
#define TEXTUREFROMFIXEDSHIFT 4
#define TEXTUREMASK     (TEXTURESIZE*(TEXTURESIZE-1))

#define SPRITESCALEFACTOR 2

#endif

#define NORTH   0
#define EAST    1
#define SOUTH   2
#define WEST    3


#define STATUSLINES     32

#define SCREENSIZE      (SCREENBWIDE*208)
#define PAGE1START      0
#define PAGE2START      (SCREENSIZE)
#define PAGE3START      (SCREENSIZE*2u)
#define FREESTART       (SCREENSIZE*3u)


#define PIXRADIUS       512

#define STARTAMMO       8


// object flag values

typedef enum
{
    FL_SHOOTABLE        = 0x00000001,
    FL_BONUS            = 0x00000002,
    FL_NEVERMARK        = 0x00000004,
    FL_VISABLE          = 0x00000008,
    FL_ATTACKMODE       = 0x00000010,
    FL_FIRSTATTACK      = 0x00000020,
    FL_AMBUSH           = 0x00000040,
    FL_NONMARK          = 0x00000080,
    FL_FULLBRIGHT       = 0x00000100,
#ifdef USE_DIR3DSPR
    // you can choose one of the following values in wl_act1.cpp
    // to make a static sprite a directional 3d sprite
    // (see example at the end of the statinfo array)
    FL_DIR_HORIZ_MID    = 0x00000200,
    FL_DIR_HORIZ_FW     = 0x00000400,
    FL_DIR_HORIZ_BW     = 0x00000600,
    FL_DIR_VERT_MID     = 0x00000a00,
    FL_DIR_VERT_FW      = 0x00000c00,
    FL_DIR_VERT_BW      = 0x00000e00,

    // these values are just used to improve readability of code
    FL_DIR_NONE         = 0x00000000,
    FL_DIR_POS_MID      = 0x00000200,
    FL_DIR_POS_FW       = 0x00000400,
    FL_DIR_POS_BW       = 0x00000600,
    FL_DIR_POS_MASK     = 0x00000600,
    FL_DIR_VERT_FLAG    = 0x00000800,
    FL_DIR_MASK         = 0x00000e00,
#endif
    // next free bit is   0x00001000
} objflag_t;


//
// sprite constants
//

enum	{
        SPR_DEMO,
        SPR_DEATHCAM,
//
// static sprites
//
        SPR_STAT_0,SPR_STAT_1,SPR_STAT_2,SPR_STAT_3,
        SPR_STAT_4,SPR_STAT_5,SPR_STAT_6,SPR_STAT_7,

        SPR_STAT_8,SPR_STAT_9,SPR_STAT_10,SPR_STAT_11,
        SPR_STAT_12,SPR_STAT_13,SPR_STAT_14,SPR_STAT_15,

        SPR_STAT_16,SPR_STAT_17,SPR_STAT_18,SPR_STAT_19,
        SPR_STAT_20,SPR_STAT_21,SPR_STAT_22,SPR_STAT_23,

        SPR_STAT_24,SPR_STAT_25,SPR_STAT_26,SPR_STAT_27,
        SPR_STAT_28,SPR_STAT_29,SPR_STAT_30,SPR_STAT_31,

        SPR_STAT_32,SPR_STAT_33,SPR_STAT_34,SPR_STAT_35,
        SPR_STAT_36,SPR_STAT_37,SPR_STAT_38,SPR_STAT_39,

        SPR_STAT_40,SPR_STAT_41,SPR_STAT_42,SPR_STAT_43,
        SPR_STAT_44,SPR_STAT_45,SPR_STAT_46,SPR_STAT_47,

        SPR_STAT_48,SPR_STAT_49,SPR_STAT_50,SPR_STAT_51,
        SPR_STAT_51A,SPR_STAT_52,SPR_STAT_53,SPR_STAT_54,

        SPR_STAT_55,SPR_STAT_56,SPR_STAT_57,SPR_STAT_58,
        SPR_STAT_59,SPR_STAT_60,SPR_STAT_61,

//
// Rocket, smoke and small explosion
//
        SPR_ROCKET_1,SPR_ROCKET_2,SPR_ROCKET_3,SPR_ROCKET_4,
        SPR_ROCKET_5,SPR_ROCKET_6,SPR_ROCKET_7,SPR_ROCKET_8,

        SPR_BOOM_1,SPR_BOOM_2,SPR_BOOM_3,

        SPR_BFG_1,SPR_BFG_2,SPR_BFG_BOOM1,SPR_BFG_BOOM2,
        SPR_BFG_BOOM3,SPR_BFG_BOOM4,SPR_BFG_BOOM5,SPR_BFG_BOOM6,

        SPR_PLASMA_1,SPR_PLASMA_2,SPR_PLASMA_BOOM1,SPR_PLASMA_BOOM2,SPR_PLASMA_BOOM3,
        SPR_PLASMA_BOOM4,SPR_PLASMA_BOOM5,

        SPR_FIRE_1,SPR_FIRE_2,SPR_FIRE_BOOM1,SPR_FIRE_BOOM2,
        SPR_FIRE_BOOM3,

        SPR_FIRE2_1,SPR_FIRE2_2,SPR_FIRE2_BOOM1,SPR_FIRE2_BOOM2,
        SPR_FIRE2_BOOM3,


//
// mutant
//
        SPR_MUTANT_W1_1,SPR_MUTANT_W1_2,SPR_MUTANT_W1_3,SPR_MUTANT_W1_4,
        SPR_MUTANT_W1_5,SPR_MUTANT_W1_6,SPR_MUTANT_W1_7,SPR_MUTANT_W1_8,

        SPR_MUTANT_W2_1,SPR_MUTANT_W2_2,SPR_MUTANT_W2_3,SPR_MUTANT_W2_4,
        SPR_MUTANT_W2_5,SPR_MUTANT_W2_6,SPR_MUTANT_W2_7,SPR_MUTANT_W2_8,

        SPR_MUTANT_W3_1,SPR_MUTANT_W3_2,SPR_MUTANT_W3_3,SPR_MUTANT_W3_4,
        SPR_MUTANT_W3_5,SPR_MUTANT_W3_6,SPR_MUTANT_W3_7,SPR_MUTANT_W3_8,

        SPR_MUTANT_W4_1,SPR_MUTANT_W4_2,SPR_MUTANT_W4_3,SPR_MUTANT_W4_4,
        SPR_MUTANT_W4_5,SPR_MUTANT_W4_6,SPR_MUTANT_W4_7,SPR_MUTANT_W4_8,

        SPR_MUTANT_PAIN_1,SPR_MUTANT_PAIN_2,SPR_MUTANT_PAIN_3,SPR_MUTANT_PAIN_4,
        SPR_MUTANT_PAIN_5,SPR_MUTANT_PAIN_6,SPR_MUTANT_PAIN_7,SPR_MUTANT_PAIN_8,

        SPR_MUTANT_DIE_1,SPR_MUTANT_DIE_2,SPR_MUTANT_DIE_3,SPR_MUTANT_DIE_4,
        SPR_MUTANT_DEAD,

        SPR_MUTANT_DIE2_1,SPR_MUTANT_DIE2_2,SPR_MUTANT_DIE2_3,SPR_MUTANT_DIE2_4,
        SPR_MUTANT_DIE2_5,SPR_MUTANT_DIE2_6,SPR_MUTANT_DIE2_7,SPR_MUTANT_DIE2_8,
        SPR_MUTANT_DIE2_9,

        SPR_MUTANT_SHOOT_1,SPR_MUTANT_SHOOT_2,SPR_MUTANT_SHOOT_3,SPR_MUTANT_SHOOT_4,
        SPR_MUTANT_SHOOT_5,SPR_MUTANT_SHOOT_6,SPR_MUTANT_SHOOT_7,SPR_MUTANT_SHOOT_8,

        SPR_MUTANT_SHOOT2_1,SPR_MUTANT_SHOOT2_2,SPR_MUTANT_SHOOT2_3,SPR_MUTANT_SHOOT2_4,
        SPR_MUTANT_SHOOT2_5,SPR_MUTANT_SHOOT2_6,SPR_MUTANT_SHOOT2_7,SPR_MUTANT_SHOOT2_8,

//
// demon
//
        SPR_DEMON_W1_1,SPR_DEMON_W1_2,SPR_DEMON_W1_3,SPR_DEMON_W1_4,
        SPR_DEMON_W1_5,SPR_DEMON_W1_6,SPR_DEMON_W1_7,SPR_DEMON_W1_8,

        SPR_DEMON_W2_1,SPR_DEMON_W2_2,SPR_DEMON_W2_3,SPR_DEMON_W2_4,
        SPR_DEMON_W2_5,SPR_DEMON_W2_6,SPR_DEMON_W2_7,SPR_DEMON_W2_8,

        SPR_DEMON_W3_1,SPR_DEMON_W3_2,SPR_DEMON_W3_3,SPR_DEMON_W3_4,
        SPR_DEMON_W3_5,SPR_DEMON_W3_6,SPR_DEMON_W3_7,SPR_DEMON_W3_8,

        SPR_DEMON_W4_1,SPR_DEMON_W4_2,SPR_DEMON_W4_3,SPR_DEMON_W4_4,
        SPR_DEMON_W4_5,SPR_DEMON_W4_6,SPR_DEMON_W4_7,SPR_DEMON_W4_8,

        SPR_DEMON_DIE_1,SPR_DEMON_DIE_2,SPR_DEMON_DIE_3,SPR_DEMON_DIE_4,
        SPR_DEMON_DIE_5,SPR_DEMON_DEAD,

        SPR_DEMON_JUMP1,SPR_DEMON_JUMP2,SPR_DEMON_JUMP3,



//
// cacodemon
//
        SPR_CACO_W1_1,SPR_CACO_W1_2,SPR_CACO_W1_3,SPR_CACO_W1_4,
        SPR_CACO_W1_5,SPR_CACO_W1_6,SPR_CACO_W1_7,SPR_CACO_W1_8,

        SPR_CACO_PAIN_1,SPR_CACO_PAIN_2,SPR_CACO_PAIN_3,SPR_CACO_PAIN_4,
        SPR_CACO_PAIN_5,SPR_CACO_PAIN_6,SPR_CACO_PAIN_7,SPR_CACO_PAIN_8,

        SPR_CACO_DIE_1,SPR_CACO_DIE_2,SPR_CACO_DIE_3,SPR_CACO_DIE_4,
        SPR_CACO_DIE_5,SPR_CACO_DEAD,

        SPR_CACO_SHOOT1,SPR_CACO_SHOOT2,SPR_CACO_SHOOT3,SPR_CACO_SHOOT4,

//
// shotgun guy
//
        SPR_SHOTGUY_W1_1,SPR_SHOTGUY_W1_2,SPR_SHOTGUY_W1_3,SPR_SHOTGUY_W1_4,
        SPR_SHOTGUY_W1_5,SPR_SHOTGUY_W1_6,SPR_SHOTGUY_W1_7,SPR_SHOTGUY_W1_8,

        SPR_SHOTGUY_W2_1,SPR_SHOTGUY_W2_2,SPR_SHOTGUY_W2_3,SPR_SHOTGUY_W2_4,
        SPR_SHOTGUY_W2_5,SPR_SHOTGUY_W2_6,SPR_SHOTGUY_W2_7,SPR_SHOTGUY_W2_8,

        SPR_SHOTGUY_W3_1,SPR_SHOTGUY_W3_2,SPR_SHOTGUY_W3_3,SPR_SHOTGUY_W3_4,
        SPR_SHOTGUY_W3_5,SPR_SHOTGUY_W3_6,SPR_SHOTGUY_W3_7,SPR_SHOTGUY_W3_8,

        SPR_SHOTGUY_W4_1,SPR_SHOTGUY_W4_2,SPR_SHOTGUY_W4_3,SPR_SHOTGUY_W4_4,
        SPR_SHOTGUY_W4_5,SPR_SHOTGUY_W4_6,SPR_SHOTGUY_W4_7,SPR_SHOTGUY_W4_8,

        SPR_SHOTGUY_PAIN_1,SPR_SHOTGUY_PAIN_2,SPR_SHOTGUY_PAIN_3,SPR_SHOTGUY_PAIN_4,
        SPR_SHOTGUY_PAIN_5,SPR_SHOTGUY_PAIN_6,SPR_SHOTGUY_PAIN_7,SPR_SHOTGUY_PAIN_8,

        SPR_SHOTGUY_DIE_1,SPR_SHOTGUY_DIE_2,SPR_SHOTGUY_DIE_3,SPR_SHOTGUY_DIE_4,
        SPR_SHOTGUY_DEAD,

        SPR_SHOTGUY_DIE2_1,SPR_SHOTGUY_DIE2_2,SPR_SHOTGUY_DIE2_3,SPR_SHOTGUY_DIE2_4,
        SPR_SHOTGUY_DIE2_5,SPR_SHOTGUY_DIE2_6,SPR_SHOTGUY_DIE2_7,SPR_SHOTGUY_DIE2_8,
        SPR_SHOTGUY_DIE2_9,

        SPR_SHOTGUY_SHOOT_1,SPR_SHOTGUY_SHOOT_2,SPR_SHOTGUY_SHOOT_3,SPR_SHOTGUY_SHOOT_4,
        SPR_SHOTGUY_SHOOT_5,SPR_SHOTGUY_SHOOT_6,SPR_SHOTGUY_SHOOT_7,SPR_SHOTGUY_SHOOT_8,

        SPR_SHOTGUY_SHOOT2_1,SPR_SHOTGUY_SHOOT2_2,SPR_SHOTGUY_SHOOT2_3,SPR_SHOTGUY_SHOOT2_4,
        SPR_SHOTGUY_SHOOT2_5,SPR_SHOTGUY_SHOOT2_6,SPR_SHOTGUY_SHOOT2_7,SPR_SHOTGUY_SHOOT2_8,

//
// chaingun guy
//
        SPR_CHAINGUY_W1_1,SPR_CHAINGUY_W1_2,SPR_CHAINGUY_W1_3,SPR_CHAINGUY_W1_4,
        SPR_CHAINGUY_W1_5,SPR_CHAINGUY_W1_6,SPR_CHAINGUY_W1_7,SPR_CHAINGUY_W1_8,

        SPR_CHAINGUY_W2_1,SPR_CHAINGUY_W2_2,SPR_CHAINGUY_W2_3,SPR_CHAINGUY_W2_4,
        SPR_CHAINGUY_W2_5,SPR_CHAINGUY_W2_6,SPR_CHAINGUY_W2_7,SPR_CHAINGUY_W2_8,

        SPR_CHAINGUY_W3_1,SPR_CHAINGUY_W3_2,SPR_CHAINGUY_W3_3,SPR_CHAINGUY_W3_4,
        SPR_CHAINGUY_W3_5,SPR_CHAINGUY_W3_6,SPR_CHAINGUY_W3_7,SPR_CHAINGUY_W3_8,

        SPR_CHAINGUY_W4_1,SPR_CHAINGUY_W4_2,SPR_CHAINGUY_W4_3,SPR_CHAINGUY_W4_4,
        SPR_CHAINGUY_W4_5,SPR_CHAINGUY_W4_6,SPR_CHAINGUY_W4_7,SPR_CHAINGUY_W4_8,

        SPR_CHAINGUY_PAIN_1,SPR_CHAINGUY_PAIN_2,SPR_CHAINGUY_PAIN_3,SPR_CHAINGUY_PAIN_4,
        SPR_CHAINGUY_PAIN_5,SPR_CHAINGUY_PAIN_6,SPR_CHAINGUY_PAIN_7,SPR_CHAINGUY_PAIN_8,

        SPR_CHAINGUY_DIE_1,SPR_CHAINGUY_DIE_2,SPR_CHAINGUY_DIE_3,SPR_CHAINGUY_DIE_4,
        SPR_CHAINGUY_DIE_5,SPR_CHAINGUY_DIE_6,SPR_CHAINGUY_DEAD,

        SPR_CHAINGUY_DIE2_1,SPR_CHAINGUY_DIE2_2,SPR_CHAINGUY_DIE2_3,SPR_CHAINGUY_DIE2_4,
        SPR_CHAINGUY_DIE2_5,SPR_CHAINGUY_DIE2_6,

        SPR_CHAINGUY_SHOOT_1,SPR_CHAINGUY_SHOOT_2,SPR_CHAINGUY_SHOOT_3,SPR_CHAINGUY_SHOOT_4,
        SPR_CHAINGUY_SHOOT_5,SPR_CHAINGUY_SHOOT_6,SPR_CHAINGUY_SHOOT_7,SPR_CHAINGUY_SHOOT_8,

        SPR_CHAINGUY_SHOOT2_1,SPR_CHAINGUY_SHOOT2_2,SPR_CHAINGUY_SHOOT2_3,SPR_CHAINGUY_SHOOT2_4,
        SPR_CHAINGUY_SHOOT2_5,SPR_CHAINGUY_SHOOT2_6,SPR_CHAINGUY_SHOOT2_7,SPR_CHAINGUY_SHOOT2_8,

//
// Imp
//
        SPR_IMP_W1_1,SPR_IMP_W1_2,SPR_IMP_W1_3,SPR_IMP_W1_4,
        SPR_IMP_W1_5,SPR_IMP_W1_6,SPR_IMP_W1_7,SPR_IMP_W1_8,

        SPR_IMP_W2_1,SPR_IMP_W2_2,SPR_IMP_W2_3,SPR_IMP_W2_4,
        SPR_IMP_W2_5,SPR_IMP_W2_6,SPR_IMP_W2_7,SPR_IMP_W2_8,

        SPR_IMP_W3_1,SPR_IMP_W3_2,SPR_IMP_W3_3,SPR_IMP_W3_4,
        SPR_IMP_W3_5,SPR_IMP_W3_6,SPR_IMP_W3_7,SPR_IMP_W3_8,

        SPR_IMP_W4_1,SPR_IMP_W4_2,SPR_IMP_W4_3,SPR_IMP_W4_4,
        SPR_IMP_W4_5,SPR_IMP_W4_6,SPR_IMP_W4_7,SPR_IMP_W4_8,

        SPR_IMP_PAIN_1,SPR_IMP_PAIN_2,SPR_IMP_PAIN_3,SPR_IMP_PAIN_4,
        SPR_IMP_PAIN_5,SPR_IMP_PAIN_6,SPR_IMP_PAIN_7,SPR_IMP_PAIN_8,

        SPR_IMP_DIE_1,SPR_IMP_DIE_2,SPR_IMP_DIE_3,SPR_IMP_DIE_4,
        SPR_IMP_DEAD,

        SPR_IMP_DIE2_1,SPR_IMP_DIE2_2,SPR_IMP_DIE2_3,SPR_IMP_DIE2_4,
        SPR_IMP_DIE2_5,SPR_IMP_DIE2_6,SPR_IMP_DIE2_7,SPR_IMP_DIE2_8,

        SPR_IMP_SHOOT_1,SPR_IMP_SHOOT_2,SPR_IMP_SHOOT_3,SPR_IMP_SHOOT_4,
        SPR_IMP_SHOOT_5,SPR_IMP_SHOOT_6,SPR_IMP_SHOOT_7,SPR_IMP_SHOOT_8,

        SPR_IMP_SHOOT2_1,SPR_IMP_SHOOT2_2,SPR_IMP_SHOOT2_3,SPR_IMP_SHOOT2_4,
        SPR_IMP_SHOOT2_5,SPR_IMP_SHOOT2_6,SPR_IMP_SHOOT2_7,SPR_IMP_SHOOT2_8,

        SPR_IMP_SHOOT3_1,SPR_IMP_SHOOT3_2,SPR_IMP_SHOOT3_3,SPR_IMP_SHOOT3_4,
        SPR_IMP_SHOOT3_5,SPR_IMP_SHOOT3_6,SPR_IMP_SHOOT3_7,SPR_IMP_SHOOT3_8,

//
// pain elemental
//
        SPR_PAIN_W1_1,SPR_PAIN_W1_2,SPR_PAIN_W1_3,SPR_PAIN_W1_4,
        SPR_PAIN_W1_5,SPR_PAIN_W1_6,SPR_PAIN_W1_7,SPR_PAIN_W1_8,

        SPR_PAIN_W2_1,SPR_PAIN_W2_2,SPR_PAIN_W2_3,SPR_PAIN_W2_4,
        SPR_PAIN_W2_5,SPR_PAIN_W2_6,SPR_PAIN_W2_7,SPR_PAIN_W2_8,

        SPR_PAIN_W3_1,SPR_PAIN_W3_2,SPR_PAIN_W3_3,SPR_PAIN_W3_4,
        SPR_PAIN_W3_5,SPR_PAIN_W3_6,SPR_PAIN_W3_7,SPR_PAIN_W3_8,

        SPR_PAIN_SHOOT_1,SPR_PAIN_SHOOT_2,SPR_PAIN_SHOOT_3,SPR_PAIN_SHOOT_4,
        SPR_PAIN_SHOOT_5,SPR_PAIN_SHOOT_6,SPR_PAIN_SHOOT_7,SPR_PAIN_SHOOT_8,

        SPR_PAIN_SHOOT2_1,SPR_PAIN_SHOOT2_2,SPR_PAIN_SHOOT2_3,SPR_PAIN_SHOOT2_4,
        SPR_PAIN_SHOOT2_5,SPR_PAIN_SHOOT2_6,SPR_PAIN_SHOOT2_7,SPR_PAIN_SHOOT2_8,

        SPR_PAIN_SHOOT3_1,SPR_PAIN_SHOOT3_2,SPR_PAIN_SHOOT3_3,SPR_PAIN_SHOOT3_4,
        SPR_PAIN_SHOOT3_5,SPR_PAIN_SHOOT3_6,SPR_PAIN_SHOOT3_7,SPR_PAIN_SHOOT3_8,

        SPR_PAIN_PAIN_1,SPR_PAIN_PAIN_2,SPR_PAIN_PAIN_3,SPR_PAIN_PAIN_4,
        SPR_PAIN_PAIN_5,SPR_PAIN_PAIN_6,SPR_PAIN_PAIN_7,SPR_PAIN_PAIN_8,

        SPR_PAIN_DIE_1,SPR_PAIN_DIE_2,SPR_PAIN_DIE_3,SPR_PAIN_DIE_4,
        SPR_PAIN_DIE_5,

//
// lost soul
//
        SPR_SOUL_W1_1,SPR_SOUL_W1_2,SPR_SOUL_W1_3,SPR_SOUL_W1_4,
        SPR_SOUL_W1_5,SPR_SOUL_W1_6,SPR_SOUL_W1_7,SPR_SOUL_W1_8,

        SPR_SOUL_W2_1,SPR_SOUL_W2_2,SPR_SOUL_W2_3,SPR_SOUL_W2_4,
        SPR_SOUL_W2_5,SPR_SOUL_W2_6,SPR_SOUL_W2_7,SPR_SOUL_W2_8,

        SPR_SOUL_PAIN_1,SPR_SOUL_PAIN_2,SPR_SOUL_PAIN_3,SPR_SOUL_PAIN_4,
        SPR_SOUL_PAIN_5,SPR_SOUL_PAIN_6,SPR_SOUL_PAIN_7,SPR_SOUL_PAIN_8,

        SPR_SOUL_DIE_1,SPR_SOUL_DIE_2,SPR_SOUL_DIE_3,SPR_SOUL_DIE_4,
        SPR_SOUL_DIE_5,

        SPR_SOUL_SHOOT_1,SPR_SOUL_SHOOT_2,SPR_SOUL_SHOOT_3,SPR_SOUL_SHOOT_4,
        SPR_SOUL_SHOOT_5,SPR_SOUL_SHOOT_6,SPR_SOUL_SHOOT_7,SPR_SOUL_SHOOT_8,

        SPR_SOUL_SHOOT2_1,SPR_SOUL_SHOOT2_2,SPR_SOUL_SHOOT2_3,SPR_SOUL_SHOOT2_4,
        SPR_SOUL_SHOOT2_5,SPR_SOUL_SHOOT2_6,SPR_SOUL_SHOOT2_7,SPR_SOUL_SHOOT2_8,
//
// cyberdemon
//
        SPR_CYBER_W1_1,SPR_CYBER_W1_2,SPR_CYBER_W1_3,SPR_CYBER_W1_4,
        SPR_CYBER_W1_5,SPR_CYBER_W1_6,SPR_CYBER_W1_7,SPR_CYBER_W1_8,

        SPR_CYBER_W2_1,SPR_CYBER_W2_2,SPR_CYBER_W2_3,SPR_CYBER_W2_4,
        SPR_CYBER_W2_5,SPR_CYBER_W2_6,SPR_CYBER_W2_7,SPR_CYBER_W2_8,

        SPR_CYBER_W3_1,SPR_CYBER_W3_2,SPR_CYBER_W3_3,SPR_CYBER_W3_4,
        SPR_CYBER_W3_5,SPR_CYBER_W3_6,SPR_CYBER_W3_7,SPR_CYBER_W3_8,

        SPR_CYBER_W4_1,SPR_CYBER_W4_2,SPR_CYBER_W4_3,SPR_CYBER_W4_4,
        SPR_CYBER_W4_5,SPR_CYBER_W4_6,SPR_CYBER_W4_7,SPR_CYBER_W4_8,

        SPR_CYBER_SHOOT_1,SPR_CYBER_SHOOT_2,SPR_CYBER_SHOOT_3,SPR_CYBER_SHOOT_4,
        SPR_CYBER_SHOOT_5,SPR_CYBER_SHOOT_6,SPR_CYBER_SHOOT_7,SPR_CYBER_SHOOT_8,

        SPR_CYBER_SHOOT2_1,SPR_CYBER_SHOOT2_2,SPR_CYBER_SHOOT2_3,SPR_CYBER_SHOOT2_4,
        SPR_CYBER_SHOOT2_5,SPR_CYBER_SHOOT2_6,SPR_CYBER_SHOOT2_7,SPR_CYBER_SHOOT2_8,

        SPR_CYBER_DEAD,

        SPR_CYBER_DIE_1,SPR_CYBER_DIE_2,SPR_CYBER_DIE_3,SPR_CYBER_DIE_4,
        SPR_CYBER_DIE_5,SPR_CYBER_DIE_6,SPR_CYBER_DIE_7,SPR_CYBER_DIE_8,

        SPR_CYBER_DIE2,SPR_CYBER_DIE3,SPR_CYBER_DIE4,SPR_CYBER_DIE5,
        SPR_CYBER_DIE6,SPR_CYBER_DIE7,SPR_CYBER_DIE8,SPR_CYBER_DIE9,

//
// spider demon
//
        SPR_SPIDER_W1_1,SPR_SPIDER_W1_2,SPR_SPIDER_W1_3,SPR_SPIDER_W1_4,
        SPR_SPIDER_W1_5,SPR_SPIDER_W1_6,SPR_SPIDER_W1_7,SPR_SPIDER_W1_8,

        SPR_SPIDER_W2_1,SPR_SPIDER_W2_2,SPR_SPIDER_W2_3,SPR_SPIDER_W2_4,
        SPR_SPIDER_W2_5,SPR_SPIDER_W2_6,SPR_SPIDER_W2_7,SPR_SPIDER_W2_8,

        SPR_SPIDER_W3_1,SPR_SPIDER_W3_2,SPR_SPIDER_W3_3,SPR_SPIDER_W3_4,
        SPR_SPIDER_W3_5,SPR_SPIDER_W3_6,SPR_SPIDER_W3_7,SPR_SPIDER_W3_8,

        SPR_SPIDER_W4_1,SPR_SPIDER_W4_2,SPR_SPIDER_W4_3,SPR_SPIDER_W4_4,
        SPR_SPIDER_W4_5,SPR_SPIDER_W4_6,SPR_SPIDER_W4_7,SPR_SPIDER_W4_8,

        SPR_SPIDER_W5_1,SPR_SPIDER_W5_2,SPR_SPIDER_W5_3,SPR_SPIDER_W5_4,
        SPR_SPIDER_W5_5,SPR_SPIDER_W5_6,SPR_SPIDER_W5_7,SPR_SPIDER_W5_8,

        SPR_SPIDER_W6_1,SPR_SPIDER_W6_2,SPR_SPIDER_W6_3,SPR_SPIDER_W6_4,
        SPR_SPIDER_W6_5,SPR_SPIDER_W6_6,SPR_SPIDER_W6_7,SPR_SPIDER_W6_8,

        SPR_SPIDER_SHOOT1_1,SPR_SPIDER_SHOOT1_2,SPR_SPIDER_SHOOT1_3,SPR_SPIDER_SHOOT1_4,
        SPR_SPIDER_SHOOT1_5,SPR_SPIDER_SHOOT1_6,SPR_SPIDER_SHOOT1_7,SPR_SPIDER_SHOOT1_8,

        SPR_SPIDER_SHOOT2_1,SPR_SPIDER_SHOOT2_2,SPR_SPIDER_SHOOT2_3,SPR_SPIDER_SHOOT2_4,
        SPR_SPIDER_SHOOT2_5,SPR_SPIDER_SHOOT2_6,SPR_SPIDER_SHOOT2_7,SPR_SPIDER_SHOOT2_8,

        SPR_SPIDER_DIE1_1,SPR_SPIDER_DIE1_2,SPR_SPIDER_DIE1_3,SPR_SPIDER_DIE1_4,
        SPR_SPIDER_DIE1_5,SPR_SPIDER_DIE1_6,SPR_SPIDER_DIE1_7,SPR_SPIDER_DIE1_8,

        SPR_SPIDER_DIE2,SPR_SPIDER_DIE3,SPR_SPIDER_DIE4,SPR_SPIDER_DIE5,
        SPR_SPIDER_DIE6,SPR_SPIDER_DIE7,SPR_SPIDER_DEAD,

//
// spider projectile
//
        SPR_SPLASMA_1,SPR_SPLASMA_2,SPR_SPLASMA_BOOM1,SPR_SPLASMA_BOOM2,
        SPR_SPLASMA_BOOM3,SPR_SPLASMA_BOOM4,SPR_SPLASMA_BOOM5,

//
// revenant
//
        SPR_REVENANT_W1_1,SPR_REVENANT_W1_2,SPR_REVENANT_W1_3,SPR_REVENANT_W1_4,
        SPR_REVENANT_W1_5,SPR_REVENANT_W1_6,SPR_REVENANT_W1_7,SPR_REVENANT_W1_8,

        SPR_REVENANT_W2_1,SPR_REVENANT_W2_2,SPR_REVENANT_W2_3,SPR_REVENANT_W2_4,
        SPR_REVENANT_W2_5,SPR_REVENANT_W2_6,SPR_REVENANT_W2_7,SPR_REVENANT_W2_8,

        SPR_REVENANT_W3_1,SPR_REVENANT_W3_2,SPR_REVENANT_W3_3,SPR_REVENANT_W3_4,
        SPR_REVENANT_W3_5,SPR_REVENANT_W3_6,SPR_REVENANT_W3_7,SPR_REVENANT_W3_8,

        SPR_REVENANT_W4_1,SPR_REVENANT_W4_2,SPR_REVENANT_W4_3,SPR_REVENANT_W4_4,
        SPR_REVENANT_W4_5,SPR_REVENANT_W4_6,SPR_REVENANT_W4_7,SPR_REVENANT_W4_8,

        SPR_REVENANT_W5_1,SPR_REVENANT_W5_2,SPR_REVENANT_W5_3,SPR_REVENANT_W5_4,
        SPR_REVENANT_W5_5,SPR_REVENANT_W5_6,SPR_REVENANT_W5_7,SPR_REVENANT_W5_8,

        SPR_REVENANT_W6_1,SPR_REVENANT_W6_2,SPR_REVENANT_W6_3,SPR_REVENANT_W6_4,
        SPR_REVENANT_W6_5,SPR_REVENANT_W6_6,SPR_REVENANT_W6_7,SPR_REVENANT_W6_8,

        SPR_REVENANT_SHOOT1_1,SPR_REVENANT_SHOOT1_2,SPR_REVENANT_SHOOT1_3,SPR_REVENANT_SHOOT1_4,
        SPR_REVENANT_SHOOT1_5,SPR_REVENANT_SHOOT1_6,SPR_REVENANT_SHOOT1_7,SPR_REVENANT_SHOOT1_8,

        SPR_REVENANT_SHOOT2_1,SPR_REVENANT_SHOOT2_2,SPR_REVENANT_SHOOT2_3,SPR_REVENANT_SHOOT2_4,
        SPR_REVENANT_SHOOT2_5,SPR_REVENANT_SHOOT2_6,SPR_REVENANT_SHOOT2_7,SPR_REVENANT_SHOOT2_8,

        SPR_REVENANT_SHOOT3_1,SPR_REVENANT_SHOOT3_2,SPR_REVENANT_SHOOT3_3,SPR_REVENANT_SHOOT3_4,
        SPR_REVENANT_SHOOT3_5,SPR_REVENANT_SHOOT3_6,SPR_REVENANT_SHOOT3_7,SPR_REVENANT_SHOOT3_8,

        SPR_REVENANT_SHOOT4_1,SPR_REVENANT_SHOOT4_2,SPR_REVENANT_SHOOT4_3,SPR_REVENANT_SHOOT4_4,
        SPR_REVENANT_SHOOT4_5,SPR_REVENANT_SHOOT4_6,SPR_REVENANT_SHOOT4_7,SPR_REVENANT_SHOOT4_8,

        SPR_REVENANT_DIE1_1,SPR_REVENANT_DIE1_2,SPR_REVENANT_DIE1_3,SPR_REVENANT_DIE1_4,
        SPR_REVENANT_DIE1_5,SPR_REVENANT_DIE1_6,SPR_REVENANT_DIE1_7,SPR_REVENANT_DIE1_8,

        SPR_REVENANT_DIE2_1,SPR_REVENANT_DIE2_2,SPR_REVENANT_DIE2_3,SPR_REVENANT_DIE2_4,
        SPR_REVENANT_DIE2_5,SPR_REVENANT_DIE2_6,SPR_REVENANT_DIE2_7,SPR_REVENANT_DIE2_8,

        SPR_REVENANT_DIE3,SPR_REVENANT_DIE4,SPR_REVENANT_DIE5,SPR_REVENANT_DIE6,
        SPR_REVENANT_DEAD,

//
// boss
//

        SPR_BOSS_W1_1,SPR_BOSS_W1_2,SPR_BOSS_W1_3,SPR_BOSS_W1_4,
        SPR_BOSS_W1_5,SPR_BOSS_W1_6,SPR_BOSS_W1_7,SPR_BOSS_W1_8,

        SPR_BOSS_W2_1,SPR_BOSS_W2_2,SPR_BOSS_W2_3,SPR_BOSS_W2_4,
        SPR_BOSS_W2_5,SPR_BOSS_W2_6,SPR_BOSS_W2_7,SPR_BOSS_W2_8,

        SPR_BOSS_W3_1,SPR_BOSS_W3_2,SPR_BOSS_W3_3,SPR_BOSS_W3_4,
        SPR_BOSS_W3_5,SPR_BOSS_W3_6,SPR_BOSS_W3_7,SPR_BOSS_W3_8,

        SPR_BOSS_W4_1,SPR_BOSS_W4_2,SPR_BOSS_W4_3,SPR_BOSS_W4_4,
        SPR_BOSS_W4_5,SPR_BOSS_W4_6,SPR_BOSS_W4_7,SPR_BOSS_W4_8,

        SPR_BOSS_SHOOT_1,SPR_BOSS_SHOOT_2,SPR_BOSS_SHOOT_3,SPR_BOSS_SHOOT_4,
        SPR_BOSS_SHOOT_5,SPR_BOSS_SHOOT_6,SPR_BOSS_SHOOT_7,SPR_BOSS_SHOOT_8,

        SPR_BOSS_SHOOT2_1,SPR_BOSS_SHOOT2_2,SPR_BOSS_SHOOT2_3,SPR_BOSS_SHOOT2_4,
        SPR_BOSS_SHOOT2_5,SPR_BOSS_SHOOT2_6,SPR_BOSS_SHOOT2_7,SPR_BOSS_SHOOT2_8,

        SPR_BOSS_SHOOT3_1,SPR_BOSS_SHOOT3_2,SPR_BOSS_SHOOT3_3,SPR_BOSS_SHOOT3_4,
        SPR_BOSS_SHOOT3_5,SPR_BOSS_SHOOT3_6,SPR_BOSS_SHOOT3_7,SPR_BOSS_SHOOT3_8,

        SPR_BOSS_DIE_1,SPR_BOSS_DIE_2,SPR_BOSS_DIE_3,SPR_BOSS_DIE_4,
        SPR_BOSS_DIE_5,SPR_BOSS_DIE_6,SPR_BOSS_DIE_7,SPR_BOSS_DIE_8,

        SPR_BOSS_DIE2,SPR_BOSS_DIE3,SPR_BOSS_DIE4,SPR_BOSS_DIE5,
        SPR_BOSS_DIE6,SPR_BOSS_DIE7,SPR_BOSS_DEAD,

//
// boss2
//

        SPR_BOSS2_W1_1,SPR_BOSS2_W1_2,SPR_BOSS2_W1_3,SPR_BOSS2_W1_4,
        SPR_BOSS2_W1_5,SPR_BOSS2_W1_6,SPR_BOSS2_W1_7,SPR_BOSS2_W1_8,

        SPR_BOSS2_W2_1,SPR_BOSS2_W2_2,SPR_BOSS2_W2_3,SPR_BOSS2_W2_4,
        SPR_BOSS2_W2_5,SPR_BOSS2_W2_6,SPR_BOSS2_W2_7,SPR_BOSS2_W2_8,

        SPR_BOSS2_W3_1,SPR_BOSS2_W3_2,SPR_BOSS2_W3_3,SPR_BOSS2_W3_4,
        SPR_BOSS2_W3_5,SPR_BOSS2_W3_6,SPR_BOSS2_W3_7,SPR_BOSS2_W3_8,

        SPR_BOSS2_W4_1,SPR_BOSS2_W4_2,SPR_BOSS2_W4_3,SPR_BOSS2_W4_4,
        SPR_BOSS2_W4_5,SPR_BOSS2_W4_6,SPR_BOSS2_W4_7,SPR_BOSS2_W4_8,

        SPR_BOSS2_SHOOT_1,SPR_BOSS2_SHOOT_2,SPR_BOSS2_SHOOT_3,SPR_BOSS2_SHOOT_4,
        SPR_BOSS2_SHOOT_5,SPR_BOSS2_SHOOT_6,SPR_BOSS2_SHOOT_7,SPR_BOSS2_SHOOT_8,

        SPR_BOSS2_SHOOT2_1,SPR_BOSS2_SHOOT2_2,SPR_BOSS2_SHOOT2_3,SPR_BOSS2_SHOOT2_4,
        SPR_BOSS2_SHOOT2_5,SPR_BOSS2_SHOOT2_6,SPR_BOSS2_SHOOT2_7,SPR_BOSS2_SHOOT2_8,

        SPR_BOSS2_SHOOT3_1,SPR_BOSS2_SHOOT3_2,SPR_BOSS2_SHOOT3_3,SPR_BOSS2_SHOOT3_4,
        SPR_BOSS2_SHOOT3_5,SPR_BOSS2_SHOOT3_6,SPR_BOSS2_SHOOT3_7,SPR_BOSS2_SHOOT3_8,

        SPR_BOSS2_DIE_1,SPR_BOSS2_DIE_2,SPR_BOSS2_DIE_3,SPR_BOSS2_DIE_4,
        SPR_BOSS2_DIE_5,SPR_BOSS2_DIE_6,SPR_BOSS2_DIE_7,SPR_BOSS2_DIE_8,

        SPR_BOSS2_DIE2,SPR_BOSS2_DIE3,SPR_BOSS2_DIE4,SPR_BOSS2_DIE5,
        SPR_BOSS2_DIE6,SPR_BOSS2_DIE7,SPR_BOSS2_DEAD,

//
// boss projectile
//

        SPR_GREEN_1,SPR_GREEN_2,SPR_GREEN_3,SPR_GREEN_4,
        SPR_GREEN_5,SPR_GREEN_6,SPR_GREEN_7,SPR_GREEN_8,

        SPR_GREEN2_1,SPR_GREEN2_2,SPR_GREEN2_3,SPR_GREEN2_4,
                SPR_GREEN2_5,SPR_GREEN2_6,SPR_GREEN2_7,SPR_GREEN2_8,

        SPR_GREEN_BOOM_1,SPR_GREEN_BOOM_2,SPR_GREEN_BOOM_3,

//
// fat mancubus
//
        SPR_MANCUBUS_W1_1,SPR_MANCUBUS_W1_2,SPR_MANCUBUS_W1_3,SPR_MANCUBUS_W1_4,
        SPR_MANCUBUS_W1_5,SPR_MANCUBUS_W1_6,SPR_MANCUBUS_W1_7,SPR_MANCUBUS_W1_8,

        SPR_MANCUBUS_W2_1,SPR_MANCUBUS_W2_2,SPR_MANCUBUS_W2_3,SPR_MANCUBUS_W2_4,
        SPR_MANCUBUS_W2_5,SPR_MANCUBUS_W2_6,SPR_MANCUBUS_W2_7,SPR_MANCUBUS_W2_8,

        SPR_MANCUBUS_W3_1,SPR_MANCUBUS_W3_2,SPR_MANCUBUS_W3_3,SPR_MANCUBUS_W3_4,
        SPR_MANCUBUS_W3_5,SPR_MANCUBUS_W3_6,SPR_MANCUBUS_W3_7,SPR_MANCUBUS_W3_8,

        SPR_MANCUBUS_W4_1,SPR_MANCUBUS_W4_2,SPR_MANCUBUS_W4_3,SPR_MANCUBUS_W4_4,
        SPR_MANCUBUS_W4_5,SPR_MANCUBUS_W4_6,SPR_MANCUBUS_W4_7,SPR_MANCUBUS_W4_8,

        SPR_MANCUBUS_W5_1,SPR_MANCUBUS_W5_2,SPR_MANCUBUS_W5_3,SPR_MANCUBUS_W5_4,
        SPR_MANCUBUS_W5_5,SPR_MANCUBUS_W5_6,SPR_MANCUBUS_W5_7,SPR_MANCUBUS_W5_8,

        SPR_MANCUBUS_W6_1,SPR_MANCUBUS_W6_2,SPR_MANCUBUS_W6_3,SPR_MANCUBUS_W6_4,
        SPR_MANCUBUS_W6_5,SPR_MANCUBUS_W6_6,SPR_MANCUBUS_W6_7,SPR_MANCUBUS_W6_8,

        SPR_MANCUBUS_SHOOT_1,SPR_MANCUBUS_SHOOT_2,SPR_MANCUBUS_SHOOT_3,SPR_MANCUBUS_SHOOT_4,
        SPR_MANCUBUS_SHOOT_5,SPR_MANCUBUS_SHOOT_6,SPR_MANCUBUS_SHOOT_7,SPR_MANCUBUS_SHOOT_8,

        SPR_MANCUBUS_SHOOT2_1,SPR_MANCUBUS_SHOOT2_2,SPR_MANCUBUS_SHOOT2_3,SPR_MANCUBUS_SHOOT2_4,
        SPR_MANCUBUS_SHOOT2_5,SPR_MANCUBUS_SHOOT2_6,SPR_MANCUBUS_SHOOT2_7,SPR_MANCUBUS_SHOOT2_8,

        SPR_MANCUBUS_DIE_1,SPR_MANCUBUS_DIE_2,SPR_MANCUBUS_DIE_3,SPR_MANCUBUS_DIE_4,
        SPR_MANCUBUS_DIE_5,SPR_MANCUBUS_DIE_6,SPR_MANCUBUS_DIE_7,SPR_MANCUBUS_DIE_8,

        SPR_MANCUBUS_DIE2_1,SPR_MANCUBUS_DIE2_2,SPR_MANCUBUS_DIE2_3,SPR_MANCUBUS_DIE2_4,
        SPR_MANCUBUS_DIE2_5,SPR_MANCUBUS_DIE2_6,SPR_MANCUBUS_DIE2_7,SPR_MANCUBUS_DIE2_8,

        SPR_MANCUBUS_DIE3,SPR_MANCUBUS_DIE4,SPR_MANCUBUS_DIE5,SPR_MANCUBUS_DIE6,
        SPR_MANCUBUS_DIE7,SPR_MANCUBUS_DIE8,SPR_MANCUBUS_DIE9,SPR_MANCUBUS_DIE10,

        SPR_MANCUBUS_DIE11,SPR_MANCUBUS_DEAD,

//
// fat projectile
//

        SPR_FAT_1,SPR_FAT_2,SPR_FAT_3,SPR_FAT_4,
        SPR_FAT_5,SPR_FAT_6,SPR_FAT_7,SPR_FAT_8,

        SPR_FAT2_1,SPR_FAT2_2,SPR_FAT2_3,SPR_FAT2_4,
        SPR_FAT2_5,SPR_FAT2_6,SPR_FAT2_7,SPR_FAT2_8,

//
// player
//
        SPR_BJ_W1_1,SPR_BJ_W1_2,SPR_BJ_W1_3,SPR_BJ_W1_4,
        SPR_BJ_W1_5,SPR_BJ_W1_6,SPR_BJ_W1_7,SPR_BJ_W1_8,

        SPR_BJ_W2_1,SPR_BJ_W2_2,SPR_BJ_W2_3,SPR_BJ_W2_4,
        SPR_BJ_W2_5,SPR_BJ_W2_6,SPR_BJ_W2_7,SPR_BJ_W2_8,

        SPR_BJ_W3_1,SPR_BJ_W3_2,SPR_BJ_W3_3,SPR_BJ_W3_4,
        SPR_BJ_W3_5,SPR_BJ_W3_6,SPR_BJ_W3_7,SPR_BJ_W3_8,

        SPR_BJ_W4_1,SPR_BJ_W4_2,SPR_BJ_W4_3,SPR_BJ_W4_4,
        SPR_BJ_W4_5,SPR_BJ_W4_6,SPR_BJ_W4_7,SPR_BJ_W4_8,

        SPR_BJ_PAIN_1,SPR_BJ_PAIN_2,SPR_BJ_PAIN_3,SPR_BJ_PAIN_4,
        SPR_BJ_PAIN_5,SPR_BJ_PAIN_6,SPR_BJ_PAIN_7,SPR_BJ_PAIN_8,

        SPR_BJ_DIE1,SPR_BJ_DIE2,SPR_BJ_DIE3,SPR_BJ_DIE4,
        SPR_BJ_DIE5,SPR_BJ_DIE6,SPR_BJ_DEAD,

/////////////
//
// WEAPONS
//
/////////////

                SPR_FISTS_READY,SPR_FISTS_READY2,SPR_FISTS_ATK1,SPR_FISTS_ATK2,
                SPR_FISTS_ATK2A,

                SPR_PISTOL_READY,SPR_PISTOL_ATK1,SPR_PISTOL_ATK2,SPR_PISTOL_ATK3,
                SPR_PISTOL_ATK4,

                SPR_SHOTGUN_READY,SPR_SHOTGUN_ATK1,SPR_SHOTGUN_ATK2,SPR_SHOTGUN_RLD1,
                SPR_SHOTGUN_RLD2,SPR_SHOTGUN_RLD3,

                SPR_SUPERSHOT_READY,SPR_SUPERSHOT_ATK1,SPR_SUPERSHOT_ATK2,SPR_SUPERSHOT_RLD1,
                SPR_SUPERSHOT_RLD2,SPR_SUPERSHOT_RLD3,SPR_SUPERSHOT_RLD4,SPR_SUPERSHOT_RLD4A,
                SPR_SUPERSHOT_RLD5,SPR_SUPERSHOT_RLD6,SPR_SUPERSHOT_RLD7,

                SPR_CHAINGUN_READY,SPR_CHAINGUN_ATK1,SPR_CHAINGUN_ATK2,SPR_CHAINGUN_ATK3,

                SPR_LAUNCHER_READY,SPR_LAUNCHER_ATK1,SPR_LAUNCHER_ATK2,SPR_LAUNCHER_ATK3,
                SPR_LAUNCHER_ATK4,SPR_LAUNCHER_ATK5,

                SPR_PLASMAGUN_READY,SPR_PLASMAGUN_ATK1,SPR_PLASMAGUN_ATK2,SPR_PLASMAGUN_ATK3,

                SPR_BFGGUN_READY,SPR_BFGGUN_READYA,SPR_BFGGUN_ATK1,SPR_BFGGUN_ATK1A,
                SPR_BFGGUN_ATK2,SPR_BFGGUN_ATK2A,SPR_BFGGUN_ATK3,SPR_BFGGUN_ATK3A,

                SPR_CHAINSAW_IDLE1,SPR_CHAINSAW_IDLE2,SPR_CHAINSAW_IDLE1A,SPR_CHAINSAW_IDLE2A,
                SPR_CHAINSAW_ATK1,SPR_CHAINSAW_ATK1A,SPR_CHAINSAW_ATK2,SPR_CHAINSAW_ATK2A,

/////////////////
//
// Rest of the stuffs
//
////////////////

        SPR_STAT2_0,SPR_STAT2_1,SPR_STAT2_2,

        SPR_BARREL,SPR_BARREL2,SPR_BARRELBOOM1,SPR_BARRELBOOM2,
        SPR_BARRELBOOM3,

        SPR_SPLAT1,SPR_SPLAT2,SPR_SPLAT3,

        SPR_PUFF1,SPR_PUFF2,SPR_PUFF3,SPR_PUFF4,

        SPR_BFGSPLAT1,SPR_BFGSPLAT2,SPR_BFGSPLAT3,SPR_BFGSPLAT4,

        SPR_STAT3_0,SPR_STAT3_1,SPR_STAT3_2,SPR_STAT3_3,
        SPR_STAT3_4,SPR_STAT3_5,SPR_STAT3_6,SPR_STAT3_7,

        SPR_STAT3_8,SPR_STAT3_9,SPR_STAT3_10,SPR_STAT3_11,
        SPR_STAT3_12,SPR_STAT3_13,SPR_STAT3_14,SPR_STAT3_15,

        SPR_STAT3_16,SPR_STAT3_17,SPR_STAT3_18,SPR_STAT3_19,
        SPR_STAT3_20,SPR_STAT3_21,SPR_STAT3_22,SPR_STAT3_23,

        SPR_STAT3_24,SPR_STAT3_25,SPR_STAT3_26,SPR_STAT3_27,
        SPR_STAT3_28,SPR_STAT3_29,SPR_STAT3_30,SPR_STAT3_31,

        SPR_STAT3_32,SPR_STAT3_33,SPR_STAT3_34,SPR_STAT3_35,
        SPR_STAT3_36,SPR_STAT3_37,SPR_STAT3_38,SPR_STAT3_39,

        SPR_STAT3_40,SPR_STAT3_41,SPR_STAT3_42,SPR_STAT3_43,
        SPR_STAT3_44,SPR_STAT3_45,SPR_STAT3_46,SPR_STAT3_47,

        SPR_STAT3_48,SPR_STAT3_49,SPR_STAT3_50,SPR_STAT3_51,
        SPR_STAT3_52,SPR_STAT3_53,SPR_STAT3_54,SPR_STAT3_55,

        SPR_STAT3_56,SPR_STAT3_57,SPR_STAT3_58,SPR_STAT3_59,
        SPR_STAT3_60,SPR_STAT3_61,SPR_STAT3_62,SPR_STAT3_63,

        SPR_STAT3_64,SPR_TOILET,SPR_EXIT,SPR_BIOSUIT,
        SPR_HELLORB,SPR_HELLORB2,SPR_HELLORB3,SPR_HELLORB4,

        SPR_VILE_W1_1,SPR_VILE_W1_2,SPR_VILE_W1_3,SPR_VILE_W1_4,
        SPR_VILE_W1_5,SPR_VILE_W1_6,SPR_VILE_W1_7,SPR_VILE_W1_8,

        SPR_VILE_W2_1,SPR_VILE_W2_2,SPR_VILE_W2_3,SPR_VILE_W2_4,
        SPR_VILE_W2_5,SPR_VILE_W2_6,SPR_VILE_W2_7,SPR_VILE_W2_8,

        SPR_VILE_W3_1,SPR_VILE_W3_2,SPR_VILE_W3_3,SPR_VILE_W3_4,
        SPR_VILE_W3_5,SPR_VILE_W3_6,SPR_VILE_W3_7,SPR_VILE_W3_8,

        SPR_VILE_W4_1,SPR_VILE_W4_2,SPR_VILE_W4_3,SPR_VILE_W4_4,
        SPR_VILE_W4_5,SPR_VILE_W4_6,SPR_VILE_W4_7,SPR_VILE_W4_8,

        SPR_VILE_W5_1,SPR_VILE_W5_2,SPR_VILE_W5_3,SPR_VILE_W5_4,
        SPR_VILE_W5_5,SPR_VILE_W5_6,SPR_VILE_W5_7,SPR_VILE_W5_8,

        SPR_VILE_W6_1,SPR_VILE_W6_2,SPR_VILE_W6_3,SPR_VILE_W6_4,
        SPR_VILE_W6_5,SPR_VILE_W6_6,SPR_VILE_W6_7,SPR_VILE_W6_8,

        SPR_VILE_SHOOT1_1,SPR_VILE_SHOOT1_2,SPR_VILE_SHOOT1_3,SPR_VILE_SHOOT1_4,
        SPR_VILE_SHOOT1_5,SPR_VILE_SHOOT1_6,SPR_VILE_SHOOT1_7,SPR_VILE_SHOOT1_8,

        SPR_VILE_SHOOT2_1,SPR_VILE_SHOOT2_2,SPR_VILE_SHOOT2_3,SPR_VILE_SHOOT2_4,
        SPR_VILE_SHOOT2_5,SPR_VILE_SHOOT2_6,SPR_VILE_SHOOT2_7,SPR_VILE_SHOOT2_8,

        SPR_VILE_SHOOT3_1,SPR_VILE_SHOOT3_2,SPR_VILE_SHOOT3_3,SPR_VILE_SHOOT3_4,
        SPR_VILE_SHOOT3_5,SPR_VILE_SHOOT3_6,SPR_VILE_SHOOT3_7,SPR_VILE_SHOOT3_8,

        SPR_VILE_SHOOT4_1,SPR_VILE_SHOOT4_2,SPR_VILE_SHOOT4_3,SPR_VILE_SHOOT4_4,
        SPR_VILE_SHOOT4_5,SPR_VILE_SHOOT4_6,SPR_VILE_SHOOT4_7,SPR_VILE_SHOOT4_8,

        SPR_VILE_SHOOT5_1,SPR_VILE_SHOOT5_2,SPR_VILE_SHOOT5_3,SPR_VILE_SHOOT5_4,
        SPR_VILE_SHOOT5_5,SPR_VILE_SHOOT5_6,SPR_VILE_SHOOT5_7,SPR_VILE_SHOOT5_8,

        SPR_VILE_SHOOT6_1,SPR_VILE_SHOOT6_2,SPR_VILE_SHOOT6_3,SPR_VILE_SHOOT6_4,
        SPR_VILE_SHOOT6_5,SPR_VILE_SHOOT6_6,SPR_VILE_SHOOT6_7,SPR_VILE_SHOOT6_8,

        SPR_VILE_SHOOT7_1,SPR_VILE_SHOOT7_2,SPR_VILE_SHOOT7_3,SPR_VILE_SHOOT7_4,
        SPR_VILE_SHOOT7_5,SPR_VILE_SHOOT7_6,SPR_VILE_SHOOT7_7,SPR_VILE_SHOOT7_8,

        SPR_VILE_SHOOT8_1,SPR_VILE_SHOOT8_2,SPR_VILE_SHOOT8_3,SPR_VILE_SHOOT8_4,
        SPR_VILE_SHOOT8_5,SPR_VILE_SHOOT8_6,SPR_VILE_SHOOT8_7,SPR_VILE_SHOOT8_8,

        SPR_VILE_SHOOT9_1,SPR_VILE_SHOOT9_2,SPR_VILE_SHOOT9_3,SPR_VILE_SHOOT9_4,
        SPR_VILE_SHOOT9_5,SPR_VILE_SHOOT9_6,SPR_VILE_SHOOT9_7,SPR_VILE_SHOOT9_8,

        SPR_VILE_SHOOT10_1,SPR_VILE_SHOOT10_2,SPR_VILE_SHOOT10_3,SPR_VILE_SHOOT10_4,
        SPR_VILE_SHOOT10_5,SPR_VILE_SHOOT10_6,SPR_VILE_SHOOT10_7,SPR_VILE_SHOOT10_8,

        SPR_VILE_DIE1_1,SPR_VILE_DIE1_2,SPR_VILE_DIE1_3,SPR_VILE_DIE1_4,
        SPR_VILE_DIE1_5,SPR_VILE_DIE1_6,SPR_VILE_DIE1_7,SPR_VILE_DIE1_8,

        SPR_VILE_DIE2,SPR_VILE_DIE3,SPR_VILE_DIE4,SPR_VILE_DIE5,
        SPR_VILE_DIE6,SPR_VILE_DIE7,SPR_VILE_DIE8,SPR_VILE_DIE9,

        SPR_VILE_DEAD,

        SPR_VILE_FIRE1,SPR_VILE_FIRE2,SPR_VILE_FIRE3,SPR_VILE_FIRE4,
        SPR_VILE_FIRE5,SPR_VILE_FIRE6,SPR_VILE_FIRE7,SPR_VILE_FIRE8,

        SPR_REVBALL_1,SPR_REVBALL_2,SPR_REVBALL_3,SPR_REVBALL_4,
        SPR_REVBALL_5,SPR_REVBALL_6,SPR_REVBALL_7,SPR_REVBALL_8,

        SPR_REVBALL2_1,SPR_REVBALL2_2,SPR_REVBALL2_3,SPR_REVBALL2_4,
        SPR_REVBALL2_5,SPR_REVBALL2_6,SPR_REVBALL2_7,SPR_REVBALL2_8,

        SPR_REVBALL_BOOM1,SPR_REVBALL_BOOM2,SPR_REVBALL_BOOM3,

        SPR_HELLORB_READY,SPR_HELLORB_READY2,SPR_HELLORB_READY3,SPR_HELLORB_READY4,
        SPR_HELLORB_ATK1,SPR_HELLORB_ATK2,SPR_HELLORB_ATK3,SPR_HELLORB_ATK4,


        SPR_HELLORB_ATK5,SPR_HELLORB_ATK6,SPR_HELLORB_ATK7,SPR_HELLORB_ATK8,
        SPR_HELLBALL1,SPR_HELLBALL2,SPR_HELLBALL3,SPR_HELLBALL4,

};


/*
=============================================================================

                               GLOBAL TYPES

=============================================================================
*/

typedef enum {
    di_north,
    di_east,
    di_south,
    di_west
} controldir_t;

typedef enum {
    dr_normal,
    dr_lock1,
    dr_lock2,
    dr_lock3,
    dr_lock4,
    dr_elevator,
    dr_green,
    dr_brown,
} door_t;

typedef enum {
    ac_badobject = -1,
    ac_no,
    ac_yes,
    ac_allways
} activetype;

typedef enum {
    nothing,
    playerobj,
    inertobj,
    mutantobj,
    chainguyobj,
    cacoobj,
    demonobj,
    cyberobj,
    spiderobj,
    revenantobj,
    shotguyobj,
    impobj,
    painobj,
    soulobj,
    bjobj,
    bossobj,
    boss2obj,
    mancubusobj,
    vileobj,
    rocketobj,
    plasmaobj,
    bossballobj,
    fatballobj, 
    revballobj,
    bfgobj,
    splatobj,
    puffobj,
    barrelobj,
    impballobj,
    cacoballobj,
    splasmaobj,
    vilefireobj,
    hellorbobj,
    bfgsplashobj,
    cyberrocketobj
} classtype;

typedef enum {
    none,
    block,
    bo_key1,
    bo_key2,
    bo_key3,
    bo_ammo1,
    bo_ammoused,
    bo_ammo1box,
    bo_ammo2,
    bo_ammo2box,
    bo_ammo3,
    bo_ammo3box,
    bo_ammo4,
    bo_ammo4box,
    bo_shotgun,
    bo_supershotgun,
    bo_chaingun,
    bo_launcher,
    bo_plasmagun,
    bo_bfggun,
    bo_chainsaw,
    bo_megaarmor,
    bo_armor,
    bo_armor1,
    bo_health1,
    bo_medpack,
    bo_healthkit,
    bo_berserk,
    bo_backpack,
    bo_godmode,
    bo_megasphere,
    bo_goggles,
    bo_cloak,
    bo_soul,
    bo_biosuit,
    bo_hellorb,
    bo_secret
} wl_stat_t;

typedef enum {
    east,
    northeast,
    north,
    northwest,
    west,
    southwest,
    south,
    southeast,
    nodir
} dirtype;


typedef enum {
    en_mutant,
    en_chaingunguy,
    en_cacodemon,
    en_demon,
    en_cyber,
    en_spider,
    en_revenant,
    en_shotgunguy,
    en_imp,
    en_painelemental,
    en_lostsoul,
    en_mancubus,
    en_barrel,
    en_boss,
    en_boss2,
    en_vile,
    NUMENEMIES
} enemy_t;

typedef void (* statefunc) (void *);

typedef struct statestruct
{
    boolean rotate;
    short   shapenum;           // a shapenum of -1 means get from ob->temp1
    short   tictime;
    void    (*think) (void *),(*action) (void *);
    struct  statestruct *next;
} statetype;


//---------------------
//
// trivial actor structure
//
//---------------------

typedef struct statstruct
{
    byte      tilex,tiley;
    short     shapenum;           // if shapenum == -1 the obj has been removed
    byte      *visspot;
    uint32_t  flags;
    byte      itemnumber;
} statobj_t;


//---------------------
//
// door actor structure
//
//---------------------

typedef enum
{
    dr_open,dr_closed,dr_opening,dr_closing
} doortype;

typedef struct doorstruct
{
    byte     tilex,tiley;
    boolean  vertical;
    byte     lock;
    doortype action;
    short    ticcount;
} doorobj_t;


//--------------------
//
// thinking actor structure
//
//--------------------

typedef struct objstruct
{
    activetype  active;
    short       ticcount;
    classtype   obclass;
    statetype   *state;

    uint32_t    flags;              // FL_SHOOTABLE, etc

    int32_t     distance;           // if negative, wait for that door to open
    dirtype     dir;

    fixed       x,y;
    word        tilex,tiley;
    byte        areanumber;

    short       viewx;
    word        viewheight;
    fixed       transx,transy;      // in global coord

    short       angle;
    short       hitpoints;
    int32_t     speed;

    short       temp1,temp2,hidden;
    struct objstruct *next,*prev;
} objtype;

enum
{
    bt_nobutton=-1,
    bt_attack=0,
    bt_strafe,
    bt_run,
    bt_use,
    bt_readyknife,
    bt_readypistol,
    bt_readymachinegun,
    bt_readychaingun,
    bt_nextweapon,
    bt_prevweapon,
    bt_esc,
    bt_pause,
    bt_strafeleft,
    bt_straferight,
    bt_moveforward,
    bt_movebackward,
    bt_turnleft,
    bt_turnright,
    NUMBUTTONS
};


typedef enum {
    wp_fists,         // 1
    wp_chainsaw,      // 2
    wp_pistol,        // 4
    wp_shotgun,       // 8
    wp_supershotgun,  // 16
    wp_chaingun,      // 32
    wp_launcher,      // 64
    wp_plasmagun,     // 128
    wp_bfggun,        // 256
    wp_hellorb,       // 512
    NUMWEAPONS
} weapontype;

typedef enum {
    ammo1,
    ammo2,
    ammo3,
    ammo4
} ammotype;

typedef enum {
    p_godmode,
    p_berserk,
    p_biosuit,
    p_invisibility,
    p_lightamp
} bonustype;

#define BONUS_TIME 30*70

#define NUMEPISODES     4
enum
{
    gd_baby,
    gd_easy,
    gd_medium,
    gd_hard,
    NUMDIFFICULTIES
};

//---------------
//
// gamestate structure
//
//---------------

typedef struct
{
    short       difficulty;
    short       mapon;
    int32_t     oldscore,score,nextextra;
    short       lives;
    short       health;
    short		armor;
    short		ammo,ammomax, ammo2,ammo2max, ammo3,ammo3max, ammo4,ammo4max;
    short       keys;
    weapontype  bestweapon,weapon,chosenweapon,nextweapon;
    bool		changingweapon, goingdown;
    short		weapchange;
    boolean		berserk, startinvistrigger, godmode, bonusgodmode, lightamp, biosuit, invisibility;
    long		godmodecount, berserkcount, lightampcount, biosuitcount, invisibilitycount;

    short       faceframe;
    short       attackframe,attackcount,weaponframe;

    short       episode,secretcount,treasurecount,killcount,
                secrettotal,treasuretotal,killtotal;
    int32_t     TimeCount;
    int32_t     killx,killy;
    boolean     victoryflag;            // set during victory animations
    char        message[48];
     
    short		bobber, bobber2, bobdir2, bobdir;
    short       lostsoulcount;
} gametype;

/*
 * Original - start with 10 bullets and only a pistol
 * Ask - Ask the player if they want to start from last save
 * Last Save - Just restart the level from the last save
*/
typedef enum {
    do_original,
    do_ask,
    do_lastsave
} deathopt;

typedef struct {
    boolean     autosave;
    deathopt    deathOption;
} optiontype;

typedef enum
{
    ex_stillplaying,
    ex_completed,
    ex_died,
    ex_warped,
    ex_resetgame,
    ex_loadedgame,
    ex_victorious,
    ex_abort,
    ex_demodone,
    ex_secretlevel
} exit_t;


extern word *mapsegs[MAPPLANES];
extern int mapon;

/*
=============================================================================

                             WL_MAIN DEFINITIONS

=============================================================================
*/

extern  boolean  loadedgame;
extern  fixed    focallength;
extern  int      viewscreenx, viewscreeny;
extern  int      viewwidth;
extern  int      viewheight;
extern  short    centerx;
extern  int32_t  heightnumerator;
extern  fixed    scale;

extern  int      dirangle[9];

extern  int      mouseadjustment;
extern  int      shootdelta;
extern  unsigned screenofs;

extern  boolean  startgame;
extern  char     str[80];
extern  char     configdir[256];
extern  char     configname[13];

//
// Command line parameter variables
//
extern  boolean  param_debugmode;
extern  boolean  param_nowait;
extern  int      param_difficulty;
extern  int      param_tedlevel;
extern  int      param_joystickindex;
extern  int      param_joystickhat;
extern  int      param_samplerate;
extern  int      param_audiobuffer;
extern  int      param_mission;
extern  boolean  param_goodtimes;
extern  boolean  param_ignorenumchunks;


void            NewGame (int difficulty,int episode);
void            CalcProjection (int32_t focal);
void            NewViewSize (int width);
boolean         SetViewSize (unsigned width, unsigned height);
boolean         LoadTheGame(FILE *file,int x,int y);
boolean         SaveTheGame(FILE *file,int x,int y);
void            ShowViewSize (int width);
void            ShutdownId (void);


/*
=============================================================================

                         WL_GAME DEFINITIONS

=============================================================================
*/

extern  gametype        gamestate;
extern optiontype      optionstate;
extern char    lastsavedgame[13];
extern  SDL_Surface     *latchpics[NUMLATCHPICS];
extern  char            demoname[13];

void    SetupGameLevel (void);
void    GameLoop (void);
void    DrawPlayScreen (void);
void    ShowActStatus();

void    PlayDemo (int demonumber);
void    RecordDemo (void);

extern  int32_t            spearx,speary;
extern  unsigned        spearangle;
extern  boolean         spearflag;

#define ClearMemory SD_StopDigitized


// JAB
#define PlaySoundLocTile(s,tx,ty)       PlaySoundLocGlobal(s,(((int32_t)(tx) << TILESHIFT) + (1L << (TILESHIFT - 1))),(((int32_t)ty << TILESHIFT) + (1L << (TILESHIFT - 1))))
#define PlaySoundLocActor(s,ob, repeatable)         PlaySoundLocGlobal(s,(ob)->x,(ob)->y, repeatable)
void    PlaySoundLocGlobal(word s,fixed gx,fixed gy, bool repeatable = true);
void UpdateSoundLoc(void);


/*
=============================================================================

                            WL_PLAY DEFINITIONS

=============================================================================
*/

#define BASEMOVE                35
#define RUNMOVE                 70
#define BASETURN                35
#define RUNTURN                 70

#define JOYSCALE                2

extern  uint16_t            tilemap[MAPSIZE][MAPSIZE];      // wall values only
extern  byte            spotvis[MAPSIZE][MAPSIZE];
extern  objtype         *actorat[MAPSIZE][MAPSIZE];
extern  uint16_t switchmap[MAPSIZE][MAPSIZE]; // switches on plane 3

extern  objtype         *player;

extern  unsigned        tics;
extern  int             viewsize;

extern  int             lastgamemusicoffset;

//
// current user input
//
extern  int         controlx,controly;              // range from -100 to 100
extern  boolean     buttonstate[NUMBUTTONS];
extern  objtype     objlist[MAXACTORS];
extern  boolean     buttonheld[NUMBUTTONS];
extern  exit_t      playstate;
extern  boolean     madenoise;
extern  statobj_t   statobjlist[MAXSTATS];
extern  statobj_t   *laststatobj;
extern  objtype     *newobj,*killerobj;
extern  doorobj_t   doorobjlist[MAXDOORS];
extern  doorobj_t   *lastdoorobj;

extern  boolean     demorecord,demoplayback;
extern  int8_t      *demoptr, *lastdemoptr;
extern  memptr      demobuffer;

//
// control info
//
extern  boolean     mouseenabled,joystickenabled;
extern  int         dirscan[4];
extern  int         buttonscan[NUMBUTTONS];
extern  int         buttonmouse[4];
extern  int         buttonjoy[32];

void    InitActorList (void);
void    GetNewActor (void);
void    PlayLoop (void);

void    CenterWindow(word w,word h);

void    InitRedShifts (void);
void    FinishPaletteShifts (void);

void    RemoveObj (objtype *gone);
void    PollControls (void);
int     StopMusic(void);
void    StartMusic(void);
void    ContinueMusic(int offs);
void    StartDamageFlash (int damage);
void    StartBonusFlash (void);
void	GodInvertFlash (void);
void	StartBiosuitFlash (void);


#ifdef SPEAR
extern  int32_t     funnyticount;           // FOR FUNNY BJ FACE
#endif

extern  objtype     *objfreelist;     // *obj,*player,*lastobj,

void TimedBonus (int bonustype, int time);

extern  boolean     noclip;
extern  int         singlestep, extravbls;

/*
=============================================================================

                                WL_INTER

=============================================================================
*/

void IntroScreen (void);
void PG13(void);
void DrawHighScores(void);
void CheckHighScore (int32_t score,word other);
void Victory (void);
void LevelCompleted (void);
void EndScrollText (void);
void ClearSplitVWB (void);

void PreloadGraphics(void);


/*
=============================================================================

                                WL_DEBUG

=============================================================================
*/

int DebugKeys (void);

/*
=============================================================================

                            WL_DRAW DEFINITIONS

=============================================================================
*/

//
// math tables
//
extern  short *pixelangle;
extern  int32_t finetangent[FINEANGLES/4];
extern  fixed sintable[];
extern  fixed *costable;
extern  int *wallheight;
extern  word horizwall[],vertwall[];
extern  int32_t    lasttimecount;
extern  int32_t    frameon;

extern  unsigned screenloc[3];

extern  boolean fizzlein, fpscounter;

extern  fixed   viewx,viewy;                    // the focal point
extern  fixed   viewsin,viewcos;

void    ThreeDRefresh (void);
void    CalcTics (void);
void LSWrite (int x, int y, const char *string);
int LSWrite_GetLength (const char *string);
void   GetMessage (char *lastmessage);
void   DrawMessage (void);  

typedef struct
{
    word leftpix,rightpix;
    word dataofs[64];
// table data after dataofs[rightpix-leftpix+1]
} t_compshape;

/*
=============================================================================

                             WL_STATE DEFINITIONS

=============================================================================
*/
#define TURNTICS        10
#define SPDPATROL       512
#define SPDDOG          1500


void    InitHitRect (objtype *ob, unsigned radius);
void    SpawnNewObj (unsigned tilex, unsigned tiley, statetype *state);
void    NewState (objtype *ob, statetype *state);

boolean TryWalk (objtype *ob);
void    SelectChaseDir (objtype *ob);
void    SelectDodgeDir (objtype *ob);
void    SelectRunDir (objtype *ob);
void    MoveObj (objtype *ob, int32_t move);
boolean SightPlayer (objtype *ob);

void    KillActor (objtype *ob);
void    DamageActor (objtype *ob, unsigned damage);

boolean CheckLine (objtype *ob);
boolean CheckSight (objtype *ob);

/*
=============================================================================

                             WL_AGENT DEFINITIONS

=============================================================================
*/

extern  short    anglefrac;
extern  int      facecount, facetimes;
extern  word     plux,pluy;         // player coordinates scaled to unsigned
extern  int32_t  thrustspeed;
extern  objtype  *LastAttacker;
extern	int warpwx,warpwy,warpex,warpey;

void    Thrust (int angle, int32_t speed);
void    CheckWeaponChange();
void    SpawnPlayer (int tilex, int tiley, int dir);
void    TakeDamage (int points,objtype *attacker);
void    GivePoints (int32_t points);
void    GetBonus (statobj_t *check);
void	GiveArmor (int points);
void    GiveWeapon (int weapon, int giveammo);
void    GiveAmmo (int ammo, int ammotype);
void    GiveKey (int key);

//
// player state info
//

void    StatusDrawFace(unsigned picnum);
void    DrawFace (void);
void    DrawHealth (void);
void    HealSelf (int points);
void	DrawArmor (void);
void    DrawLevel (void);
void    DrawLives (void);
void    GiveExtraMan (void);
void    DrawScore (void);
void    DrawWeapon (void);
void    DrawKeys (void);
void    DrawAmmo (void);
void	FullScreenDraw (void);


/*
=============================================================================

                             WL_ACT1 DEFINITIONS

=============================================================================
*/

extern  doorobj_t doorobjlist[MAXDOORS];
extern  doorobj_t *lastdoorobj;
extern  short     doornum;

extern  word      doorposition[MAXDOORS];

extern  byte      areaconnect[NUMAREAS][NUMAREAS];

extern  boolean   areabyplayer[NUMAREAS];

extern word     pwallstate;
extern word     pwallpos;        // amount a pushable wall has been moved (0-63)
extern word     pwallx,pwally;
extern word     pwalldir,pwalltile;


void InitDoorList (void);
void InitStaticList (void);
void SpawnStatic (int tilex, int tiley, int type);
void SpawnDoor (int tilex, int tiley, boolean vertical, int lock);
void MoveDoors (void);
void MovePWalls (void);
void OpenDoor (int door);
void PlaceItemType (int itemtype, int tilex, int tiley);
void PushWall (int checkx, int checky, int dir);
void OperateDoor (int door);
void InitAreas (void);

/*
=============================================================================

                             WL_ACT2 DEFINITIONS

=============================================================================
*/

#define s_nakedbody s_static10

#define INVISIBLEACCURACY 20

extern	statetype s_mutantdie1;
extern	statetype s_mutantexpl1;
extern	statetype s_demondie1;
extern	statetype s_chainguydie1;
extern	statetype s_chainguyexpl1;
extern	statetype s_shotguydie1;
extern	statetype s_shotguyexpl1;
extern	statetype s_impdie1;
extern	statetype s_impexpl1;
extern	statetype s_paindie1;
extern	statetype s_souldie1;
extern	statetype s_cacodie1;
extern	statetype s_cyberdie1;
extern	statetype s_spiderdie1;
extern	statetype s_revenantdie1;
extern	statetype s_bossdie1;
extern	statetype s_boss2die1;
extern	statetype s_mancubusdie1;
extern	statetype s_viledie1;
extern	statetype s_barreldie1;

extern	statetype s_mutantchase1;
extern	statetype s_demonchase1;
extern	statetype s_chainguychase1;
extern	statetype s_cacochase1;
extern	statetype s_impchase1;
extern	statetype s_painchase1;
extern	statetype s_soulchase1;
extern	statetype s_shotguychase1;
extern	statetype s_cyberchase1;
extern	statetype s_spiderchase1;
extern	statetype s_revenantchase1;
extern	statetype s_mancubuschase1;
extern	statetype s_vilechase1;

extern	statetype s_bosschase1;
extern	statetype s_boss2chase1;

extern	statetype s_mutantpain;
extern	statetype s_demonpain;
extern	statetype s_chainguypain;
extern	statetype s_cacopain;
extern	statetype s_imppain;
extern	statetype s_painpain;
extern	statetype s_soulpain;
extern	statetype s_shotguypain;

extern	statetype s_impballboom1;
extern	statetype s_bossballboom1;
extern	statetype s_cacoballboom1;
extern	statetype s_fatballboom1;
extern	statetype s_splasmaboom1;
extern	statetype s_vilefire1;
extern	statetype s_revballboom1;

void SpawnStand (enemy_t which, int tilex, int tiley, int dir);
void SpawnPatrol (enemy_t which, int tilex, int tiley, int dir);
void SpawnBoss (int tilex, int tiley, enemy_t which);
void KillActor (objtype *ob);

void SpawnDeadGuard (int tilex, int tiley);
void SpawnBarrel (int tilex, int tiley);
void T_BulletPuff (int x,int y,objtype *ob);
void T_BloodSplatter (int x,int y,objtype *ob);
void T_BFGSplash (int x,int y,objtype *ob);
void GunAttack (objtype *ob);

/*
=============================================================================

                             WL_TEXT DEFINITIONS

=============================================================================
*/

extern  char    helpfilename[];

extern  void    HelpScreens(void);


/*
=============================================================================

                               GP2X DEFINITIONS

=============================================================================
*/

#if defined(GP2X)

#if defined(GP2X_940)
void GP2X_MemoryInit(void);
void GP2X_Shutdown(void);
#endif
void GP2X_ButtonDown(int button);
void GP2X_ButtonUp(int button);

#endif


/*
=============================================================================

                             MISC DEFINITIONS

=============================================================================
*/

static inline fixed FixedMul(fixed a, fixed b)
{
    return (fixed)(((int64_t)a * b + 0x8000) >> 16);
}

#ifdef PLAYDEMOLIKEORIGINAL
    #define DEMOCHOOSE_ORIG_SDL(orig, sdl) ((demorecord || demoplayback) ? (orig) : (sdl))
    #define DEMOCOND_ORIG                  (demorecord || demoplayback)
    #define DEMOIF_SDL                     if(DEMOCOND_SDL)
#else
    #define DEMOCHOOSE_ORIG_SDL(orig, sdl) (sdl)
    #define DEMOCOND_ORIG                  false
    #define DEMOIF_SDL
#endif
#define DEMOCOND_SDL                   (!DEMOCOND_ORIG)

#define GetTicks() ((SDL_GetTicks()*7)/100)

#define ISPOINTER(x) ((((uintptr_t)(x)) & ~0xffff) != 0)

#define CHECKMALLOCRESULT(x) if(!(x)) Quit("Out of memory at %s:%i", __FILE__, __LINE__)

#ifdef _WIN32
    #define strcasecmp stricmp
    #define strncasecmp strnicmp
    #define snprintf _snprintf
#else
    static inline char* itoa(int value, char* string, int radix)
    {
        sprintf(string, "%d", value);
        return string;
    }

    static inline char* ltoa(long value, char* string, int radix)
    {
        sprintf(string, "%ld", value);
        return string;
    }
#endif

#define lengthof(x) (sizeof(x) / sizeof(*(x)))
#define endof(x)    ((x) + lengthof(x))

static inline word READWORD(byte *&ptr)
{
    word val = ptr[0] | ptr[1] << 8;
    ptr += 2;
    return val;
}

static inline longword READLONGWORD(byte *&ptr)
{
    longword val = ptr[0] | ptr[1] << 8 | ptr[2] << 16 | ptr[3] << 24;
    ptr += 4;
    return val;
}


/*
=============================================================================

                           FEATURE DEFINITIONS

=============================================================================
*/

#ifdef USE_FEATUREFLAGS
    // The currently available feature flags
    #define FF_STARSKY      0x0001
    #define FF_PARALLAXSKY  0x0002
    #define FF_CLOUDSKY     0x0004
    #define FF_RAIN         0x0010
    #define FF_SNOW         0x0020

    // The ffData... variables contain the 16-bit values of the according corners of the current level.
    // The corners are overwritten with adjacent tiles after initialization in SetupGameLevel
    // to avoid interpretation as e.g. doors.
    extern int ffDataTopLeft, ffDataTopRight, ffDataBottomLeft, ffDataBottomRight;

    /*************************************************************
     * Current usage of ffData... variables:
     * ffDataTopLeft:     lower 8-bit: ShadeDefID
     * ffDataTopRight:    FeatureFlags
     * ffDataBottomLeft:  CloudSkyDefID or ParallaxStartTexture
     * ffDataBottomRight: unused
     *************************************************************/

    // The feature flags are stored as a wall in the upper right corner of each level
    static inline word GetFeatureFlags()
    {
        return ffDataTopRight;
    }

#endif

#ifdef USE_FLOORCEILINGTEX
    void DrawFloorAndCeiling(byte *vbuf, unsigned vbufPitch, int min_wallheight);
#endif

#ifdef USE_PARALLAX
    void DrawParallax(byte *vbuf, unsigned vbufPitch);
#endif

#ifdef USE_DIR3DSPR
    void Scale3DShape(byte *vbuf, unsigned vbufPitch, statobj_t *ob);
#endif

#endif
    
extern byte soundvol, musicvol;
extern bool reversestereo;


////////////////////////////////////////////////////////////////////
//
// WL_LOGGER DEFS
//
////////////////////////////////////////////////////////////////////
extern void writeLog(char* funcMessage, char* message);

extern char * levelNameList[];