////////////////////////////////////////////////////////////////////
//
// WL_MENU.C
// by John Romero (C) 1992 Id Software, Inc.
//
////////////////////////////////////////////////////////////////////

#include <sys/stat.h>
#include <sys/types.h>
#ifdef _WIN32
    #include <io.h>
    #include <direct.h>
#else
    #include <unistd.h>
#endif

#include "wl_def.h"
#include <time.h>
#pragma hdrstop

extern int lastgamemusicoffset;
extern int numEpisodesMissing;

#define STARTITEM       newgame

bool autosaveAvailable = false;

// ENDSTRx constants are defined in foreign.h
char endStrings[6][80] = {
    ENDSTR1,
    ENDSTR2,
    ENDSTR3,
    ENDSTR4,
    ENDSTR5,
    ENDSTR6
};

CP_itemtype MainMenu[] = {
    {1,"",CP_NewGame},
    {1,"",CP_Options},
    {1,"",CP_LoadGame},
    {0,"",CP_SaveGame},
    {1,"",CP_ViewScores},
    {1,"",0},
    {1,"",0}
};

CP_itemtype OptMenu[]={
    {1,"",CP_Sound},
    {1,"",CP_Control},
    {1,"",CP_ChangeView},
    {1,"",0},
    {1,"", 0}
};

CP_itemtype SndMenu[] = {
    {1,"",0},
    {0,"",0},
    {1,"",0},
    {0,"",0},
    {1,"",0}
};

enum { CTL_MOUSEENABLE, CTL_MOUSESENS, CTL_JOYENABLE, CTL_CUSTOMIZE };

CP_itemtype CtlMenu[] = {
    {0,"",0},
    {0,"",MouseSensitivity},
    {0,"",0},
    {1,"",CustomControls}
};

CP_itemtype NewEmenu[] = {
    {1,"",0},
    {1,"",0},
    {1,"",0},
    {1,"",0}
};

CP_itemtype NewMenu[] = {
    {1,"",0},
    {1,"",0},
    {1,"",0},
    {1,"",0}
};

CP_itemtype LSMenu[] = {
    {1, "", 0},
    {1, "", 0},
    {1, "", 0},
    {1, "", 0},
    {1, "", 0},
    {1, "", 0},
    {1, "", 0},
    {1, "", 0},
    {1, "", 0},
    {0, "", 0}
};

CP_itemtype CusMenu[] = {
    {1,"",0},
    {0,"",0},
    {0,"",0},
    {1,"",0},
    {0,"",0},
    {0,"",0},
    {1,"",0},
    {0,"",0},
    {1,"",0}
};

// CP_iteminfo struct format: short x, y, amount, curpos, indent;
CP_iteminfo MainItems = { MENU_X, MENU_Y + 5, lengthof(MainMenu), STARTITEM, 24, 16 },
            SndItems  = { 46,52, lengthof(SndMenu), 0, 24, 16 },
            LSItems   = { LSM_X, LSM_Y - 3, lengthof(LSMenu), 0, 24, 16 },
            CtlItems  = { CTL_X + 16, CTL_Y - 44, lengthof(CtlMenu), -1, 24, 16 },
            CusItems  = { CST_X, CST_Y + 24, lengthof(CusMenu), -1, 0, 13},
            NewEitems = { NE_X + 8, NE_Y + 47, lengthof(NewEmenu), 0, 24, 16 },
            NewItems  = { NM_X - 4, NM_Y + 2, lengthof(NewMenu), 2, 24, 16 },
            OptItems  = { OPT_X, OPT_Y, lengthof(OptMenu), 0, 24, 16 };

int color_hlite[] = {
    DEACTIVE,
    HIGHLIGHT,
    READHCOLOR,
    0x67
};

int color_norml[] = {
    DEACTIVE,
    TEXTCOLOR,
    READCOLOR,
    0x6b
};

int EpisodeSelect[NUMEPISODES] = { 1 };


static int SaveGamesAvail[10];
static int StartGame;
static int SoundStatus = 1;
static int pickquick;
static char SaveGameNames[9][32];
// The level is + 1 so 0 is the default with nothing
static int SaveGameLevel[10];
static char SaveGameTime[10][20];
static char SaveName[13] = "savegam?.";


////////////////////////////////////////////////////////////////////
//
// INPUT MANAGER SCANCODE TABLES
//
////////////////////////////////////////////////////////////////////

#if 0
static const char *ScanNames[] =      // Scan code names with single chars
{
    "?", "?", "1", "2", "3", "4", "5", "6", "7", "8", "9", "0", "-", "+", "?", "?",
    "Q", "W", "E", "R", "T", "Y", "U", "I", "O", "P", "[", "]", "|", "?", "A", "S",
    "D", "F", "G", "H", "J", "K", "L", ";", "\"", "?", "?", "?", "Z", "X", "C", "V",
    "B", "N", "M", ",", ".", "/", "?", "?", "?", "?", "?", "?", "?", "?", "?", "?",
    "?", "?", "?", "?", "?", "?", "?", "?", "\xf", "?", "-", "\x15", "5", "\x11", "+", "?",
    "\x13", "?", "?", "?", "?", "?", "?", "?", "?", "?", "?", "?", "?", "?", "?", "?",
    "?", "?", "?", "?", "?", "?", "?", "?", "?", "?", "?", "?", "?", "?", "?", "?",
    "?", "?", "?", "?", "?", "?", "?", "?", "?", "?", "?", "?", "?", "?", "?", "?"
};                              // DEBUG - consolidate these
static ScanCode ExtScanCodes[] =        // Scan codes with >1 char names
{
    1, 0xe, 0xf, 0x1d, 0x2a, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e,
    0x3f, 0x40, 0x41, 0x42, 0x43, 0x44, 0x57, 0x59, 0x46, 0x1c, 0x36,
    0x37, 0x38, 0x47, 0x49, 0x4f, 0x51, 0x52, 0x53, 0x45, 0x48,
    0x50, 0x4b, 0x4d, 0x00
};
static const char *ExtScanNames[] =   // Names corresponding to ExtScanCodes
{
    "Esc", "BkSp", "Tab", "Ctrl", "LShft", "Space", "CapsLk", "F1", "F2", "F3", "F4",
    "F5", "F6", "F7", "F8", "F9", "F10", "F11", "F12", "ScrlLk", "Enter", "RShft",
    "PrtSc", "Alt", "Home", "PgUp", "End", "PgDn", "Ins", "Del", "NumLk", "Up",
    "Down", "Left", "Right", ""
};

/*#pragma warning 737 9
static byte
                                        *ScanNames[] =          // Scan code names with single chars
                                        {
        "?","?","1","2","3","4","5","6","7","8","9","0","-","+","?","?",
        "Q","W","E","R","T","Y","U","I","O","P","[","]","|","?","A","S",
        "D","F","G","H","J","K","L",";","\"","?","?","?","Z","X","C","V",
        "B","N","M",",",".","/","?","?","?","?","?","?","?","?","?","?",
        "?","?","?","?","?","?","?","?","\xf","?","-","\x15","5","\x11","+","?",
        "\x13","?","?","?","?","?","?","?","?","?","?","?","?","?","?","?",
        "?","?","?","?","?","?","?","?","?","?","?","?","?","?","?","?",
        "?","?","?","?","?","?","?","?","?","?","?","?","?","?","?","?"
                                        };      // DEBUG - consolidate these
static byte ExtScanCodes[] =    // Scan codes with >1 char names
                                        {
        1,0xe,0xf,0x1d,0x2a,0x39,0x3a,0x3b,0x3c,0x3d,0x3e,
        0x3f,0x40,0x41,0x42,0x43,0x44,0x57,0x59,0x46,0x1c,0x36,
        0x37,0x38,0x47,0x49,0x4f,0x51,0x52,0x53,0x45,0x48,
        0x50,0x4b,0x4d,0x00
                                        };
static byte *ExtScanNames[] =   // Names corresponding to ExtScanCodes
                                        {
        "Esc","BkSp","Tab","Ctrl","LShft","Space","CapsLk","F1","F2","F3","F4",
        "F5","F6","F7","F8","F9","F10","F11","F12","ScrlLk","Enter","RShft",
        "PrtSc","Alt","Home","PgUp","End","PgDn","Ins","Del","NumLk","Up",
        "Down","Left","Right",""
                                        };*/

#else
static const char* const ScanNames[SDLK_LAST] =
    {
        "?","?","?","?","?","?","?","?",                                //   0
        "BkSp","Tab","?","?","?","Return","?","?",                      //   8
        "?","?","?","Pause","?","?","?","?",                            //  16
        "?","?","?","Esc","?","?","?","?",                              //  24
        "Space","!","\"","#","$","?","&","'",                           //  32
        "(",")","*","+",",","-",".","/",                                //  40
        "0","1","2","3","4","5","6","7",                                //  48
        "8","9",":",";","<","=",">","?",                                //  56
        "@","A","B","C","D","E","F","G",                                //  64
        "H","I","J","K","L","M","N","O",                                //  72
        "P","Q","R","S","T","U","V","W",                                //  80
        "X","Y","Z","[","\\","]","^","_",                               //  88
        "`","a","b","c","d","e","f","h",                                //  96
        "h","i","j","k","l","m","n","o",                                // 104
        "p","q","r","s","t","u","v","w",                                // 112
        "x","y","z","{","|","}","~","?",                                // 120
        "?","?","?","?","?","?","?","?",                                // 128
        "?","?","?","?","?","?","?","?",                                // 136
        "?","?","?","?","?","?","?","?",                                // 144
        "?","?","?","?","?","?","?","?",                                // 152
        "?","?","?","?","?","?","?","?",                                // 160
        "?","?","?","?","?","?","?","?",                                // 168
        "?","?","?","?","?","?","?","?",                                // 176
        "?","?","?","?","?","?","?","?",                                // 184
        "?","?","?","?","?","?","?","?",                                // 192
        "?","?","?","?","?","?","?","?",                                // 200
        "?","?","?","?","?","?","?","?",                                // 208
        "?","?","?","?","?","?","?","?",                                // 216
        "?","?","?","?","?","?","?","?",                                // 224
        "?","?","?","?","?","?","?","?",                                // 232
        "?","?","?","?","?","?","?","?",                                // 240
        "?","?","?","?","?","?","?","?",                                // 248
        "?","?","?","?","?","?","?","?",                                // 256
        "?","?","?","?","?","?","?","Enter",                            // 264
        "?","Up","Down","Right","Left","Ins","Home","End",              // 272
        "PgUp","PgDn","F1","F2","F3","F4","F5","F6",                    // 280
        "F7","F8","F9","F10","F11","F12","?","?",                       // 288
        "?","?","?","?","NumLk","CapsLk","ScrlLk","RShft",              // 296
        "Shift","RCtrl","Ctrl","RAlt","Alt","?","?","?",                // 304
        "?","?","?","?","PrtSc","?","?","?",                            // 312
        "?","?"                                                         // 320
    };

#endif

////////////////////////////////////////////////////////////////////
//
// Wolfenstein Control Panel!  Ta Da!
//
////////////////////////////////////////////////////////////////////
void
US_ControlPanel (ScanCode scancode)
{
    int which;

#ifdef _arch_dreamcast
    DC_StatusClearLCD();
#endif

    if (ingame)
    {
        if (CP_CheckQuick (scancode))
            return;
        lastgamemusicoffset = StartCPMusic (MENUSONG);
    }
    else
        StartCPMusic (MENUSONG);
    SetupControlPanel ();

    //
    // F-KEYS FROM WITHIN GAME
    //
    switch (scancode)
    {
       // case sc_F1:
        //    HelpScreens ();
        //    goto finishup;

        case sc_F2:
            CP_SaveGame (0);
            goto finishup;

        case sc_F3:
            CP_LoadGame (0);
            goto finishup;

        case sc_F4:
            CP_Sound (0);
            goto finishup;

        case sc_F5:
            CP_ChangeView (0);
            goto finishup;

        case sc_F6:
            CP_Control (0);
            goto finishup;

        finishup:
            CleanupControlPanel ();
#ifdef SPEAR
            UnCacheLump (OPTIONS_LUMP_START, OPTIONS_LUMP_END);
#endif
            return;
    }

#ifdef SPEAR
    CacheLump (OPTIONS_LUMP_START, OPTIONS_LUMP_END);
#endif

    DrawMainMenu ();
    MenuFadeIn ();
    StartGame = 0;

    //
    // MAIN MENU LOOP
    //
    do
    {
        which = HandleMenu (&MainItems, &MainMenu[0], NULL);

        switch (which)
        {
            case viewscores:
                if (MainMenu[viewscores].routine == NULL)
                {
                    if (CP_EndGame (0))
                        StartGame = 1;
                }
                else
                {
                    DrawMainMenu();
                    MenuFadeIn ();
                }
                break;

            case backtodemo:
                StartGame = 1;
                if (!ingame)
                    StartCPMusic (INTROSONG);
                VL_FadeOut (0, 255, 0, 0, 0, 10);
                break;

            case -1:
            case quit:
                CP_Quit (0);
                break;

            default:
                if (!StartGame)
                {
                    DrawMainMenu ();
                    MenuFadeIn ();
                }
        }

        //
        // "EXIT OPTIONS" OR "NEW GAME" EXITS
        //
    }
    while (!StartGame);

    //
    // DEALLOCATE EVERYTHING
    //
    CleanupControlPanel ();

    //
    // CHANGE MAINMENU ITEM
    //
    if (startgame || loadedgame)
        EnableEndGameMenuItem();

    // RETURN/START GAME EXECUTION

#ifdef SPEAR
    UnCacheLump (OPTIONS_LUMP_START, OPTIONS_LUMP_END);
#endif
}

void EnableEndGameMenuItem()
{
    MainMenu[viewscores].routine = NULL;
#ifndef JAPAN
    //strcpy (MainMenu[viewscores].string, STR_EG);
#endif
}

////////////////////////
//
// DRAW MAIN MENU SCREEN
//
void
DrawMainMenu (void)
{
    int startheight = 60;
    int curspace = 16;
    ClearMScreen();
    CacheLump(MAINMENU_LUMP_START,MAINMENU_LUMP_END);
    CacheLump(MESSAGE_LUMP_START,MESSAGE_LUMP_END);
    VWB_DrawPic(98,0,M_DOOMPIC);
    VWB_DrawPic(100,startheight + curspace * newgame,M_NEWGAMEPIC);
    VWB_DrawPic(100,startheight + curspace * options,M_OPTIONSPIC);
    VWB_DrawPic(100,startheight + curspace * loadgame,M_LOADGAMEPIC);
    VWB_DrawPic(100,startheight + curspace * savegame,M_SAVEGAMEPIC);
    VWB_DrawPic(100,startheight + curspace * backtodemo,M_BACKTOPIC);
    if (ingame)
    {
        VWB_DrawPic(100,startheight + curspace * viewscores,M_ENDGAMEPIC);
        VWB_DrawPic(196,startheight + curspace * backtodemo,M_GAMEPIC);
    }
    else
    {
        VWB_DrawPic(100,startheight + curspace * viewscores,M_VIEWSCORESPIC);
        VWB_DrawPic(196,startheight + curspace * backtodemo,M_DEMOPIC);
    }
    VWB_DrawPic(100,startheight + curspace * quit,M_QUITGAMEPIC);

    // VERSION
    LSWrite(260,4, STR_VERSION);
    if (ingame) {
        MainMenu[backtodemo].active=2;
    }
    else {
        MainMenu[backtodemo].active=1;
    }
    DrawMenu(&MainItems,&MainMenu[0]);
    UnCacheLump(MAINMENU_LUMP_START,MAINMENU_LUMP_END);
    UnCacheLump(MESSAGE_LUMP_START,MESSAGE_LUMP_END);
    VW_UpdateScreen();
}

int CP_Options(int)
{
    int which;
    DrawOptionsMenu();
    MenuFadeIn();
    WaitKeyUp();
    do
    {
        which=HandleMenu(&OptItems,OptMenu,NULL);

        switch(which) {
            case 3: // Autosave
                optionstate.autosave ^= 1;
                ShootSnd();
                DrawOptionsMenu();
                WaitKeyUp ();
                break;
            case 4: // Death Option
                switch (optionstate.deathOption) {
                    case do_original:
                        optionstate.deathOption = do_ask;
                        break;
                    case do_ask:
                        optionstate.deathOption = do_lastsave;
                        break;
                    default: // do_lastsave
                        optionstate.deathOption = do_original;
                }
                ShootSnd();
                DrawOptionsMenu();
                WaitKeyUp ();
                break;
            default:
                DrawOptionsMenu();
                MenuFadeIn();
                WaitKeyUp();
                break;
        }
    } while(which>=0);
    MenuFadeOut();
    return 0;
}

void DrawOptionsMenu(void)
{
    ClearMScreen();
    CA_CacheGrChunk(M_OPTIONSPIC);
    CacheLump (OPTIONS_LUMP_START,OPTIONS_LUMP_END);
    VWB_DrawPic(114,28,M_OPTIONSPIC);
    VWB_DrawPic(OPT_T,52,O_SOUNDPIC);
    VWB_DrawPic(OPT_T,68,O_CONTROLPIC);
    VWB_DrawPic(OPT_T,84,O_SCREENSIZEPIC);
    VWB_DrawPic(OPT_T,100,O_AUTOSAVEPIC);
    VWB_DrawPic(OPT_T,116,O_DEATHOPTIONPIC);
    VWB_DrawPic (OPT_T+126,103, optionstate.autosave ? O_SELECTEDPIC : O_NOTSELECTEDPIC);
    switch (optionstate.deathOption) {
        case do_original:
            VWB_DrawPic (OPT_T+118,116, O_ORIGINALPIC);
            break;
        case do_ask:
            VWB_DrawPic (OPT_T+118,116, O_ASKPIC);
            break;
        case do_lastsave:
            VWB_DrawPic (OPT_T+118,116, O_LASTSAVEPIC);
            break;
    }

    DrawMenu(&OptItems,OptMenu);
    UNCACHEGRCHUNK (M_OPTIONSPIC);
    UnCacheLump (OPTIONS_LUMP_START,OPTIONS_LUMP_END);
    VW_UpdateScreen();
}


////////////////////////////////////////////////////////////////////
//
// CHECK QUICK-KEYS & QUIT (WHILE IN A GAME)
//
////////////////////////////////////////////////////////////////////
int CP_CheckQuick (ScanCode scancode)
{
    switch (scancode)
    {
        //
        // END GAME
        //
        case sc_F7:
            CA_CacheGrChunk (STARTFONT + 1);

            WindowH = 160;
#ifdef JAPAN
            if (GetYorN (7, 8, C_JAPQUITPIC))
#else
            if (Confirm (ENDGAMESTR))
#endif
            {
                playstate = ex_died;
                killerobj = NULL;
                pickquick = gamestate.lives = 0;
            }

            WindowH = 200;
            fontnumber = 0;
            MainMenu[savegame].active = 0;
            return 1;

        //
        // QUICKSAVE
        //
        case sc_F8:
            VL_SetSaveGameSlot();
            if (SaveGamesAvail[LSItems.curpos] && pickquick)
            {
                CA_CacheGrChunk (STARTFONT + 1);
                fontnumber = 1;
                Message (STR_SAVING "...");
                CP_SaveGame (1);
                fontnumber = 0;
            }
            else
            {
#ifndef SPEAR
                CA_CacheGrChunk (STARTFONT + 1);
                CA_CacheGrChunk (C_DISKLOADING1PIC);
                CA_CacheGrChunk (C_DISKLOADING2PIC);
                // TODO: Change
                //CA_CacheGrChunk (C_SAVEGAMEPIC);
                //CA_CacheGrChunk (C_MOUSELBACKPIC);
#else
                CacheLump (BACKDROP_LUMP_START, BACKDROP_LUMP_END);
                CA_CacheGrChunk (C_CURSOR1PIC);
#endif

                VW_FadeOut ();
                if(screenHeight % 200 != 0)
                    VL_ClearScreen(0);

                lastgamemusicoffset = StartCPMusic (MENUSONG);
                pickquick = CP_SaveGame (0);

                SETFONTCOLOR (0, 15);
                IN_ClearKeysDown ();
                VW_FadeOut();

                if (!startgame && !loadedgame)
                    ContinueMusic (lastgamemusicoffset);

                if (loadedgame)
                    playstate = ex_abort;
                lasttimecount = GetTimeCount ();

                if (MousePresent && IN_IsInputGrabbed())
                    IN_CenterMouse();     // Clear accumulated mouse movement

#ifndef SPEAR
                UNCACHEGRCHUNK (C_DISKLOADING1PIC);
                UNCACHEGRCHUNK (C_DISKLOADING2PIC);
                // TODO: Change
                //UNCACHEGRCHUNK (C_SAVEGAMEPIC);
                //UNCACHEGRCHUNK (C_MOUSELBACKPIC);
#else
                UnCacheLump (BACKDROP_LUMP_START, BACKDROP_LUMP_END);
#endif
            }
            return 1;

        //
        // QUICKLOAD
        //
        case sc_F9:
            if (SaveGamesAvail[LSItems.curpos] && pickquick)
            {
                char string[100] = STR_LGC;


                CA_CacheGrChunk (STARTFONT + 1);
                fontnumber = 1;

                strcat (string, SaveGameNames[LSItems.curpos]);
                strcat (string, "\"?");

                if (Confirm (string))
                    CP_LoadGame (1);

                fontnumber = 0;
            }
            else
            {
#ifndef SPEAR
                CA_CacheGrChunk (STARTFONT + 1);
                CA_CacheGrChunk (C_DISKLOADING1PIC);
                CA_CacheGrChunk (C_DISKLOADING2PIC);
                // TODO: Change
                //CA_CacheGrChunk (C_LOADGAMEPIC);
                //CA_CacheGrChunk (C_MOUSELBACKPIC);
#else
                CA_CacheGrChunk (C_CURSOR1PIC);
                CacheLump (BACKDROP_LUMP_START, BACKDROP_LUMP_END);
#endif

                VW_FadeOut ();
                if(screenHeight % 200 != 0)
                    VL_ClearScreen(0);

                lastgamemusicoffset = StartCPMusic (MENUSONG);
                pickquick = CP_LoadGame (0);    // loads lastgamemusicoffs

                SETFONTCOLOR (0, 15);
                IN_ClearKeysDown ();
                VW_FadeOut();

                if (!startgame && !loadedgame)
                    ContinueMusic (lastgamemusicoffset);

                if (loadedgame)
                    playstate = ex_abort;

                lasttimecount = GetTimeCount ();

                if (MousePresent && IN_IsInputGrabbed())
                    IN_CenterMouse();     // Clear accumulated mouse movement

#ifndef SPEAR
                UNCACHEGRCHUNK (C_DISKLOADING1PIC);
                UNCACHEGRCHUNK (C_DISKLOADING2PIC);
                // TODO: Change
                //UNCACHEGRCHUNK (C_LOADGAMEPIC);
                //UNCACHEGRCHUNK (C_MOUSELBACKPIC);
#else
                UnCacheLump (BACKDROP_LUMP_START, BACKDROP_LUMP_END);
#endif
            }
            return 1;

        //
        // QUIT
        //
        case sc_F10:
            CA_CacheGrChunk (STARTFONT + 1);

            WindowX = WindowY = 0;
            WindowW = 320;
            WindowH = 160;
#ifdef JAPAN
            if (GetYorN (7, 8, C_QUITMSGPIC))
#else
#ifdef SPANISH
            if (Confirm (ENDGAMESTR))
#else
            int rand = US_RndT();
            int len = lengthof(endStrings);
            if (Confirm (endStrings[rand % len]))
#endif
#endif
            {
                VW_UpdateScreen ();
                SD_MusicOff ();
                SD_StopSound ();
                MenuFadeOut ();

                Quit (NULL);
            }

            WindowH = 200;
            fontnumber = 0;
            return 1;
    }

    return 0;
}


////////////////////////////////////////////////////////////////////
//
// END THE CURRENT GAME
//
////////////////////////////////////////////////////////////////////
int
CP_EndGame (int)
{
    int res;
#ifdef JAPAN
    res = GetYorN (7, 8, C_JAPQUITPIC);
#else
    res = Confirm (ENDGAMESTR);
#endif
    DrawMainMenu();
    if(!res) return 0;

    pickquick = gamestate.lives = 0;
    playstate = ex_died;
    killerobj = NULL;

    MainMenu[savegame].active = 0;
    MainMenu[viewscores].routine = CP_ViewScores;
#ifndef JAPAN
//    strcpy (MainMenu[viewscores].string, STR_VS);
#endif

    return 1;
}


////////////////////////////////////////////////////////////////////
//
// VIEW THE HIGH SCORES
//
////////////////////////////////////////////////////////////////////
int CP_ViewScores (int)
{
    fontnumber = 0;

    DrawHighScores ();
    VW_UpdateScreen ();
    MenuFadeIn ();
    fontnumber = 1;

    IN_Ack ();

    MenuFadeOut ();
    return 0;
}


////////////////////////////////////////////////////////////////////
//
// START A NEW GAME
//
////////////////////////////////////////////////////////////////////
int CP_NewGame (int)
{
    int which, episode;

  firstpart:

    DrawNewEpisode ();
    do
    {
        which = HandleMenu (&NewEitems, &NewEmenu[0], NULL);
        switch (which)
        {
            case -1:
                MenuFadeOut ();
                return 0;

            default:
                episode = which;
                which = 1;
                break;
        }

    }
    while (!which);

    ShootSnd ();

    //
    // ALREADY IN A GAME?
    //
    if (ingame) {
        if (!Confirm (CURGAME))
        {
            goto firstpart;
            //MenuFadeOut ();
            //return 0;
        }
    }

    MenuFadeOut ();

    DrawNewGame ();
    which = HandleMenu (&NewItems, NewMenu, NULL);
    if (which < 0)
    {
        MenuFadeOut ();
        goto firstpart;
    }

    ShootSnd ();
    NewGame (which, episode);
    StartGame = 1;
    MenuFadeOut ();

    pickquick = 0;
    return 0;
}


/////////////////////
//
// DRAW NEW EPISODE MENU
//
void DrawNewEpisode (void)
{
    int i;

    ClearMScreen ();
    CacheLump (EPISODE_LUMP_START,EPISODE_LUMP_END);
    
    WindowX = 0;
    VWB_DrawPic (84,52,E_WHICHEPISODEPIC);

    DrawMenu (&NewEitems, NewEmenu);
    
    for (i = 0; i < NUMEPISODES; i++) {
        VWB_DrawPic(NE_X+32,NE_Y+48+i*16,E_EPISODE1PIC+i);
    }

    UnCacheLump (EPISODE_LUMP_START,EPISODE_LUMP_END);

    VW_UpdateScreen ();
    MenuFadeIn ();
    WaitKeyUp ();
}

/////////////////////
//
// DRAW NEW GAME MENU
//
void DrawNewGame (void)
{
    ClearMScreen();

    CacheLump (SKILL_LUMP_START,SKILL_LUMP_END);

    VWB_DrawPic (NM_X + 20,36,S_SKILLLEVELPIC);
    VWB_DrawPic (NM_X + 20,63,S_SKILL1PIC);
    VWB_DrawPic (NM_X + 20,79,S_SKILL2PIC);
    VWB_DrawPic (NM_X + 20,95,S_SKILL3PIC);
    VWB_DrawPic (NM_X + 20,111,S_SKILL4PIC);
    DrawMenu(&NewItems,NewMenu);

    VW_UpdateScreen ();
    MenuFadeIn ();
    WaitKeyUp ();
}

////////////////////////////////////////////////////////////////////
//
// HANDLE SOUND MENU
//
////////////////////////////////////////////////////////////////////

#define SOUNDBAR 68
#define MUSICBAR 102

int CP_Sound (int)
{
    enum {SOUNDS,SNDBAR,MUSIC, MSCBAR, REVERSE};
    int which;
    ControlInfo ci;
    int exit = 0, oldSV = soundvol, oldMV = musicvol;

    DrawSoundMenu ();
    MenuFadeIn ();
    WaitKeyUp ();

    do
    {
        which = HandleMenu (&SndItems, SndMenu, NULL);
        //
        // HANDLE MENU CHOICES
        //
        switch (which)
        {
            case SOUNDS:
            CA_CacheGrChunk (METERPIC);
            CA_CacheGrChunk (BARPIC);
            do
            {
                SDL_Delay(5);
                ReadAnyControl (&ci);
                switch (ci.dir)
                {
                    case dir_North:
                    case dir_South:
                        VWB_DrawPic (80,SOUNDBAR,BARPIC);
                        VWB_DrawPic ((int)(82+(soundvol*0.85)),SOUNDBAR+1,METERPIC);
                        DrawSoundMenu();
                        VW_UpdateScreen ();
                        SD_PlaySound (DSPSTOP);
                        IN_ClearKeysDown();
                        break;
                    case dir_West:
                        if (soundvol > 0)
                        {
                            soundvol--;
                            VWB_DrawPic (80,SOUNDBAR,BARPIC);
                            VWB_DrawPic ((int)(82+(soundvol*0.85)),SOUNDBAR+1,METERPIC);
                            SD_ChangeVolume((byte) (soundvol * 1.28), (byte) (musicvol * 1.28));
                            DrawSoundMenu();
                            VW_UpdateScreen ();
                            SD_PlaySound (DSSTNMOV);
                            TicDelay(2);
                        }
                        break;
                    case dir_East:
                        if (soundvol < 100)
                        {
                            soundvol++;
                            VWB_DrawPic (80,SOUNDBAR,METERPIC);
                            VWB_DrawPic ((int)(82+(soundvol*0.85)),SOUNDBAR+1,BARPIC);
                            SD_ChangeVolume((byte) (soundvol * 1.28), (byte) (musicvol * 1.28));
                            DrawSoundMenu();
                            VW_UpdateScreen ();
                            SD_PlaySound (DSSTNMOV);
                            TicDelay(2);
                        }
                        break;
                }

                if (ci.button0 || Keyboard[sc_Space] || Keyboard[sc_Enter])
                    exit = 1;
                else if (ci.button1 || Keyboard[sc_Escape])
                    exit = 2;

                }while (!exit);

                if (exit == 2)
                {
                    soundvol = oldSV;
                    SD_ChangeVolume((byte) (soundvol * 1.28), (byte) (musicvol * 1.28));
                    SD_PlaySound (DSSWTCHX);
                }
                else {
                    SD_PlaySound (DSPISTOL);
                    oldSV = soundvol;
                }

                exit=0;
                DrawSoundMenu();
                //MenuFadeIn();
                WaitKeyUp();
                break;
            case MUSIC:
                CA_CacheGrChunk (METERPIC);
                CA_CacheGrChunk (BARPIC);
                do
                {
                    SDL_Delay(5);
                    ReadAnyControl (&ci);
                    switch (ci.dir)
                    {
                        case dir_North:
                        case dir_South:
                            VWB_DrawPic (80,MUSICBAR,METERPIC);
                            VWB_DrawPic ((int)(82+(musicvol*0.85)),MUSICBAR+1,BARPIC);
                            VW_UpdateScreen ();
                            SD_PlaySound (DSPSTOP);
                            IN_ClearKeysDown();
                            break;
                        case dir_West:
                            if (musicvol > 0)
                            {
                                musicvol--;
                                VWB_DrawPic (80,MUSICBAR,METERPIC);
                                VWB_DrawPic ((int)(82+(musicvol*0.85)),MUSICBAR+1,BARPIC);
                                SD_ChangeVolume((byte) (soundvol * 1.28), (byte) (musicvol * 1.28));
                                VW_UpdateScreen ();
                                SD_PlaySound (DSSTNMOV);
                                TicDelay(2);
                            }
                            break;
                        case dir_East:
                            if (musicvol < 100)
                            {
                                musicvol++;
                                VWB_DrawPic (80,MUSICBAR,METERPIC);
                                VWB_DrawPic ((int)(82+(musicvol*0.85)),MUSICBAR+1,BARPIC);
                                SD_ChangeVolume((byte) (soundvol * 1.28), (byte) (musicvol * 1.28));
                                VW_UpdateScreen ();
                                SD_PlaySound (DSSTNMOV);
                                TicDelay(2);
                            }
                            break;
                    }

                    if (ci.button0 || Keyboard[sc_Space] || Keyboard[sc_Enter])
                        exit = 1;
                    else if (ci.button1 || Keyboard[sc_Escape])
                        exit = 2;

                } while (!exit);
                if (exit == 2)
                {
                    musicvol = oldMV;
                    SD_ChangeVolume((byte) (soundvol * 1.28), (byte) (musicvol * 1.28));
                    SD_PlaySound (DSSWTCHX);
                }
                else {
                    SD_PlaySound (DSPISTOL);
                    oldMV = musicvol;
                }
                    exit=0;
                    DrawSoundMenu();
                    //MenuFadeIn();
                    WaitKeyUp();
            break;
            case REVERSE:
                reversestereo ^= 1;
                SD_Reverse(reversestereo);
                DrawSoundMenu();
                ShootSnd();
                break;
        }

        DrawSoundMenu();
    }
    while (which >= 0);

    MenuFadeOut ();
    return 0;
}


//////////////////////
//
// DRAW THE SOUND MENU
//
void DrawSoundMenu (void)
{
    CacheLump (OPTIONS_LUMP_START,OPTIONS_LUMP_END);
    CA_CacheGrChunk (METERPIC);
    CA_CacheGrChunk (BARPIC);

    ClearMScreen();
    
    VWB_DrawPic (122,32,O_SOUNDPIC);
    VWB_DrawPic (64,52,O_SOUNDPIC);
    VWB_DrawPic (140,52,O_COLONPIC);
    VWB_DrawPic (64,86,O_MUSICPIC);
    VWB_DrawPic (64,116,O_REVERSEPIC);
    DrawMenu(&SndItems,SndMenu);
    
    VWB_DrawPic (80,MUSICBAR,METERPIC);
    VWB_DrawPic ((int)(82+(musicvol*0.85)),MUSICBAR+1,BARPIC);
    VWB_DrawPic (80,SOUNDBAR,METERPIC);
    VWB_DrawPic ((int)(82+(soundvol*0.85)),SOUNDBAR+1,BARPIC);
           
    if (reversestereo)
        VWB_DrawPic (170,119,O_SELECTEDPIC);
    else
        VWB_DrawPic (170,119,O_NOTSELECTEDPIC);

    UnCacheLump (OPTIONS_LUMP_START,OPTIONS_LUMP_END);
    VW_UpdateScreen();
}


//
// DRAW LOAD/SAVE IN PROGRESS
//
void DrawLSAction (int which)
{
#define LSA_X   96
#define LSA_Y   80
#define LSA_W   130
#define LSA_H   42
    
    CA_CacheGrChunk (C_DISKLOADING1PIC);
    CA_CacheGrChunk (C_DISKLOADING2PIC);

    DrawWindow (LSA_X, LSA_Y, LSA_W, LSA_H, BKGDCOLOR);
    VWB_DrawPic (LSA_X + 8, LSA_Y + 5, C_DISKLOADING1PIC);

    fontnumber = 1;
    SETFONTCOLOR (0, TEXTCOLOR);
    PrintX = LSA_X + 46;
    PrintY = LSA_Y + 13;

    if (!which)
        US_Print (STR_LOADING "...");
    else
        US_Print (STR_SAVING "...");

    VW_UpdateScreen ();
}


////////////////////////////////////////////////////////////////////
//
// LOAD SAVED GAMES
//
////////////////////////////////////////////////////////////////////
int CP_LoadGame (int quick)
{
    FILE *file;
    int which, exit = 0;
    char name[13];
    char loadpath[300];
    CacheLump (MESSAGE_LUMP_START,MESSAGE_LUMP_END);

    strcpy (name, SaveName);

    //
    // QUICKLOAD?
    //
    if (quick)
    {
        which = LSItems.curpos;

        if (SaveGamesAvail[which])
        {
            name[7] = which + '0';

#ifdef _arch_dreamcast
            DC_LoadFromVMU(name);
#endif

            if(configdir[0])
                snprintf(loadpath, sizeof(loadpath), "%s/%s", configdir, name);
            else
                strcpy(loadpath, name);

            file = fopen (loadpath, "rb");
            fseek (file, READ_GAMESTATE, SEEK_SET);
            loadedgame = true;
            LoadTheGame (file, 0, 0);
            strcpy(lastsavedgame,name);
            loadedgame = false;
            fclose (file);

            ContinueMusic (lastgamemusicoffset);
            return 1;
        }
    }

    DrawLoadSaveScreen (0);

    do
    {
        which = HandleMenu (&LSItems, &LSMenu[0], TrackWhichGame);
        if (which >= 0 && SaveGamesAvail[which])
        {
            ShootSnd ();
            if (which == 9) {
                strcpy(name, "autosave.");
                strcat(name, extension);
            } else {
                name[7] = which + '0';
            }

#ifdef _arch_dreamcast
            DC_LoadFromVMU(name);
#endif

            if(configdir[0])
                snprintf(loadpath, sizeof(loadpath), "%s/%s", configdir, name);
            else
                strcpy(loadpath, name);

            file = fopen (loadpath, "rb");
            fseek (file, READ_GAMESTATE, SEEK_SET);

            DrawLSAction (0);
            loadedgame = true;

            LoadTheGame (file, LSA_X + 8, LSA_Y + 5);
            strcpy(lastsavedgame,name);
            fclose (file);

            StartGame = 1;
            ShootSnd ();

            exit = 1;
            break;
        }

    }
    while (which >= 0);

    MenuFadeOut ();
    UnCacheLump (MESSAGE_LUMP_START,MESSAGE_LUMP_END);

    return exit;
}


///////////////////////////////////
//
// HIGHLIGHT CURRENT SELECTED ENTRY
//
void TrackWhichGame (int w)
{
    for (int i = 0; i < (autosaveAvailable ? 10 : 9); i++) {
        VWB_DrawPic(LSM_X + 19, (LSM_Y - 1) + (i * LSItems.curspace),I_LSENTRYPIC);
        PrintLSEntry(i);
    }

    static int lastgameon = 0;

    PrintLSEntry (lastgameon);
    PrintLSEntry (w);

    char loadpath[300];
    char bmpName[13] = BMP_SAVE;
    if (w == 9) {
        strcpy(bmpName, "autosave.svg");
    } else {
        bmpName[7] = w + '0';
    }

    if(configdir[0]) {
        snprintf(loadpath, sizeof(loadpath), "%s/%s", configdir, bmpName);
    }
    else {
        strcpy(loadpath, bmpName);
    }

    DrawWindow (LSP_X - 1, LSP_Y - 1, LSP_W + 1, LSP_H + 1, 0x00);
    SDL_Surface *bmpSurface = SDL_LoadBMP(bmpName);
    if (bmpSurface != NULL) {
        VL_LatchToScreenScaledCoord(bmpSurface, 0, 0, LSP_W, LSP_H, LSP_X*scaleFactor, LSP_Y*scaleFactor);
    }
    
    // Draw level name
    CacheLump (MESSAGE_LUMP_START,MESSAGE_LUMP_END);
    
    byte *background = grsegs[BACKDROPPIC];
    VL_MemToScreenScaledCoord(background, 320, 200, LSP_X, LSP_Y + 84, LSP_X*scaleFactor, (LSP_Y + 84)*scaleFactor, 136, 20);
    LSWrite(LSP_X, LSP_Y + 84, SaveGameLevel[w] == 0 || SaveGameLevel[w] > NUMMAPS ? "N/A" : levelNameList[SaveGameLevel[w] - 1]);
   
    LSWrite(LSP_X, LSP_Y + 94, SaveGameTime[w]);
    UnCacheLump (MESSAGE_LUMP_START,MESSAGE_LUMP_END);

    lastgameon = w;
}


////////////////////////////
//
// DRAW THE LOAD/SAVE SCREEN
//
void DrawLoadSaveScreen (int loadsave)
{
#define DISKX   100
#define DISKY   0

    int i;

    CacheLump (LOADSAVE_LUMP_START,LOADSAVE_LUMP_END);
    CacheLump (MESSAGE_LUMP_START, MESSAGE_LUMP_END);
    CA_CacheGrChunk (M_LOADGAMEPIC);
    CA_CacheGrChunk (M_SAVEGAMEPIC);
    ClearMScreen ();
    fontnumber = 1;
    DrawWindow (LSP_X - 1, LSP_Y - 1, LSP_W + 1, LSP_H + 1, 0x00);
    if (!loadsave)
        VWB_DrawPic(98,8,M_LOADGAMEPIC);
    else
        VWB_DrawPic(98,8,M_SAVEGAMEPIC);

    for (i = 0; i < (autosaveAvailable ? 10 : 9); i++) {
        VWB_DrawPic(LSM_X + 19, (LSM_Y - 1) + (i * LSItems.curspace),I_LSENTRYPIC);
        PrintLSEntry(i);
    }
    
    DrawMenu (&LSItems, &LSMenu[0]);
    UNCACHEGRCHUNK(M_LOADGAMEPIC);
    UNCACHEGRCHUNK(M_SAVEGAMEPIC);
    UnCacheLump (LOADSAVE_LUMP_START,LOADSAVE_LUMP_END);
    UnCacheLump (MESSAGE_LUMP_START, MESSAGE_LUMP_END);
    VW_UpdateScreen ();
    MenuFadeIn ();
    WaitKeyUp ();
}

static const int alpha[] = {
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, STCFN033,STCFN034,STCFN035,STCFN036,STCFN037,STCFN038,STCFN039,
    STCFN040,STCFN041,STCFN042,STCFN043,STCFN044,STCFN045,STCFN046,STCFN047,
    STCFN048,STCFN049,STCFN050,STCFN051,STCFN052,STCFN053,STCFN054,STCFN055,
    STCFN056,STCFN057,STCFN058,STCFN059,STCFN060,STCFN061,STCFN062,STCFN063,
    STCFN064,STCFN065,STCFN066,STCFN067,STCFN068,STCFN069,STCFN070,STCFN071,
    STCFN072,STCFN073,STCFN074,STCFN075,STCFN076,STCFN077,STCFN078,STCFN079,
    STCFN080,STCFN081,STCFN082,STCFN083,STCFN084,STCFN085,STCFN086,STCFN087,
    STCFN088,STCFN089,STCFN090,STCFN091,STCFN092,STCFN093,STCFN094,STCFN095,
    STCFN039,STCFN065,STCFN066,STCFN067,STCFN068,STCFN069,STCFN070,STCFN071,
    STCFN072,STCFN073,STCFN074,STCFN075,STCFN076,STCFN077,STCFN078,STCFN079,
    STCFN080,STCFN081,STCFN082,STCFN083,STCFN084,STCFN085,STCFN086,STCFN087,
    STCFN088,STCFN089,STCFN090, STCFN091, STCFN121, STCFN093, STCFN045, 0
};

static const int alphaLength[] = {
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    4, 4, 7, 7, 7, 9, 8, 4,
    7, 7, 7, 5, 4, 6, 4, 7,
    8, 5, 8, 8, 7, 7, 8, 8,
    8, 8, 4, 4, 5, 5, 5, 8,
    9, 8, 8, 8, 8, 8, 8, 8,
    8, 4, 8, 8, 8, 9, 8, 8,
    8, 8, 8, 7, 8, 8, 7, 9,
    9, 8, 7, 5, 7, 5, 7, 8,
    8, 8, 8, 8, 8, 8, 8, 8,
    8, 4, 8, 8, 8, 9, 8, 8,
    8, 8, 8, 7, 8, 8, 7, 9,
    9, 8, 7, 5, 4, 5, 6, 0
};

int LSWrite_GetLength(const char *string)
{
    int x = 0, len = (int) strlen(string);
    char ch;

    for (int i = 0; i < len; i++)
    {
        ch = string[i];
        x += alphaLength[ch];
    }

    return x;
}

void LSWrite (int x, int y, const char *string)
{
    int ox, oy, nx, ny, len = (int) strlen(string);
    char ch;

    ox = nx = x; // start of line (ox), next char (nx)
    oy = ny = y;
    for (int i = 0; i < len; i++)
    {
        ch = string[i];
        ny = oy;
        // switch is for heights
        switch (ch) {
        case ',':
        case '.':
        case '_':
            ny+=4;
            break;
        case '-':
        case '+':
            ny+=2;
            break;
        case '\n':
            nx = ox;
            oy = ny += 8;
            continue;
        }

        if (alpha[ch] > 0)
        {
            VWB_DrawPic (nx,ny,alpha[ch]);
        }

        nx+=alphaLength[ch];
    }
}

///////////////////////////////////////////
//
// PRINT LOAD/SAVE GAME ENTRY W/BOX OUTLINE
//
void PrintLSEntry (int w)
{
    PrintX = LSM_X + LSItems.indent;
    PrintY = LSM_Y + w * LSItems.curspace + 3;
    
    if (w == 9) {
        LSWrite (PrintX,PrintY,"Autosave");
    }
    else if (SaveGamesAvail[w]) {
        LSWrite (PrintX,PrintY,SaveGameNames[w]);
    }
    else {
        LSWrite (PrintX, PrintY, STR_EMPTY);
    }
}


////////////////////////////////////////////////////////////////////
//
// SAVE CURRENT GAME
//
////////////////////////////////////////////////////////////////////
int loadsaveWhich = 0;
void DrawLoadSaveBar(int which)
{
    CA_CacheGrChunk(I_LSENTRYPIC);
    VWB_DrawPic (LSM_X+LSItems.indent - 5,LSM_Y+loadsaveWhich*LSItems.curspace - 1,I_LSENTRYPIC);
}
int CP_SaveGame (int quick)
{
    int which, exit = 0;
    FILE *file;
    char name[13];
    char savepath[300];
    char input[32];
    char picpath[300];
    char bmpName[13] = BMP_SAVE; 
    CA_CacheGrChunk (I_LSENTRYPIC);
    CacheLump (MESSAGE_LUMP_START,MESSAGE_LUMP_END);

    strcpy (name, SaveName);

    //
    // QUICKSAVE?
    //
    if (quick)
    {
        which = LSItems.curpos;

        if (SaveGamesAvail[which])
        {
            name[7] = which + '0';

            if(configdir[0])
                snprintf(savepath, sizeof(savepath), "%s/%s", configdir, name);
            else
                strcpy(savepath, name);

            unlink (savepath);
            file = fopen (savepath, "wb");

            strcpy (input, &SaveGameNames[which][0]);

            fwrite (input, 1, 32, file);
            LSWrite (PrintX,PrintY,input);

            time_t timeclock;
            struct tm * timeinfo;
            time(&timeclock);
            timeinfo = localtime(&timeclock);
            char timeStr[20];
            strftime(timeStr, 20, "%x %I:%M %p ", timeinfo);
            fwrite(timeStr, 1, 20, file);
            strcpy(SaveGameTime[which], timeStr);

            fseek (file, READ_GAMESTATE, SEEK_SET);
            SaveTheGame (file, 0, 0);
            strcpy(lastsavedgame,name);
            bmpName[7] = which + '0';
            if(configdir[0]) {
                snprintf(picpath, sizeof(picpath), "%s/%s", configdir, bmpName);
            }
            else {
                strcpy(picpath, bmpName);
            }

            unlink (picpath);
            SDL_SaveBMP(lastGameSurface, picpath); 
            fclose (file);

#ifdef _arch_dreamcast
            DC_SaveToVMU(name, input);
#endif

            return 1;
        }
    }

    DrawLoadSaveScreen (1);

    do
    {
        which = HandleMenu (&LSItems, &LSMenu[0], TrackWhichGame);
        if (which >= 0)
        {
            //
            // OVERWRITE EXISTING SAVEGAME?
            //
            if (SaveGamesAvail[which])
            {
#ifdef JAPAN
                if (!GetYorN (7, 8, C_JAPSAVEOVERPIC))
#else
                if (!Confirm (GAMESVD))
#endif
                {
                    DrawLoadSaveScreen (1);
                    continue;
                }
                else
                {
                    DrawLoadSaveScreen (1);
                    PrintLSEntry (which);
                    VW_UpdateScreen ();
                }
            }

            ShootSnd ();

            strcpy (input, &SaveGameNames[which][0]);
            name[7] = which + '0';

            fontnumber = 0;
            if (!SaveGamesAvail[which])
                VWB_DrawPic (LSM_X+LSItems.indent - 5,LSM_Y+loadsaveWhich*LSItems.curspace - 1,I_LSENTRYPIC);
            VW_UpdateScreen ();
            loadsaveWhich = which;

            if (US_LineInput
                (LSM_X + LSItems.indent, LSM_Y + which * LSItems.curspace + 3, input, input, true, 31,
                 LSM_W - LSItems.indent - 30, DrawLoadSaveBar))
            {
                SaveGamesAvail[which] = 1;
                strcpy (&SaveGameNames[which][0], input);
                SaveGameLevel[which] = gamestate.episode * 10 + gamestate.mapon + 1;

                if(configdir[0])
                    snprintf(savepath, sizeof(savepath), "%s/%s", configdir, name);
                else
                    strcpy(savepath, name);

                unlink (savepath);
                file = fopen (savepath, "wb");
                fwrite (input, 32, 1, file);

                time_t timeclock;
                struct tm * timeinfo;
                time(&timeclock);
                timeinfo = localtime(&timeclock);
                char timeStr[20];
                strftime(timeStr, 20, "%x %I:%M %p ", timeinfo);
                fwrite(timeStr, 1, 20, file);
                strcpy(SaveGameTime[which], timeStr);
                fseek (file, READ_GAMESTATE, SEEK_SET);

                DrawLSAction (1);
                SaveTheGame (file, LSA_X + 8, LSA_Y + 5);
                strcpy(lastsavedgame,name);
                bmpName[7] = which + '0';
                if(configdir[0]) {
                    snprintf(picpath, sizeof(picpath), "%s/%s", configdir, bmpName);
                }
                else {
                    strcpy(picpath, bmpName);
                }

                unlink (picpath);
                SDL_SaveBMP(lastGameSurface, picpath); 

                fclose (file);

#ifdef _arch_dreamcast
                DC_SaveToVMU(name, input);
#endif

                ShootSnd ();
                exit = 1;
            }
            else
            {
                DrawLoadSaveScreen(1);
                //VWB_DrawPic (LSM_X+LSItems.indent - 5,LSM_Y+loadsaveWhich*LSItems.curspace - 1,I_LSENTRYPIC);
                //PrintLSEntry (which);
                VW_UpdateScreen ();
                SD_PlaySound (DSSWTCHX);
                continue;
            }

            fontnumber = 1;
            break;
        }

    }
    while (which >= 0);

    MenuFadeOut ();
    
    UNCACHEGRCHUNK (I_LSENTRYPIC);
    UnCacheLump (MESSAGE_LUMP_START, MESSAGE_LUMP_END);

    return exit;
}
#define AUTOSAVESLOT 9
int AutoSave ()
{
    FILE *file;
    char name[13] = "autosave.";
    char savepath[300];
    char input[32] = "Autosave";
    char picpath[300];
    char bmpName[13] = "autosave.svg";

    //if (SaveGamesAvail[AUTOSAVESLOT])
    //{
        strcat(name, extension);

        if(configdir[0])
            snprintf(savepath, sizeof(savepath), "%s/%s", configdir, name);
        else
            strcpy(savepath, name);

        unlink (savepath);
        file = fopen (savepath, "wb");
        fwrite (input, 1, 32, file);

        time_t timeclock;
        struct tm * timeinfo;
        time(&timeclock);
        timeinfo = localtime(&timeclock);
        char timeStr[20];
        strftime(timeStr, 20, "%x %I:%M %p ", timeinfo);
        fwrite(timeStr, 1, 20, file);
        strcpy(SaveGameTime[AUTOSAVESLOT], timeStr);
        SaveGameLevel[AUTOSAVESLOT] = gamestate.episode * 10 + gamestate.mapon + 1;

        fseek (file, READ_GAMESTATE, SEEK_SET);
        SaveTheGame (file, 0, 0);
        strcpy(lastsavedgame,name);
        if(configdir[0]) {
            snprintf(picpath, sizeof(picpath), "%s/%s", configdir, bmpName);
        }
        else {
            strcpy(picpath, bmpName);
        }

        unlink (picpath);
        if (lastGameSurface != NULL) {
            SDL_SaveBMP(lastGameSurface, picpath); 
        }
        fclose (file);

#ifdef _arch_dreamcast
        DC_SaveToVMU(name, input);
#endif
        SaveGamesAvail[AUTOSAVESLOT] = 1;
        LSMenu[AUTOSAVESLOT].active = 1; // activate the autosave slot
        autosaveAvailable = true;
        return 1;
    //}

    //return 0;
}
////////////////////////////////////////////////////////////////////
//
// DEFINE CONTROLS
//
////////////////////////////////////////////////////////////////////
int
CP_Control (int)
{
    int which;

    DrawCtlScreen ();
    MenuFadeIn ();
    WaitKeyUp ();

    do
    {
        which = HandleMenu (&CtlItems, CtlMenu, NULL);
        switch (which)
        {
            case CTL_MOUSEENABLE:
                mouseenabled ^= 1;
                if(IN_IsInputGrabbed())
                    IN_CenterMouse();
                DrawCtlScreen ();
                CusItems.curpos = -1;
                ShootSnd ();
                WaitKeyUp ();
                break;

            case CTL_JOYENABLE:
                joystickenabled ^= 1;
                DrawCtlScreen ();
                CusItems.curpos = -1;
                ShootSnd ();
                WaitKeyUp ();
                break;

            case CTL_MOUSESENS:
            case CTL_CUSTOMIZE:
                DrawCtlScreen ();
                MenuFadeIn ();
                WaitKeyUp ();
                break;
        }
    }
    while (which >= 0);

    MenuFadeOut ();

    return 0;
}


////////////////////////////////
//
// DRAW MOUSE SENSITIVITY SCREEN
//
void DrawMouseSens (void)
{
    ClearMScreen ();
    CA_CacheGrChunk (O_MOUSESENSPIC);
    CA_CacheGrChunk (BARPIC);
    CA_CacheGrChunk (METERPIC);
    
    VWB_DrawPic (58,72,O_MOUSESENSPIC);
    VWB_DrawPic (114,96,METERPIC);
    VWB_DrawPic (120+8*mouseadjustment,97,BARPIC);

    VW_UpdateScreen ();
    UNCACHEGRCHUNK (O_MOUSESENSPIC);
    MenuFadeIn ();
}


///////////////////////////
//
// ADJUST MOUSE SENSITIVITY
//
int MouseSensitivity (int)
{
    ControlInfo ci;
    int exit = 0, oldMA;


    oldMA = mouseadjustment;
    DrawMouseSens ();
    do
    {
        SDL_Delay(5);
        ReadAnyControl (&ci);
        switch (ci.dir)
        {
            case dir_North:
            case dir_West:
                if (mouseadjustment)
                {
                    mouseadjustment--;
                    VWB_DrawPic (114,96,METERPIC);
                    VWB_DrawPic (120+8*mouseadjustment,97,BARPIC);
                    SD_PlaySound(DSSTNMOV);
                    VW_UpdateScreen ();
                    TicDelay(20);
                }
                break;

            case dir_South:
            case dir_East:
                if (mouseadjustment < 9)
                {
                    mouseadjustment++;
                    VWB_DrawPic (114,96,METERPIC);
                    VWB_DrawPic (120+8*mouseadjustment,97,BARPIC);
                    VW_UpdateScreen();
                    SD_PlaySound(DSSTNMOV);
                    VW_UpdateScreen ();
                    TicDelay(20);
                }
                break;
        }

        if (ci.button0 || Keyboard[sc_Space] || Keyboard[sc_Enter])
            exit = 1;
        else if (ci.button1 || Keyboard[sc_Escape])
            exit = 2;

    }
    while (!exit);

    if (exit == 2)
    {
        mouseadjustment = oldMA;
        SD_PlaySound(DSSWTCHX);
    }
    else
        SD_PlaySound(DSPISTOL);

    WaitKeyUp ();
    MenuFadeOut ();

    return 0;
}


///////////////////////////
//
// DRAW CONTROL MENU SCREEN
//
void DrawCtlScreen (void)
{
    int i;
    
    CacheLump (OPTIONS_LUMP_START,OPTIONS_LUMP_END);
    ClearMScreen ();
    VWB_DrawPic(110,16,O_CONTROLPIC);
    VWB_DrawPic(68,42,O_MOUSEENABLEPIC);
    VWB_DrawPic(68,58,O_MOUSESENSPIC);
    VWB_DrawPic(68,74,O_JOYSTICKENABLEPIC);
    VWB_DrawPic(68,90,O_CUSTOMIZEPIC);
    WindowX = 0;
    WindowW = 320;
    SETFONTCOLOR (TEXTCOLOR, BKGDCOLOR);

    if (IN_JoyPresent())
        CtlMenu[CTL_JOYENABLE].active = 1;

    if (MousePresent)
    {
        CtlMenu[CTL_MOUSESENS].active = CtlMenu[CTL_MOUSEENABLE].active = 1;
    }

    CtlMenu[CTL_MOUSESENS].active = mouseenabled;


    DrawMenu (&CtlItems, CtlMenu);

    if (mouseenabled) {
        VWB_DrawPic(151,45,O_SELECTEDPIC);
    }
    else {
        VWB_DrawPic(151,45,O_NOTSELECTEDPIC);
    }

    if (joystickenabled) {
        VWB_DrawPic(183,77,O_SELECTEDPIC);
    }
    else {
        VWB_DrawPic(183,77,O_NOTSELECTEDPIC);
    }

    //
    // PICK FIRST AVAILABLE SPOT
    //
    if (CtlItems.curpos < 0 || !CtlMenu[CtlItems.curpos].active)
    {
        for (i = 0; i < CtlItems.amount; i++)
        {
            if (CtlMenu[i].active)
            {
                CtlItems.curpos = i;
                break;
            }
        }
    }

    //DrawMenuGun (&CtlItems);
    UnCacheLump(OPTIONS_LUMP_START,OPTIONS_LUMP_END);
    VW_UpdateScreen ();
}


////////////////////////////////////////////////////////////////////
//
// CUSTOMIZE CONTROLS
//
////////////////////////////////////////////////////////////////////
enum
{ FIRE, STRAFE, RUN, OPEN };
char mbarray[4][3] = { "b0", "b1", "b2", "b3" };
int8_t order[4] = { RUN, OPEN, FIRE, STRAFE };


int
CustomControls (int)
{
    int which;
    fontnumber = 0;
    DrawCustomScreen ();
    do
    {
        which = HandleMenu (&CusItems, &CusMenu[0], FixupCustom);
        switch (which)
        {
            case 0:
                DefineMouseBtns ();
                DrawCustMouse (1);
                break;
            case 3:
                DefineJoyBtns ();
                DrawCustJoy (0);
                break;
            case 6:
                DefineKeyBtns ();
                DrawCustKeybd (0);
                break;
            case 8:
                DefineKeyMove ();
                DrawCustKeys (0);
        }
    }
    while (which >= 0);

    MenuFadeOut ();

    return 0;
}


////////////////////////
//
// DEFINE THE MOUSE BUTTONS
//
void
DefineMouseBtns (void)
{
    CustomCtrls mouseallowed = { 0, 1, 1, 1 };
    EnterCtrlData (2, &mouseallowed, DrawCustMouse, PrintCustMouse, MOUSE);
}


////////////////////////
//
// DEFINE THE JOYSTICK BUTTONS
//
void
DefineJoyBtns (void)
{
    CustomCtrls joyallowed = { 1, 1, 1, 1 };
    EnterCtrlData (5, &joyallowed, DrawCustJoy, PrintCustJoy, JOYSTICK);
}


////////////////////////
//
// DEFINE THE KEYBOARD BUTTONS
//
void
DefineKeyBtns (void)
{
    CustomCtrls keyallowed = { 1, 1, 1, 1 };
    EnterCtrlData (8, &keyallowed, DrawCustKeybd, PrintCustKeybd, KEYBOARDBTNS);
}


////////////////////////
//
// DEFINE THE KEYBOARD BUTTONS
//
void
DefineKeyMove (void)
{
    CustomCtrls keyallowed = { 1, 1, 1, 1 };
    EnterCtrlData (10, &keyallowed, DrawCustKeys, PrintCustKeys, KEYBOARDMOVE);
}


////////////////////////
//
// ENTER CONTROL DATA FOR ANY TYPE OF CONTROL
//
enum
{ FWRD, RIGHT, BKWD, LEFT };
int moveorder[4] = { LEFT, RIGHT, FWRD, BKWD };

void
EnterCtrlData (int index, CustomCtrls * cust, void (*DrawRtn) (int), void (*PrintRtn) (int),
               int type)
{
    int j, exit, tick, redraw, which, x, picked, lastFlashTime;
    ControlInfo ci;
    byte *source = grsegs[BACKDROPPIC];

    ShootSnd ();
    PrintY = CST_Y + 13 * index;
    IN_ClearKeysDown ();
    exit = 0;
    redraw = 1;
    //
    // FIND FIRST SPOT IN ALLOWED ARRAY
    //
    for (j = 0; j < 4; j++)
        if (cust->allowed[j])
        {
            which = j;
            break;
        }

    do
    {
        if (redraw)
        {
            x = CST_START + CST_SPC * which;
            VL_MemToScreenScaledCoord(source, 320,200, 45, PrintY - 1, 45*scaleFactor, (PrintY - 1)*scaleFactor,270,13);

            DrawRtn (1);
            DrawWindow (x - 2, PrintY - 1, CST_SPC, 11, BKGDCOLOR);
            //DrawOutline (x - 2, PrintY, CST_SPC, 11, 0, HIGHLIGHT);
            SETFONTCOLOR (0, TEXTCOLOR);
            PrintRtn (which);
            PrintX = x;
            SETFONTCOLOR (TEXTCOLOR, BKGDCOLOR);
            VW_UpdateScreen ();
            WaitKeyUp ();
            redraw = 0;
        }

        SDL_Delay(5);
        ReadAnyControl (&ci);

        if (type == MOUSE || type == JOYSTICK)
            if (IN_KeyDown (sc_Enter) || IN_KeyDown (sc_Control) || IN_KeyDown (sc_Alt))
            {
                IN_ClearKeysDown ();
                ci.button0 = ci.button1 = false;
            }

        //
        // CHANGE BUTTON VALUE?
        //
        if ((type != KEYBOARDBTNS && type != KEYBOARDMOVE) && (ci.button0 | ci.button1 | ci.button2 | ci.button3) ||
            ((type == KEYBOARDBTNS || type == KEYBOARDMOVE) && LastScan == sc_Enter))
        {
            lastFlashTime = GetTimeCount();
            tick = picked = 0;
            SETFONTCOLOR (0, TEXTCOLOR);

            if (type == KEYBOARDBTNS || type == KEYBOARDMOVE)
                IN_ClearKeysDown ();

            while(1)
            {
                int button, result = 0;

                //
                // FLASH CURSOR
                //
                if (GetTimeCount() - lastFlashTime > 20)
                {
                    switch (tick)
                    {
                        case 0:
                            VWB_Bar (x, PrintY + 1, CST_SPC - 2, 9, BKGDCOLOR);
                            break;
                        case 1:
                            PrintX = x;
                            US_Print ("?");
                            SD_PlaySound (DSSTNMOV);
                    }
                    tick ^= 1;
                    lastFlashTime = GetTimeCount();
                    VW_UpdateScreen ();
                }
                else SDL_Delay(5);

                //
                // WHICH TYPE OF INPUT DO WE PROCESS?
                //
                switch (type)
                {
                    case MOUSE:
                        button = IN_MouseButtons();
                        switch (button)
                        {
                            case 1:
                                result = 1;
                                break;
                            case 2:
                                result = 2;
                                break;
                            case 4:
                                result = 3;
                                break;
                        }

                        if (result)
                        {
                            for (int z = 0; z < 4; z++)
                                if (order[which] == buttonmouse[z])
                                {
                                    buttonmouse[z] = bt_nobutton;
                                    break;
                                }

                            buttonmouse[result - 1] = order[which];
                            picked = 1;
                            SD_PlaySound (DSPISTOL);
                        }
                        break;

                    case JOYSTICK:
                        if (ci.button0)
                            result = 1;
                        else if (ci.button1)
                            result = 2;
                        else if (ci.button2)
                            result = 3;
                        else if (ci.button3)
                            result = 4;

                        if (result)
                        {
                            for (int z = 0; z < 4; z++)
                            {
                                if (order[which] == buttonjoy[z])
                                {
                                    buttonjoy[z] = bt_nobutton;
                                    break;
                                }
                            }

                            buttonjoy[result - 1] = order[which];
                            picked = 1;
                            SD_PlaySound (DSPISTOL);
                        }
                        break;

                    case KEYBOARDBTNS:
                        if (LastScan && LastScan != sc_Escape)
                        {
                            buttonscan[order[which]] = LastScan;
                            picked = 1;
                            ShootSnd ();
                            IN_ClearKeysDown ();
                        }
                        break;

                    case KEYBOARDMOVE:
                        if (LastScan && LastScan != sc_Escape)
                        {
                            dirscan[moveorder[which]] = LastScan;
                            picked = 1;
                            ShootSnd ();
                            IN_ClearKeysDown ();
                        }
                        break;
                }

                //
                // EXIT INPUT?
                //
                if (IN_KeyDown (sc_Escape) || type != JOYSTICK && ci.button1)
                {
                    picked = 1;
                    SD_PlaySound (DSSWTCHX);
                }

                if(picked) break;

                ReadAnyControl (&ci);
            }

            SETFONTCOLOR (TEXTCOLOR, BKGDCOLOR);
            redraw = 1;
            WaitKeyUp ();
            continue;
        }

        if (ci.button1 || IN_KeyDown (sc_Escape))
            exit = 1;

        //
        // MOVE TO ANOTHER SPOT?
        //
        switch (ci.dir)
        {
            case dir_West:
                do
                {
                    which--;
                    if (which < 0)
                        which = 3;
                }
                while (!cust->allowed[which]);
                redraw = 1;
                SD_PlaySound (DSPSTOP);
                while (ReadAnyControl (&ci), ci.dir != dir_None) SDL_Delay(5);
                IN_ClearKeysDown ();
                break;

            case dir_East:
                do
                {
                    which++;
                    if (which > 3)
                        which = 0;
                }
                while (!cust->allowed[which]);
                redraw = 1;
                SD_PlaySound (DSPSTOP);
                while (ReadAnyControl (&ci), ci.dir != dir_None) SDL_Delay(5);
                IN_ClearKeysDown ();
                break;
            case dir_North:
            case dir_South:
                exit = 1;
        }
    }
    while (!exit);

    SD_PlaySound (DSSWTCHX);
    WaitKeyUp ();
    VL_MemToScreenScaledCoord(source, 320,200, 45, PrintY - 1, 45*scaleFactor, (PrintY - 1)*scaleFactor,270,13);

}


////////////////////////
//
// FIXUP GUN CURSOR OVERDRAW SHIT
//
void
FixupCustom (int w)
{
    static int lastwhich = -1;

    switch (w)
    {
        case 0:
            DrawCustMouse (1);
            break;
        case 3:
            DrawCustJoy (1);
            break;
        case 6:
            DrawCustKeybd (1);
            break;
        case 8:
            DrawCustKeys (1);
    }

    if (lastwhich >= 0 && lastwhich != w)
    {
        switch (lastwhich)
        {
            case 0:
                DrawCustMouse (0);
                break;
            case 3:
                DrawCustJoy (0);
                break;
            case 6:
                DrawCustKeybd (0);
                break;
            case 8:
                DrawCustKeys (0);
        }
    }

    lastwhich = w;
}


////////////////////////
//
// DRAW CUSTOMIZE SCREEN
//
void DrawCustomScreen (void)
{
    int x, y;
    CacheLump(CUSTOMIZE_LUMP_START, CUSTOMIZE_LUMP_END);
    CacheLump(MESSAGE_LUMP_START, MESSAGE_LUMP_END);
    ClearMScreen ();
    WindowX = 0;
    WindowW = 320;

    //
    // MOUSE
    //

    VWB_DrawPic(122, CST_Y, Q_MOUSEPIC);
    
    x = CST_START; y = CST_Y + 16;
    LSWrite(x, y, STR_CRUN);
    x += CST_SPC;
    LSWrite(x, y, STR_COPEN);
    x += CST_SPC;
    LSWrite(x, y, STR_CFIRE);
    x += CST_SPC;
    LSWrite(x, y, STR_CSTRAFE);

    DrawCustMouse (0);

    //
    // JOYSTICK/PAD
    //
    
    x = CST_START; y = CST_Y + 55;
    VWB_DrawPic(64, y - 16, Q_JOYGRAVISPIC);
    
    LSWrite(x, y, STR_CRUN);
    x += CST_SPC;
    LSWrite(x, y, STR_COPEN);
    x += CST_SPC;
    LSWrite(x, y, STR_CFIRE);
    x += CST_SPC;
    LSWrite(x, y, STR_CSTRAFE);
    DrawCustJoy (0);

    //
    // KEYBOARD
    //
    x = CST_START; y = CST_Y + 94;
    VWB_DrawPic(102, y - 16, Q_KEYBOARDPIC);

    LSWrite(x, y, STR_CRUN);
    x += CST_SPC;
    LSWrite(x, y, STR_COPEN);
    x += CST_SPC;
    LSWrite(x, y, STR_CFIRE);
    x += CST_SPC;
    LSWrite(x, y, STR_CSTRAFE);
    DrawCustKeybd (0);


    //
    // KEYBOARD MOVE KEYS
    //
    x = CST_START; y = CST_Y + 120;
    LSWrite(x, y, STR_LEFT);
    x += CST_SPC;
    LSWrite(x, y, STR_RIGHT);
    x += CST_SPC;
    LSWrite(x, y, STR_FRWD);
    x += CST_SPC;
    LSWrite(x, y, STR_BKWD);
    DrawCustKeys (0);

    //
    // PICK STARTING POINT IN MENU
    //
    if (CusItems.curpos < 0)
        for (int i = 0; i < CusItems.amount; i++) {
            if (CusMenu[i].active)
            {
                CusItems.curpos = i;
                break;
            }
        }

    VW_UpdateScreen ();
    MenuFadeIn ();

    UnCacheLump(CUSTOMIZE_LUMP_START, CUSTOMIZE_LUMP_END);
    UnCacheLump(MESSAGE_LUMP_START, MESSAGE_LUMP_END);
}


void
PrintCustMouse (int i)
{
    int j;

    for (j = 0; j < 4; j++)
        if (order[i] == buttonmouse[j])
        {
            PrintX = CST_START + CST_SPC * i;
            US_Print (mbarray[j]);
            break;
        }
}

void
DrawCustMouse (int hilight)
{
    int i, color;

    color = TEXTCOLOR;
    if (hilight)
        color = HIGHLIGHT;
    SETFONTCOLOR (color, BKGDCOLOR);

    if (!mouseenabled)
    {
        SETFONTCOLOR (DEACTIVE, BKGDCOLOR);
        CusMenu[0].active = 0;
    }
    else
        CusMenu[0].active = 1;

    PrintY = CST_Y + 13 * 2;
    for (i = 0; i < 4; i++)
        PrintCustMouse (i);
}

void
PrintCustJoy (int i)
{
    for (int j = 0; j < 4; j++)
    {
        if (order[i] == buttonjoy[j])
        {
            PrintX = CST_START + CST_SPC * i;
            US_Print (mbarray[j]);
            break;
        }
    }
}

void
DrawCustJoy (int hilight)
{
    int i, color;

    color = TEXTCOLOR;
    if (hilight)
        color = HIGHLIGHT;
    SETFONTCOLOR (color, BKGDCOLOR);

    if (!joystickenabled)
    {
        SETFONTCOLOR (DEACTIVE, BKGDCOLOR);
        CusMenu[3].active = 0;
    }
    else
        CusMenu[3].active = 1;

    PrintY = CST_Y + 13 * 5;
    for (i = 0; i < 4; i++)
        PrintCustJoy (i);
}


void
PrintCustKeybd (int i)
{
    PrintX = CST_START + CST_SPC * i;
    US_Print ((const char *) IN_GetScanName (buttonscan[order[i]]));
}

void
DrawCustKeybd (int hilight)
{
    int i, color;


    color = TEXTCOLOR;
    if (hilight)
        color = HIGHLIGHT;
    SETFONTCOLOR (color, BKGDCOLOR);

    PrintY = CST_Y + 13 * 8;
    for (i = 0; i < 4; i++)
        PrintCustKeybd (i);
}

void
PrintCustKeys (int i)
{
    PrintX = CST_START + CST_SPC * i;
    US_Print ((const char *) IN_GetScanName (dirscan[moveorder[i]]));
}

void
DrawCustKeys (int hilight)
{
    int i, color;


    color = TEXTCOLOR;
    if (hilight)
        color = HIGHLIGHT;
    SETFONTCOLOR (color, BKGDCOLOR);

    PrintY = CST_Y + 13 * 10;
    for (i = 0; i < 4; i++)
        PrintCustKeys (i);
}


////////////////////////////////////////////////////////////////////
//
// CHANGE SCREEN VIEWING SIZE
//
////////////////////////////////////////////////////////////////////
int CP_ChangeView (int)
{
    int exit = 0, oldview, newview;
    ControlInfo ci;
    WindowX = WindowY = 0;
    WindowW = 320;
    WindowH = 200;
    newview = oldview = viewsize;
    DrawChangeView (oldview);
    MenuFadeIn ();

    do
    {
        CheckPause ();
        SDL_Delay(5);
        ReadAnyControl (&ci);
        switch (ci.dir)
        {
            case dir_South:
            case dir_West:
                newview--;
                if (newview < 20)
                    newview = 20;
                else
                    SD_PlaySound (DSPSTOP);
                DrawChangeView(newview);
                VW_UpdateScreen ();
                TicDelay (20);
                break;

            case dir_North:
            case dir_East:
                newview++;
                // only play sound moving the screen size
                if (newview <= 22)
                    SD_PlaySound (DSPSTOP);

                if (newview >= 22)
                    newview = 22;
                else
                    ShowViewSize (newview);
                DrawChangeView(newview);
                VW_UpdateScreen ();
                TicDelay (20);
                break;
        }

        if (ci.button0 || Keyboard[sc_Enter])
            exit = 1;
        else if (ci.button1 || Keyboard[sc_Escape])
        {
            SD_PlaySound (DSSWTCHX);
            MenuFadeOut ();
            if(screenHeight % 200 != 0)
                VL_ClearScreen(0);
            return 0;
        }
    }
    while (!exit);

    if (oldview != newview)
    {
        ShootSnd();
        Message (STR_THINK "...");
        NewViewSize (newview);
    }

    ShootSnd ();
    MenuFadeOut ();

    return 0;
}


/////////////////////////////
//
// DRAW THE CHANGEVIEW SCREEN
//
void DrawChangeView (int view)
{
    //int rescaledHeight = screenHeight / scaleFactor;
    //if(view < 21) ClearMScreen();
    VWB_Bar(0, 0, 320, 200, 0x03);

    ShowViewSize (view);

    fontnumber = 0;
    PrintY = (screenHeight / scaleFactor) - 31;
    WindowX = 0;
    WindowY = 320;                                  // TODO: Check this!
    SETFONTCOLOR (HIGHLIGHT, BKGDCOLOR);

    US_CPrint (STR_SIZE1 "\n");
    US_CPrint (STR_SIZE2 "\n");
    US_CPrint (STR_SIZE3);
    VW_UpdateScreen ();
    fontnumber = 1;
}


////////////////////////////////////////////////////////////////////
//
// QUIT THIS INFERNAL GAME!
//
////////////////////////////////////////////////////////////////////
int
CP_Quit (int)
{
    int rand = US_RndT();
    int len = lengthof(endStrings);
    if (Confirm (endStrings[rand % len]))
    {
        VW_UpdateScreen ();
        SD_MusicOff ();
        SD_StopSound ();
        MenuFadeOut ();
        Quit (NULL);
    }

    DrawMainMenu ();
    return 0;
}


////////////////////////////////////////////////////////////////////
//
// HANDLE INTRO SCREEN (SYSTEM CONFIG)
//
////////////////////////////////////////////////////////////////////
void
IntroScreen (void)
{
#ifdef SPEAR

#define MAINCOLOR       0x4f
#define EMSCOLOR        0x4f
#define XMSCOLOR        0x4f

#else

#define MAINCOLOR       0x79
#define EMSCOLOR        0xb9    // 0x4f
#define XMSCOLOR        0xc9    // 0x7f

#endif
#define FILLCOLOR       0xe7

//      long memory;
//      long emshere,xmshere;
    int i;
/*      int ems[10]={100,200,300,400,500,600,700,800,900,1000},
                xms[10]={100,200,300,400,500,600,700,800,900,1000};
        int main[10]={32,64,96,128,160,192,224,256,288,320};*/


    //
    // DRAW MAIN MEMORY
    //
#ifdef ABCAUS
    memory = (1023l + mminfo.nearheap + mminfo.farheap) / 1024l;
    for (i = 0; i < 10; i++)
        if (memory >= main[i])
            VWB_Bar (49, 163 - 8 * i, 6, 5, MAINCOLOR - i);

    //
    // DRAW EMS MEMORY
    //
    if (EMSPresent)
    {
        emshere = 4l * EMSPagesAvail;
        for (i = 0; i < 10; i++)
            if (emshere >= ems[i])
                VWB_Bar (89, 163 - 8 * i, 6, 5, EMSCOLOR - i);
    }

    //
    // DRAW XMS MEMORY
    //
    if (XMSPresent)
    {
        xmshere = 4l * XMSPagesAvail;
        for (i = 0; i < 10; i++)
            if (xmshere >= xms[i])
                VWB_Bar (129, 163 - 8 * i, 6, 5, XMSCOLOR - i);
    }
#else
    for (i = 0; i < 10; i++)
        VWB_Bar (49, 163 - 8 * i, 6, 5, MAINCOLOR - i);
    for (i = 0; i < 10; i++)
        VWB_Bar (89, 163 - 8 * i, 6, 5, EMSCOLOR - i);
    for (i = 0; i < 10; i++)
        VWB_Bar (129, 163 - 8 * i, 6, 5, XMSCOLOR - i);
#endif


    //
    // FILL BOXES
    //
    if (MousePresent)
        VWB_Bar (164, 82, 12, 2, FILLCOLOR);

    if (IN_JoyPresent())
        VWB_Bar (164, 105, 12, 2, FILLCOLOR);
}


////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////
//
// SUPPORT ROUTINES
//
////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////
//
// Clear Menu screens to dark red
//
////////////////////////////////////////////////////////////////////
void
ClearMScreen (void)
{
    CA_CacheGrChunk (BACKDROPPIC);
    VWB_DrawPic (0,0,BACKDROPPIC);
    UNCACHEGRCHUNK (BACKDROPPIC);
}


////////////////////////////////////////////////////////////////////
//
// Un/Cache a LUMP of graphics
//
////////////////////////////////////////////////////////////////////
void
CacheLump (int lumpstart, int lumpend)
{
    int i;

    for (i = lumpstart; i <= lumpend; i++)
        CA_CacheGrChunk (i);
}


void
UnCacheLump (int lumpstart, int lumpend)
{
    int i;

    for (i = lumpstart; i <= lumpend; i++)
        if (grsegs[i])
            UNCACHEGRCHUNK (i);
}


////////////////////////////////////////////////////////////////////
//
// Draw a window for a menu
//
////////////////////////////////////////////////////////////////////
void
DrawWindow (int x, int y, int w, int h, int wcolor)
{
    VWB_Bar (x, y, w, h, wcolor);
    DrawOutline (x, y, w, h, BORDCOLOR, BORD2COLOR);
}


void
DrawOutline (int x, int y, int w, int h, int color1, int color2)
{
    VWB_Hlin (x, x + w, y, color2);
    VWB_Vlin (y, y + h, x, color2);
    VWB_Hlin (x, x + w, y + h, color1);
    VWB_Vlin (y, y + h, x + w, color1);
}


////////////////////////////////////////////////////////////////////
//
// Setup Control Panel stuff - graphics, etc.
//
////////////////////////////////////////////////////////////////////
void
SetupControlPanel (void)
{
    //
    // CACHE GRAPHICS & SOUNDS
    //
    CA_CacheGrChunk (STARTFONT + 1);
#ifndef SPEAR
//    CacheLump (CONTROLS_LUMP_START, CONTROLS_LUMP_END);
#else
    CacheLump (BACKDROP_LUMP_START, BACKDROP_LUMP_END);
#endif

    SETFONTCOLOR (TEXTCOLOR, BKGDCOLOR);
    fontnumber = 1;
    WindowH = 200;
    if(screenHeight % 200 != 0)
        VL_ClearScreen(0);

    if (!ingame)
        CA_LoadAllSounds ();
    else
        MainMenu[savegame].active = 1;

    //
    // CENTER MOUSE
    //
    if(IN_IsInputGrabbed())
        IN_CenterMouse();
}

////////////////////////////////////////////////////////////////////
//
// SEE WHICH SAVE GAME FILES ARE AVAILABLE & READ STRING IN
//
////////////////////////////////////////////////////////////////////
void SetupSaveGames()
{
    char name[13];
    char savepath[300];

    strcpy(name, SaveName);
    for(int i = 0; i < 9; i++)
    {
        name[7] = '0' + i;
#ifdef _arch_dreamcast
        // Try to unpack file
        if(DC_LoadFromVMU(name))
        {
#endif
            if(configdir[0])
                snprintf(savepath, sizeof(savepath), "%s/%s", configdir, name);
            else
                strcpy(savepath, name);
            FILE *file = fopen (savepath, "rb");
            if (file != NULL)
            {
                gametype newstate;
                char temp[32], temptime[20];
                SaveGamesAvail[i] = 1;
                fread (&temp,sizeof(temp),1,file);
                strcpy(&SaveGameNames[i][0], temp);
                fseek (file, READ_SAVEGAMETIME, SEEK_SET);
                fread (&temptime,sizeof(temptime),1,file);
                strcpy(&SaveGameTime[i][0], temptime);
                fseek (file, READ_GAMESTATE, SEEK_SET);
                fread (&newstate,sizeof(newstate),1,file);
                SaveGameLevel[i] = newstate.episode * 10 + newstate.mapon + 1;
                fclose(file);
            }
#ifdef _arch_dreamcast
            // Remove unpacked version of file
            fs_unlink(name);
        }
#endif
    }
}

////////////////////////////////////////////////////////////////////
//
// Clean up all the Control Panel stuff
//
////////////////////////////////////////////////////////////////////
void
CleanupControlPanel (void)
{
#ifndef SPEAR
//    UnCacheLump (CONTROLS_LUMP_START, CONTROLS_LUMP_END);
#else
    UnCacheLump (BACKDROP_LUMP_START, BACKDROP_LUMP_END);
#endif

    fontnumber = 0;
}


////////////////////////////////////////////////////////////////////
//
// Handle moving gun around a menu
//
////////////////////////////////////////////////////////////////////
int HandleMenu (CP_iteminfo * item_i, CP_itemtype * items, void (*routine) (int w))
{
    char key;
    static int redrawitem = 1, lastitem = -1;
    int i, x, y, basey, exit, which, shape;
    int32_t lastBlinkTime, timer;
    ControlInfo ci;
    
    CA_CacheGrChunk (BACKDROPPIC);
    byte *source = grsegs[BACKDROPPIC];
    CacheLump (CURSOR_LUMP_START,CURSOR_LUMP_END);

    which = item_i->curpos;
    x = item_i->x & -8;
    basey = item_i->y - 2;
    y = basey + which * item_i->curspace;

    VWB_DrawPic (x, y, C_CURSOR1PIC);
    //SetTextColor (items + which, 1);

    //
    // CALL CUSTOM ROUTINE IF IT IS NEEDED
    //
    if (routine)
        routine (which);
    VW_UpdateScreen ();

    shape = C_CURSOR1PIC;
    timer = 8;
    exit = 0;
    lastBlinkTime = GetTimeCount ();
    IN_ClearKeysDown ();


    do
    {
        //
        // CHANGE GUN SHAPE
        //
        if ((int32_t)GetTimeCount () - lastBlinkTime > timer)
        {
            lastBlinkTime = GetTimeCount ();
            if (shape == C_CURSOR1PIC)
            {
                shape = C_CURSOR2PIC;
                timer = 8;
            }
            else
            {
                shape = C_CURSOR1PIC;
                timer = 70;
            }
            VWB_DrawPic (x, y, shape);
            if (routine)
                routine (which);
            VW_UpdateScreen ();
        }
        else SDL_Delay(5);

        CheckPause ();

        //
        // SEE IF ANY KEYS ARE PRESSED FOR INITIAL CHAR FINDING
        //
        key = LastASCII;
        if (key)
        {
            int ok = 0;

            if (key >= 'a')
                key -= 'a' - 'A';

            for (i = which + 1; i < item_i->amount; i++)
                if ((items + i)->active && (items + i)->string[0] == key)
                {
                    //EraseGun (item_i, items, x, y, which);
                    which = i;
                    DrawGun (item_i, items, x, &y, which, basey, routine);
                    ok = 1;
                    IN_ClearKeysDown ();
                    break;
                }

            //
            // DIDN'T FIND A MATCH FIRST TIME THRU. CHECK AGAIN.
            //
            if (!ok)
            {
                for (i = 0; i < which; i++)
                    if ((items + i)->active && (items + i)->string[0] == key)
                    {
                        EraseGun (item_i, items, x, y, which);
                        which = i;
                        DrawGun (item_i, items, x, &y, which, basey, routine);
                        IN_ClearKeysDown ();
                        break;
                    }
            }
        }

        //
        // GET INPUT
        //
        ReadAnyControl (&ci);
        switch (ci.dir)
        {
                ////////////////////////////////////////////////
                //
                // MOVE UP
                //
            case dir_North:

                EraseGun (item_i, items, x, y, which);

                //
                // ANIMATE HALF-STEP
                //
                /*if (which && (items + which - 1)->active)
                {
                    y -= 6;
                    DrawHalfStep (x, y);
                }*/

                //
                // MOVE TO NEXT AVAILABLE SPOT
                //
                do
                {
                    if (!which)
                        which = item_i->amount - 1;
                    else
                        which--;
                }
                while (!(items + which)->active);

                DrawGun (item_i, items, x, &y, which, basey, routine);
                //
                // WAIT FOR BUTTON-UP OR DELAY NEXT MOVE
                //
                TicDelay (20);
                break;

                ////////////////////////////////////////////////
                //
                // MOVE DOWN
                //
            case dir_South:

                EraseGun (item_i, items, x, y, which);
                //
                // ANIMATE HALF-STEP
                //
                /*if (which != item_i->amount - 1 && (items + which + 1)->active)
                {
                    y += 6;
                    DrawHalfStep (x, y);
                }*/

                do
                {
                    if (which == item_i->amount - 1)
                        which = 0;
                    else
                        which++;
                }
                while (!(items + which)->active);

                DrawGun (item_i, items, x, &y, which, basey, routine);

                //
                // WAIT FOR BUTTON-UP OR DELAY NEXT MOVE
                //
                TicDelay (20);
                break;
        }

        if (ci.button0 || Keyboard[sc_Space] || Keyboard[sc_Enter])
            exit = 1;

        if (ci.button1 && !Keyboard[sc_Alt] || Keyboard[sc_Escape])
            exit = 2;

    }
    while (!exit);


    IN_ClearKeysDown ();

    //
    // ERASE EVERYTHING
    //
    if (lastitem != which)
    {
        VL_MemToScreenScaledCoord(source, 320,200, x, y, x*scaleFactor, y*scaleFactor,21,19);
        //VWB_Bar (x - 1, y, 25, 16, BKGDCOLOR);
        //PrintX = item_i->x + item_i->indent;
        //PrintY = item_i->y + which * 13;
        //US_Print ((items + which)->string);
        redrawitem = 1;
    }
    else
        redrawitem = 0;

    if (routine)
        routine (which);
    VW_UpdateScreen ();

    item_i->curpos = which;

    lastitem = which;
    switch (exit)
    {
        case 1:
            //
            // CALL THE ROUTINE
            //
            if ((items + which)->routine != NULL)
            {
                ShootSnd ();
                MenuFadeOut ();
                (items + which)->routine (0);
            }
            return which;

        case 2:
            SD_PlaySound (DSSWTCHX);
            return -1;
    }

    UnCacheLump (CURSOR_LUMP_START,CURSOR_LUMP_END);
    UNCACHEGRCHUNK (BACKDROPPIC);
    return 0;                   // JUST TO SHUT UP THE ERROR MESSAGES!
}


//
// ERASE GUN & DE-HIGHLIGHT STRING
//
void
EraseGun (CP_iteminfo * item_i, CP_itemtype * items, int x, int y, int which)
{
    byte *source = grsegs[BACKDROPPIC];
    VL_MemToScreenScaledCoord(source, 320,200, x, y, x*scaleFactor, y*scaleFactor,21,19);
    SetTextColor (items + which, 0);

    PrintX = item_i->x + item_i->indent;
    PrintY = item_i->y + which * item_i->curspace;
    US_Print ((items + which)->string);
    VW_UpdateScreen ();
}

//
// DRAW GUN AT NEW POSITION
//
void
DrawGun (CP_iteminfo * item_i, CP_itemtype * items, int x, int *y, int which, int basey,
         void (*routine) (int w))
{
    byte *source = grsegs[BACKDROPPIC];
    VL_MemToScreenScaledCoord(source, 320,200, x, *y, x*scaleFactor, *y*scaleFactor,21,19);
    *y = basey + which * item_i->curspace;
    VWB_DrawPic (x, *y, C_CURSOR1PIC);
    SetTextColor (items + which, 1);

    PrintX = item_i->x + item_i->indent;
    PrintY = item_i->y + which * item_i->curspace;
    US_Print ((items + which)->string);

    //
    // CALL CUSTOM ROUTINE IF IT IS NEEDED
    //
    if (routine)
        routine (which);
    VW_UpdateScreen ();
    SD_PlaySound (DSPSTOP);
}

////////////////////////////////////////////////////////////////////
//
// DELAY FOR AN AMOUNT OF TICS OR UNTIL CONTROLS ARE INACTIVE
//
////////////////////////////////////////////////////////////////////
void
TicDelay (int count)
{
    ControlInfo ci;

    int32_t startTime = GetTimeCount ();
    do
    {
        SDL_Delay(5);
        ReadAnyControl (&ci);
    }
    while ((int32_t) GetTimeCount () - startTime < count && ci.dir != dir_None);
}


////////////////////////////////////////////////////////////////////
//
// Draw a menu
//
////////////////////////////////////////////////////////////////////
void
DrawMenu (CP_iteminfo * item_i, CP_itemtype * items)
{
    int i, which = item_i->curpos;


    WindowX = PrintX = item_i->x + item_i->indent;
    WindowY = PrintY = item_i->y;
    WindowW = 320;
    WindowH = 200;

    for (i = 0; i < item_i->amount; i++)
    {
        SetTextColor (items + i, which == i);

        PrintY = item_i->y + i * item_i->curspace;
        if ((items + i)->active)
            US_Print ((items + i)->string);
        else
        {
            SETFONTCOLOR (DEACTIVE, BKGDCOLOR);
            US_Print ((items + i)->string);
            SETFONTCOLOR (TEXTCOLOR, BKGDCOLOR);
        }

        US_Print ("\n");
    }
}


////////////////////////////////////////////////////////////////////
//
// SET TEXT COLOR (HIGHLIGHT OR NO)
//
////////////////////////////////////////////////////////////////////
void
SetTextColor (CP_itemtype * items, int hlight)
{
    if (hlight)
    {
        SETFONTCOLOR (color_hlite[items->active], BKGDCOLOR);
    }
    else
    {
        SETFONTCOLOR (color_norml[items->active], BKGDCOLOR);
    }
}


////////////////////////////////////////////////////////////////////
//
// WAIT FOR CTRLKEY-UP OR BUTTON-UP
//
////////////////////////////////////////////////////////////////////
void
WaitKeyUp (void)
{
    ControlInfo ci;
    while (ReadAnyControl (&ci), ci.button0 |
           ci.button1 |
           ci.button2 | ci.button3 | Keyboard[sc_Space] | Keyboard[sc_Enter] | Keyboard[sc_Escape])
    {
        IN_WaitAndProcessEvents();
    }
}


////////////////////////////////////////////////////////////////////
//
// READ KEYBOARD, JOYSTICK AND MOUSE FOR INPUT
//
////////////////////////////////////////////////////////////////////
void
ReadAnyControl (ControlInfo * ci)
{
    int mouseactive = 0;

    IN_ReadControl (0, ci);

    if (mouseenabled && IN_IsInputGrabbed())
    {
        int mousex, mousey, buttons;
        buttons = SDL_GetMouseState(&mousex, &mousey);
        int middlePressed = buttons & SDL_BUTTON(SDL_BUTTON_MIDDLE);
        int rightPressed = buttons & SDL_BUTTON(SDL_BUTTON_RIGHT);
        buttons &= ~(SDL_BUTTON(SDL_BUTTON_MIDDLE) | SDL_BUTTON(SDL_BUTTON_RIGHT));
        if(middlePressed) buttons |= 1 << 2;
        if(rightPressed) buttons |= 1 << 1;

        if(mousey - CENTERY < -SENSITIVE)
        {
            ci->dir = dir_North;
            mouseactive = 1;
        }
        else if(mousey - CENTERY > SENSITIVE)
        {
            ci->dir = dir_South;
            mouseactive = 1;
        }

        if(mousex - CENTERX < -SENSITIVE)
        {
            ci->dir = dir_West;
            mouseactive = 1;
        }
        else if(mousex - CENTERX > SENSITIVE)
        {
            ci->dir = dir_East;
            mouseactive = 1;
        }

        if(mouseactive)
            IN_CenterMouse();

        if (buttons)
        {
            ci->button0 = buttons & 1;
            ci->button1 = buttons & 2;
            ci->button2 = buttons & 4;
            ci->button3 = false;
            mouseactive = 1;
        }
    }

    if (joystickenabled && !mouseactive)
    {
        int jx, jy, jb;

        IN_GetJoyDelta (&jx, &jy);
        if (jy < -SENSITIVE)
            ci->dir = dir_North;
        else if (jy > SENSITIVE)
            ci->dir = dir_South;

        if (jx < -SENSITIVE)
            ci->dir = dir_West;
        else if (jx > SENSITIVE)
            ci->dir = dir_East;

        jb = IN_JoyButtons ();
        if (jb)
        {
            ci->button0 = jb & 1;
            ci->button1 = jb & 2;
            ci->button2 = jb & 4;
            ci->button3 = jb & 8;
        }
    }
}


////////////////////////////////////////////////////////////////////
//
// DRAW DIALOG AND CONFIRM YES OR NO TO QUESTION
//
////////////////////////////////////////////////////////////////////
int Confirm (const char *string)
{
    int xit = 0, tick = 0;
    ControlInfo ci;
    
    ClearMScreen();
    Message (string);
    IN_ClearKeysDown ();
    WaitKeyUp ();

    do
    {
        ReadAnyControl(&ci);
    }
    while (!Keyboard[sc_Y] && !Keyboard[sc_N] && !Keyboard[sc_Escape] && !ci.button0 && !ci.button1);

    if (Keyboard[sc_Y] || ci.button0)
    {
        xit = 1;
        ShootSnd ();
        SDL_Delay(1250);
    }
    else if (Keyboard[sc_N] || Keyboard[sc_Escape] || ci.button1)
    {
        SD_PlaySound(DSSWTCHX);
        xit=0;
    }

    IN_ClearKeysDown ();
    WaitKeyUp ();

    return xit;
}

#ifdef JAPAN
////////////////////////////////////////////////////////////////////
//
// DRAW MESSAGE & GET Y OR N
//
////////////////////////////////////////////////////////////////////
int
GetYorN (int x, int y, int pic)
{
    int xit = 0;
    soundnames whichsnd[2] = { ESCPRESSEDSND, SHOOTSND };


    CA_CacheGrChunk (pic);
    VWB_DrawPic (x * 8, y * 8, pic);
    UNCACHEGRCHUNK (pic);
    VW_UpdateScreen ();
    IN_ClearKeysDown ();

    do
    {
        IN_WaitAndProcessEvents();
    }
#ifdef SPANISH
    while (!Keyboard[sc_S] && !Keyboard[sc_N] && !Keyboard[sc_Escape]);
#else
    while (!Keyboard[sc_Y] && !Keyboard[sc_N] && !Keyboard[sc_Escape]);
#endif

#ifdef SPANISH
    if (Keyboard[sc_S])
    {
        xit = 1;
        ShootSnd ();
    }

    while (Keyboard[sc_S] || Keyboard[sc_N] || Keyboard[sc_Escape])
        IN_WaitAndProcessEvents();

#else

    if (Keyboard[sc_Y])
    {
        xit = 1;
        ShootSnd ();
    }

    while (Keyboard[sc_Y] || Keyboard[sc_N] || Keyboard[sc_Escape])
        IN_WaitAndProcessEvents();
#endif

    IN_ClearKeysDown ();
    SD_PlaySound (whichsnd[xit]);
    return xit;
}
#endif


////////////////////////////////////////////////////////////////////
//
// PRINT A MESSAGE IN A WINDOW
//
////////////////////////////////////////////////////////////////////
void Message (const char *string)
{
    int h = 0, w = 0, mw = 0, i, len = (int) strlen(string);
    CacheLump (MESSAGE_LUMP_START,MESSAGE_LUMP_END);
    
    h=7;
    for (i = 0; i < len; i++)
    {
        if (string[i] == '\n')
        {
            if (w > mw)
                mw = w;
            w = 0;
            h+=7;
        }
        else
            w+=7;
    }

    if (w + 10 > mw)
        mw = w + 10;

    PrintY = (WindowH / 2) - h / 2;
    PrintX = WindowX = 160 - mw / 2;

    LSWrite (PrintX,PrintY,string);
    VW_UpdateScreen ();
}

////////////////////////////////////////////////////////////////////
//
// THIS MAY BE FIXED A LITTLE LATER...
//
////////////////////////////////////////////////////////////////////
static int lastmusic;

int
StartCPMusic (int song)
{
    int lastoffs;

    lastmusic = song;
    lastoffs = SD_MusicOff ();

    SD_StartMusic(STARTMUSIC + song);
    return lastoffs;
}


///////////////////////////////////////////////////////////////////////////
//
//      IN_GetScanName() - Returns a string containing the name of the
//              specified scan code
//
///////////////////////////////////////////////////////////////////////////
const char *
IN_GetScanName (ScanCode scan)
{
/*    const char **p;
    ScanCode *s;

    for (s = ExtScanCodes, p = ExtScanNames; *s; p++, s++)
        if (*s == scan)
            return (*p);*/

    return (ScanNames[scan]);
}


///////////////////////////////////////////////////////////////////////////
//
// CHECK FOR PAUSE KEY (FOR MUSIC ONLY)
//
///////////////////////////////////////////////////////////////////////////
void
CheckPause (void)
{
    if (Paused)
    {
        /*switch (SoundStatus)
        {
            case 0:
                SD_MusicOn ();
                break;
            case 1:
                SD_MusicOff ();
                break;
        }*/

        SoundStatus ^= 1;
        VW_WaitVBL (3);
        IN_ClearKeysDown ();
        Paused = false;
    }
}


///////////////////////////////////////////////////////////////////////////
//
// DRAW GUN CURSOR AT CORRECT POSITION IN MENU
//
///////////////////////////////////////////////////////////////////////////
void
DrawMenuGun (CP_iteminfo * iteminfo)
{
    int x, y;


    x = iteminfo->x;
    y = iteminfo->y + iteminfo->curpos * iteminfo->curspace - 2;
    VWB_DrawPic (x, y, C_CURSOR1PIC);
}

void ShootSnd (void)
{
    SD_PlaySound (DSPISTOL);
}


///////////////////////////////////////////////////////////////////////////
//
// CHECK FOR EPISODES
//
///////////////////////////////////////////////////////////////////////////
void
CheckForEpisodes (void)
{
    struct stat statbuf;

    // On Linux like systems, the configdir defaults to $HOME/.wolf4sdl
#if !defined(_WIN32) && !defined(_arch_dreamcast)
    if(configdir[0] == 0)
    {
        // Set config location to home directory for multi-user support
        char *homedir = getenv("HOME");
        if(homedir == NULL)
        {
            Quit("Your $HOME directory is not defined. You must set this before playing.");
        }
        #define WOLFDIR "/.wolf4sdl"
        if(strlen(homedir) + sizeof(WOLFDIR) > sizeof(configdir))
        {
            Quit("Your $HOME directory path is too long. It cannot be used for saving games.");
        }
        snprintf(configdir, sizeof(configdir), "%s" WOLFDIR, homedir);
    }
#endif

    if(configdir[0] != 0)
    {
        // Ensure config directory exists and create if necessary
        if(stat(configdir, &statbuf) != 0)
        {
#ifdef _WIN32
            if(_mkdir(configdir) != 0)
#else
            if(mkdir(configdir, 0755) != 0)
#endif
            {
                Quit("The configuration directory \"%s\" could not be created.", configdir);
            }
        }
    }


//
// ENGLISH
//
    if(!stat("vswap.dlh", &statbuf))
    {
        strcpy (extension, "dlh");
        NewEmenu[1].active =
            NewEmenu[2].active =
            NewEmenu[3].active =
            EpisodeSelect[1] =
            EpisodeSelect[2] = EpisodeSelect[3] = 1;
    }

    strcpy (graphext, extension);

    strcat (configname, extension);
    strcat (SaveName, extension);
    strcat (demoname, extension);
}
