#include "version.h"

#ifdef USE_PARALLAX

#include "wl_def.h"

#ifdef USE_FEATUREFLAGS

// The lower left tile of every map determines the start texture of the parallax sky.
static int GetParallaxStartTexture()
{
    int startTex = ffDataBottomLeft;
    assert(startTex >= 0 && startTex < PMSpriteStart);
    return startTex;
}

#else

static int GetParallaxStartTexture()
{
    int startTex;
    switch(gamestate.episode * 10 + mapon)
    {
        // Episode 1
        case 0:
        case 2:
        case 3:
        case 6:
        case 8:
        case 9:
            startTex = PLX_MOSSY; // mossed mountains
            break;
        
        // Episode 2
        case 10:
        case 11:
        case 12:
        case 13:
        case 15:
        case 16:
        case 17:
        case 18:
        case 19:
            startTex = PLX_ORG_SKY; // sunset clouds
            break;
        case 20:
        case 21:
        case 23:
        case 24:
        case 25:
        case 26:
        case 27:
        case 28:
        case 29:
            startTex = PLX_HLL_MTN;
            break;
        case 30:
        case 31:
        case 32:
        case 33:
        case 35:
        case 38:
        case 40:
            startTex = PLX_RED_MTN;
            break;
        case 39:
            startTex = PLX_HLL_COL;
            break;
        default: startTex =  0; break;
    }
    assert(startTex >= 0 && startTex < PMSpriteStart);
    return startTex;
}

#endif

void DrawParallax(byte *vbuf, unsigned vbufPitch)
{
    int startpage = GetParallaxStartTexture();
    if (startpage == 0)
    {
        return;
    }
    int midangle = player->angle * (FINEANGLES / ANGLES);
    int skyheight = viewheight >> 1;
    int curtex = -1;
    byte *skytex;

    startpage += USE_PARALLAX - 1;

    for(int x = 0; x < viewwidth; x++)
    {
        int curang = pixelangle[x] + midangle;
        if(curang < 0) curang += FINEANGLES;
        else if(curang >= FINEANGLES) curang -= FINEANGLES;
        int xtex = curang * 16 * TEXTURESIZE / FINEANGLES;
        int newtex = (xtex >> TEXTURESHIFT) & (USE_PARALLAX - 1);
        if(newtex != curtex)
        {
            curtex = newtex;
            skytex = PM_GetTexture(startpage - curtex);
        }
        int texoffs = TEXTUREMASK - ((xtex & (TEXTURESIZE - 1)) << TEXTURESHIFT);
        int yend = skyheight - (wallheight[x] >> 3);
        if(yend <= 0) continue;

        for(int y = 0, offs = x; y < yend; y++, offs += vbufPitch)
            vbuf[offs] = skytex[texoffs + (y * TEXTURESIZE) / skyheight];
    }
}

#endif
