// WL_PLAY.C

#include "wl_def.h"
#pragma hdrstop

#include "wl_cloudsky.h"
#include "wl_shade.h"
#include "debug.h"

/*
=============================================================================

                                                 LOCAL CONSTANTS

=============================================================================
*/

#define sc_Question     0x35

/*
=============================================================================

                                                 GLOBAL VARIABLES

=============================================================================
*/

boolean madenoise;              // true when shooting or screaming

exit_t playstate;

static musicnames lastmusicchunk = (musicnames) 0;

static int DebugOk;

objtype objlist[MAXACTORS];
objtype *newobj, *obj, *player, *lastobj, *objfreelist, *killerobj;

boolean noclip;
int singlestep, extravbls = 0;

uint16_t tilemap[MAPSIZE][MAPSIZE]; // wall values only
byte spotvis[MAPSIZE][MAPSIZE];
objtype *actorat[MAPSIZE][MAPSIZE];
uint16_t switchmap[MAPSIZE][MAPSIZE]; // switches on plane 3

//
// replacing refresh manager
//
unsigned tics;

//
// control info
//
boolean mouseenabled, joystickenabled;
int dirscan[4] = { sc_UpArrow, sc_RightArrow, sc_DownArrow, sc_LeftArrow };
int buttonscan[NUMBUTTONS] = { sc_Control, sc_Alt, sc_LShift, sc_Space, sc_1, sc_2, sc_3, sc_4 };
int buttonmouse[4] = { bt_attack, bt_strafe, bt_use, bt_nobutton };
int buttonjoy[32] = {
#ifdef _arch_dreamcast
    bt_attack, bt_strafe, bt_use, bt_run, bt_esc, bt_prevweapon, bt_nobutton, bt_nextweapon,
    bt_pause, bt_strafeleft, bt_straferight, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton,
#else
    bt_attack, bt_strafe, bt_use, bt_run, bt_strafeleft, bt_straferight, bt_esc, bt_pause,
    bt_prevweapon, bt_nextweapon, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton,
#endif
    bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton,
    bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton
};

int viewsize;

boolean buttonheld[NUMBUTTONS];

boolean demorecord, demoplayback;
int8_t *demoptr, *lastdemoptr;
memptr demobuffer;

//
// current user input
//
int controlx, controly;         // range from -100 to 100 per tic
boolean buttonstate[NUMBUTTONS];

int lastgamemusicoffset = 0;


//===========================================================================


void CenterWindow (word w, word h);
void InitObjList (void);
void RemoveObj (objtype * gone);
void PollControls (void);
int StopMusic (void);
void StartMusic (void);
void ContinueMusic (int offs);
void PlayLoop (void);

/*
=============================================================================

                                                 LOCAL VARIABLES

=============================================================================
*/


objtype dummyobj;

//
// LIST OF SONGS FOR EACH VERSION
//
int songs[] = {
D_IN_CIT_MUS,
D_E1M2_MUS,
D_E1M3_MUS,
D_QUADM_MUS,
D_E2M6_MUS,
D_DDTBLU_MUS,
D_STALKS_MUS,
D_DEAD_MUS,
D_SHAWN_MUS,
D_BETWEEN_MUS,

D_E3M2_MUS,
D_E1M9_MUS,
D_UNDERW_MUS,
D_E3M3_MUS,
D_E2M1_MUS,
D_E2M2_MUS,
DESCENT_MUS,
D_E3M1_MUS,
D_E2M8_MUS,
D_E1M4_MUS,

D_RUNNIN_MUS,
D_E2M2_MUS,
D_E2M6_MUS,
D_UNDERW_MUS,
D_E3M3_MUS,
D_DEAD_MUS,
D_E2M1_MUS,
D_IN_CIT_MUS,
D_SHAWN_MUS,
D_E1M2_MUS,

D_DDTBLU_MUS,
D_E3M1_MUS,
D_STALKS_MUS,
D_SHAWN_MUS,
D_E3M3_MUS,
D_UNDERW_MUS,
D_IN_CIT_MUS,
DESCENT_MUS,
D_QUADM_MUS,
D_E2M1_MUS,
D_E2M6_MUS,

D_DDTBLU_MUS,
D_E3M1_MUS,
D_STALKS_MUS,
D_SHAWN_MUS,
D_E3M3_MUS,
D_UNDERW_MUS,
D_IN_CIT_MUS,
DESCENT_MUS,
D_QUADM_MUS,
D_E2M1_MUS,
D_E2M6_MUS
};


/*
=============================================================================

                               USER CONTROL

=============================================================================
*/

/*
===================
=
= PollKeyboardButtons
=
===================
*/

void PollKeyboardButtons (void)
{
    int i;

    for (i = 0; i < NUMBUTTONS; i++)
        if (Keyboard[buttonscan[i]])
            buttonstate[i] = true;
}


/*
===================
=
= PollMouseButtons
=
===================
*/

void PollMouseButtons (void)
{
    int buttons = IN_MouseButtons ();

    if (buttons & 1)
        buttonstate[buttonmouse[0]] = true;
    if (buttons & 2)
        buttonstate[buttonmouse[1]] = true;
    if (buttons & 4)
        buttonstate[buttonmouse[2]] = true;
}



/*
===================
=
= PollJoystickButtons
=
===================
*/

void PollJoystickButtons (void)
{
    int buttons = IN_JoyButtons();

    for(int i = 0, val = 1; i < JoyNumButtons; i++, val <<= 1)
    {
        if(buttons & val)
            buttonstate[buttonjoy[i]] = true;
    }
}


/*
===================
=
= PollKeyboardMove
=
===================
*/

void PollKeyboardMove (void)
{
    int delta = buttonstate[bt_run] ? RUNMOVE * tics : BASEMOVE * tics;

    if (Keyboard[dirscan[di_north]])
        controly -= delta;
    if (Keyboard[dirscan[di_south]])
        controly += delta;
    if (Keyboard[dirscan[di_west]])
        controlx -= delta;
    if (Keyboard[dirscan[di_east]])
        controlx += delta;
}


/*
===================
=
= PollMouseMove
=
===================
*/

void PollMouseMove (void)
{
    int mousexmove, mouseymove;

    SDL_GetMouseState(&mousexmove, &mouseymove);
    if(IN_IsInputGrabbed())
        IN_CenterMouse();

    mousexmove -= screenWidth / 2;
    mouseymove -= screenHeight / 2;

    controlx += mousexmove * 10 / (13 - mouseadjustment);
    controly += mouseymove * 20 / (13 - mouseadjustment);
}


/*
===================
=
= PollJoystickMove
=
===================
*/

void PollJoystickMove (void)
{
    int joyx, joyy;

    IN_GetJoyDelta (&joyx, &joyy);

    int delta = buttonstate[bt_run] ? RUNMOVE * tics : BASEMOVE * tics;

    if (joyx > 64 || buttonstate[bt_turnright])
        controlx += delta;
    else if (joyx < -64  || buttonstate[bt_turnleft])
        controlx -= delta;
    if (joyy > 64 || buttonstate[bt_movebackward])
        controly += delta;
    else if (joyy < -64 || buttonstate[bt_moveforward])
        controly -= delta;
}

/*
===================
=
= PollControls
=
= Gets user or demo input, call once each frame
=
= controlx              set between -100 and 100 per tic
= controly
= buttonheld[]  the state of the buttons LAST frame
= buttonstate[] the state of the buttons THIS frame
=
===================
*/

void PollControls (void)
{
    int max, min, i;
    byte buttonbits;

    IN_ProcessEvents();

//
// get timing info for last frame
//
    if (demoplayback || demorecord)   // demo recording and playback needs to be constant
    {
        // wait up to DEMOTICS Wolf tics
        uint32_t curtime = SDL_GetTicks();
        lasttimecount += DEMOTICS;
        int32_t timediff = (lasttimecount * 100) / 7 - curtime;
        if(timediff > 0)
            SDL_Delay(timediff);

        if(timediff < -2 * DEMOTICS)       // more than 2-times DEMOTICS behind?
            lasttimecount = (curtime * 7) / 100;    // yes, set to current timecount

        tics = DEMOTICS;
    }
    else
        CalcTics ();

    controlx = 0;
    controly = 0;
    memcpy (buttonheld, buttonstate, sizeof (buttonstate));
    memset (buttonstate, 0, sizeof (buttonstate));

    if (demoplayback)
    {
        //
        // read commands from demo buffer
        //
        buttonbits = *demoptr++;
        for (i = 0; i < NUMBUTTONS; i++)
        {
            buttonstate[i] = buttonbits & 1;
            buttonbits >>= 1;
        }

        controlx = *demoptr++;
        controly = *demoptr++;

        if (demoptr == lastdemoptr)
            playstate = ex_completed;   // demo is done

        controlx *= (int) tics;
        controly *= (int) tics;

        return;
    }


//
// get button states
//
    PollKeyboardButtons ();

    if (mouseenabled && IN_IsInputGrabbed())
        PollMouseButtons ();

    if (joystickenabled)
        PollJoystickButtons ();

//
// get movements
//
    PollKeyboardMove ();

    if (mouseenabled && IN_IsInputGrabbed())
        PollMouseMove ();

    if (joystickenabled)
        PollJoystickMove ();

//
// bound movement to a maximum
//
    max = 100 * tics;
    min = -max;
    if (controlx > max)
        controlx = max;
    else if (controlx < min)
        controlx = min;

    if (controly > max)
        controly = max;
    else if (controly < min)
        controly = min;

    if (demorecord)
    {
        //
        // save info out to demo buffer
        //
        controlx /= (int) tics;
        controly /= (int) tics;

        buttonbits = 0;

        // TODO: Support 32-bit buttonbits
        for (i = NUMBUTTONS - 1; i >= 0; i--)
        {
            buttonbits <<= 1;
            if (buttonstate[i])
                buttonbits |= 1;
        }

        *demoptr++ = buttonbits;
        *demoptr++ = controlx;
        *demoptr++ = controly;

        if (demoptr >= lastdemoptr - 8)
            playstate = ex_completed;
        else
        {
            controlx *= (int) tics;
            controly *= (int) tics;
        }
    }
}



//==========================================================================



///////////////////////////////////////////////////////////////////////////
//
//      CenterWindow() - Generates a window of a given width & height in the
//              middle of the screen
//
///////////////////////////////////////////////////////////////////////////
#define MAXX    320
#define MAXY    160

void CenterWindow (word w, word h)
{
    US_DrawWindow (((MAXX / 8) - w) / 2, ((MAXY / 8) - h) / 2, w, h);
}

//===========================================================================

/*
======================
=
= In-game changeview
=
======================
*/

void ChangeScreenSizePoll()
{
        //
        // Make screen larger
        //
        if(Keyboard[sc_Equal] && viewsize<22) // Viewsize by + - keys // needs to be no higher than 19, 20 is unstable and crashes
        {                                           // 20 needs additional coding to be made stable and work properly
            ClearMemory();
            NewViewSize(viewsize+1);
            IN_ClearKeysDown();
            //SD_PlaySound (DSSTNMOV);
        }


        //
        // Make screen smaller
        //
        if(Keyboard[sc_Minus] && viewsize>20) // Viewsize by + - keys // screen has no need to be that small (high memory will not cause mem errors)
        {
            ClearMemory();
            NewViewSize(viewsize-1);
            IN_ClearKeysDown();
            if (viewsize==21)
            DrawPlayScreen();
            //SD_PlaySound (DSSTNMOV);

        }
}

ScanCode KeyDown=0;
boolean KeyIsPressed (ScanCode scan)
{

    if (scan != 0)
    {
        KeyDown = scan;
        LastScan=0;
        return true;
    }
   return false;
}
/*
=====================
=
= CheckKeys
=
=====================
*/

//===========================================================================
// Massive bunches of cheat shit
//  to keep it from being easy to figure them out.
// Yeah, right...
unsigned char	cheat_mus_seq[] =
{
    SDLK_c, SDLK_i, SDLK_m, SDLK_u, SDLK_s, 1, 1, 0xff //cimusXX
};

unsigned char	cheat_choppers_seq[] =
{
    SDLK_c, SDLK_i, SDLK_s, SDLK_a, SDLK_w, SDLK_z, SDLK_a, SDLK_l, SDLK_l, 0xff // cisawzall
};

unsigned char	cheat_god_seq[6] =
{
    SDLK_c, SDLK_i, SDLK_g, SDLK_o, SDLK_d, 0xff // cigod
};

unsigned char	cheat_ammo_seq[] =
{
    SDLK_c, SDLK_i, SDLK_a, SDLK_m, SDLK_m, SDLK_o, 0xff	// ciammo
};

unsigned char	cheat_ammonokey_seq[] =
{
    SDLK_c, SDLK_i, SDLK_c, SDLK_h, SDLK_e, SDLK_a, SDLK_p, 0xff	// cicheap
};

unsigned char	cheat_powerup_seq[7][10] =
{
    { SDLK_c, SDLK_i, SDLK_g, SDLK_i, SDLK_m, SDLK_m, SDLK_e, SDLK_v, 0xff }, 	// cigimmev
    { SDLK_c, SDLK_i, SDLK_g, SDLK_i, SDLK_m, SDLK_m, SDLK_e, SDLK_s, 0xff }, 	// cigimmes
    { SDLK_c, SDLK_i, SDLK_g, SDLK_i, SDLK_m, SDLK_m, SDLK_e, SDLK_i, 0xff }, 	// cigimmei
    { SDLK_c, SDLK_i, SDLK_g, SDLK_i, SDLK_m, SDLK_m, SDLK_e, SDLK_r, 0xff }, 	// cigimmer
    { SDLK_c, SDLK_i, SDLK_g, SDLK_i, SDLK_m, SDLK_m, SDLK_e, SDLK_a, 0xff }, 	// cigimmea
    { SDLK_c, SDLK_i, SDLK_g, SDLK_i, SDLK_m, SDLK_m, SDLK_e, SDLK_l, 0xff }, 	// cigimmel
    { SDLK_c, SDLK_i, SDLK_g, SDLK_i, SDLK_m, SDLK_m, SDLK_e, 0xff }			// cigimme
};


unsigned char	cheat_clev_seq[] =
{
    SDLK_c, SDLK_i,  SDLK_d, SDLK_e, SDLK_s, SDLK_t, 1, 1, 0xff	// cidestXX
};

// Now what?
cheatseq_t	cheat_mus = { cheat_mus_seq, 0 };
cheatseq_t	cheat_god = { cheat_god_seq, 0 };
cheatseq_t	cheat_ammo = { cheat_ammo_seq, 0 };
cheatseq_t	cheat_ammonokey = { cheat_ammonokey_seq, 0 };

cheatseq_t	cheat_powerup[7] =
{
    { cheat_powerup_seq[0], 0 },
    { cheat_powerup_seq[1], 0 },
    { cheat_powerup_seq[2], 0 },
    { cheat_powerup_seq[3], 0 },
    { cheat_powerup_seq[4], 0 },
    { cheat_powerup_seq[5], 0 },
    { cheat_powerup_seq[6], 0 }
};

cheatseq_t	cheat_choppers = { cheat_choppers_seq, 0 };
cheatseq_t	cheat_clev = { cheat_clev_seq, 0 };

void CheatCodeCheck() {
    // godmode cheat
    if (cht_CheckCheat(&cheat_god, KeyDown))
    {
        gamestate.godmode ^= 1;
        if (gamestate.godmode)
        {
            gamestate.health = 100;
            GetMessage ("Degreelessness Mode ON");
        }
        else 
            GetMessage ("Degreelessness Mode OFF");
    }
        
    // 'fa' cheat for killer fucking arsenal
    if (cht_CheckCheat(&cheat_ammonokey, KeyDown))
    {
        gamestate.armor = 200;
        gamestate.bestweapon = (weapontype) 511;
        gamestate.ammo = gamestate.ammomax = MAXAMMO;
        gamestate.ammo2 = gamestate.ammo2max= MAXAMMO2;
        gamestate.ammo3 = gamestate.ammo3max = MAXAMMO3;
        gamestate.ammo4 = gamestate.ammo4max = MAXAMMO4;
        GetMessage ("Ammo (no keys) added!");
    }
    // 'kfa' cheat for key full ammo
    if (cht_CheckCheat(&cheat_ammo,KeyDown))
    {
        gamestate.armor = 200;
        gamestate.bestweapon = (weapontype) 511;
        gamestate.ammo = gamestate.ammomax = MAXAMMO;
        gamestate.ammo2 = gamestate.ammo2max = MAXAMMO2;
        gamestate.ammo3 = gamestate.ammo3max = MAXAMMO3;
        gamestate.ammo4 = gamestate.ammo4max = MAXAMMO4;
        gamestate.keys = 7;
        GetMessage ("Very happy ammo added!");
    }

    // 'mus' cheat for changing music
    if (cht_CheckCheat(&cheat_mus, KeyDown))
    {
        char	buf[2];
        int		musnum;
        int		epsd;
        int		map;
            
        cht_GetParam(&cheat_mus, buf);

        epsd = buf[0] - '0';
        map = buf[1] - '0';
        
        GetMessage ("Music Change");
        cht_GetParam(&cheat_mus, buf);
        // Catch invalid maps.
        if (epsd < 1 || epsd > (NUMMAPS/10)+1 || map < 0 || map > 9)
        {
            GetMessage("Invalid entry");
        }
        else
        {
            musnum = (musicnames) songs[(epsd - 1)*10 + (map == 0 ? 9 : map - 1)];
            StartCPMusic(musnum);
        }
    }

    // 'behold' power-up menu
    if (cht_CheckCheat(&cheat_powerup[6], KeyDown))
    {

        GetMessage ("Shield, invis, god, map, visor, berserk"); // if map is enabled
        //GetMessage ("Shield, invis, god, visor, berserk");
    }
    //v(god) s(berserk) i(invis) r(radiation) a(automap) l(light amp goggles)
    if (cht_CheckCheat(&cheat_powerup[0], KeyDown))
    {
        TimedBonus (p_godmode,BONUS_TIME);
        GetMessage ("Behold you godliness!");
    }
    else if (cht_CheckCheat(&cheat_powerup[1], KeyDown))
    {
        TimedBonus (p_berserk,BONUS_TIME);
        GetMessage ("Behold the berserk pack!");
    }
    else if (cht_CheckCheat(&cheat_powerup[2], KeyDown))
    {
        TimedBonus (p_invisibility,BONUS_TIME*2);
        GetMessage ("Behold your invisibility!");
    }
    else if (cht_CheckCheat(&cheat_powerup[3], KeyDown))
    {
        TimedBonus (p_biosuit,BONUS_TIME*2);
        GetMessage ("Behold the Radiation Suit!");
    }
    /*else if (cht_CheckCheat(&cheat_powerup[4], KeyDown))
    {
        //computerareamap=true;
        // Fill the entire map
        GetMessage ("Behold you have a map!");
    }*/
    else if (cht_CheckCheat(&cheat_powerup[5], KeyDown))
    {
        TimedBonus (p_lightamp,BONUS_TIME);
        GetMessage ("Behold you can see!");
    }

    if (cht_CheckCheat(&cheat_choppers, KeyDown))
    {
        GiveWeapon (wp_chainsaw,0);
        GetMessage ("I <3 Chainsaws!!!");
    }
    
    // 'clev' change-level cheat
    if (cht_CheckCheat(&cheat_clev, KeyDown))
    {
        char	buf[2];
        int		epsd;
        int		map;
            
        cht_GetParam(&cheat_clev, buf);

        epsd = buf[0] - '0';
        map = buf[1] - '0';

        // Catch invalid maps.
        if (epsd < 1 || epsd > (NUMMAPS/10) || map < 0 || map > 9)
        {
            GetMessage("Invalid map");
        }
        else
        {
            // So be it.
            GetMessage ("Changing Levels");
            gamestate.episode = epsd-1;
            gamestate.mapon = map == 0 ? 9 : map-1; // 0 goes to level 10
            playstate = ex_warped;
        }
    }
}

void PowerupCheck() {
    if (gamestate.bonusgodmode)
    {
        if (gamestate.godmodecount <=0)
        {
            gamestate.godmodecount=0;
            gamestate.bonusgodmode=false;
        }
        else
        {
            gamestate.godmodecount-=tics;
        }
    }
    if (gamestate.berserk)
    {
        gamestate.berserkcount += tics;
        if (gamestate.berserkcount > 20l*70)
        {
            gamestate.berserkcount = 0;
            gamestate.berserk=false;
        }
    }

    if (gamestate.lightamp)
    {
        if (gamestate.lightampcount <=0)
        {
            gamestate.lightampcount=0;
            gamestate.lightamp=false;
        }
        else
        {  
            //lightamp = 1;
            gamestate.lightampcount-=tics;
        }
    }
    if (gamestate.invisibility)
    {
        if (gamestate.invisibilitycount <=0)
        {
            gamestate.invisibilitycount=0;
            gamestate.invisibility=false;
        }
        else
        {  
            gamestate.invisibilitycount-=tics;
        }
    }
    if (gamestate.biosuit)
    {
        if (gamestate.biosuitcount <=0)
        {
            gamestate.biosuitcount=0;
            gamestate.biosuit=false;
        }
        else
        {  
            gamestate.biosuitcount-=tics;
        }
    }
}

void CheckKeys (void)
{
    ScanCode scan;


    if (screenfaded || demoplayback)    // don't do anything with a faded screen
        return;

    scan = LastScan;

    if (KeyIsPressed(scan))
    {
        CheckWeaponChange ();
        
        CheatCodeCheck();
    }
    
/////////////
//
// timing for power-ups
//
/////////////
    PowerupCheck();
/////////////
//
// Screen size
//
/////////////
    ChangeScreenSizePoll();
//
// pause key weirdness can't be checked as a scan code
//
    if(buttonstate[bt_pause]) Paused = true;
    if(Paused)
    {
        //int lastoffs = StopMusic();
        CacheLump(MESSAGE_LUMP_START, MESSAGE_LUMP_END);
        LSWrite(136, 96, "Paused");
        VH_UpdateScreen();
        IN_Ack ();
        Paused = false;
        //ContinueMusic(lastoffs);
        UnCacheLump(MESSAGE_LUMP_START, MESSAGE_LUMP_END);
        if (MousePresent && IN_IsInputGrabbed())
            IN_CenterMouse();     // Clear accumulated mouse movement
        lasttimecount = GetTimeCount();
        return;
    }

//
// F1-F7/ESC to enter control panel
//
    if (
#ifndef DEBCHECK
           scan == sc_F10 ||
#endif
           scan == sc_F9 || scan == sc_F7 || scan == sc_F8)     // pop up quit dialog
    {
        short oldmapon = gamestate.mapon;
        short oldepisode = gamestate.episode;
        ClearMemory ();
        ClearSplitVWB ();
        US_ControlPanel (scan);

        SETFONTCOLOR (0, 15);
        IN_ClearKeysDown ();
        return;
    }

    if ((scan >= sc_F1 && scan <= sc_F9) || scan == sc_Escape || buttonstate[bt_esc])
    {
        int lastoffs = StopMusic ();
        VL_SetSaveGameSlot();
        ClearMemory ();
        VW_FadeOut ();

        US_ControlPanel (buttonstate[bt_esc] ? sc_Escape : scan);

        SETFONTCOLOR (0, 15);
        IN_ClearKeysDown ();
        VW_FadeOut();
        if (!startgame && !loadedgame)
            ContinueMusic (lastoffs);
        if (loadedgame)
            playstate = ex_abort;
        lasttimecount = GetTimeCount();
        if (MousePresent && IN_IsInputGrabbed())
            IN_CenterMouse();     // Clear accumulated mouse movement
        return;
    }

//
// TAB-? debug keys
//
#ifdef DEBUGKEYS
    if (Keyboard[sc_Tab] && DebugOk)
    {
        CA_CacheGrChunk (STARTFONT);
        fontnumber = 0;
        SETFONTCOLOR (0, 15);

        if (MousePresent && IN_IsInputGrabbed())
            IN_CenterMouse();     // Clear accumulated mouse movement

        lasttimecount = GetTimeCount();
        return;
    }
#endif
}


//===========================================================================

/*
#############################################################################

                                  The objlist data structure

#############################################################################

objlist containt structures for every actor currently playing.  The structure
is accessed as a linked list starting at *player, ending when ob->next ==
NULL.  GetNewObj inserts a new object at the end of the list, meaning that
if an actor spawn another actor, the new one WILL get to think and react the
same frame.  RemoveObj unlinks the given object and returns it to the free
list, but does not damage the objects ->next pointer, so if the current object
removes itself, a linked list following loop can still safely get to the
next element.

<backwardly linked free list>

#############################################################################
*/


/*
=========================
=
= InitActorList
=
= Call to clear out the actor object lists returning them all to the free
= list.  Allocates a special spot for the player.
=
=========================
*/

int objcount;

void InitActorList (void)
{
    int i;

//
// init the actor lists
//
    for (i = 0; i < MAXACTORS; i++)
    {
        objlist[i].prev = &objlist[i + 1];
        objlist[i].next = NULL;
    }

    objlist[MAXACTORS - 1].prev = NULL;

    objfreelist = &objlist[0];
    lastobj = NULL;

    objcount = 0;

//
// give the player the first free spots
//
    GetNewActor ();
    player = newobj;

}

//===========================================================================

/*
=========================
=
= GetNewActor
=
= Sets the global variable new to point to a free spot in objlist.
= The free spot is inserted at the end of the liked list
=
= When the object list is full, the caller can either have it bomb out ot
= return a dummy object pointer that will never get used
=
=========================
*/

void GetNewActor (void)
{
    if (!objfreelist)
        Quit ("GetNewActor: No free spots in objlist!");

    newobj = objfreelist;
    objfreelist = newobj->prev;
    memset (newobj, 0, sizeof (*newobj));

    if (lastobj)
        lastobj->next = newobj;
    newobj->prev = lastobj;     // new->next is allready NULL from memset

    newobj->active = ac_no;
    lastobj = newobj;

    objcount++;
}

//===========================================================================

/*
=========================
=
= RemoveObj
=
= Add the given object back into the free list, and unlink it from it's
= neighbors
=
=========================
*/

void RemoveObj (objtype * gone)
{
    if (gone == player)
        Quit ("RemoveObj: Tried to remove the player!");

    gone->state = NULL;

//
// fix the next object's back link
//
    if (gone == lastobj)
        lastobj = (objtype *) gone->prev;
    else
        gone->next->prev = gone->prev;

//
// fix the previous object's forward link
//
    gone->prev->next = gone->next;

//
// add it back in to the free list
//
    gone->prev = objfreelist;
    objfreelist = gone;

    objcount--;
}

/*
=============================================================================

                                                MUSIC STUFF

=============================================================================
*/


/*
=================
=
= StopMusic
=
=================
*/
int StopMusic (void)
{
    int lastoffs = SD_MusicOff ();

    UNCACHEAUDIOCHUNK (STARTMUSIC + lastmusicchunk);

    return lastoffs;
}

//==========================================================================


/*
=================
=
= StartMusic
=
=================
*/

void StartMusic ()
{
    SD_MusicOff ();
    lastmusicchunk = (musicnames) songs[gamestate.mapon + gamestate.episode * 10];
    SD_StartMusic(STARTMUSIC + lastmusicchunk);
}

void ContinueMusic (int offs)
{
    SD_MusicOff ();
    lastmusicchunk = (musicnames) songs[gamestate.mapon + gamestate.episode * 10];
    SD_ContinueMusic(STARTMUSIC + lastmusicchunk, offs);
}

/*
=============================================================================

                                        PALETTE SHIFTING STUFF

=============================================================================
*/

#define NUMREDSHIFTS    6
#define REDSTEPS        8

#define NUMWHITESHIFTS  3
#define WHITESTEPS      20
#define WHITETICS       6

#define NUMGODSHIFTS  9
#define GODSTEPS      5

#define NUMGREENSHIFTS 3
#define GREENSTEPS     20

SDL_Color redshifts[NUMREDSHIFTS][256];
SDL_Color whiteshifts[NUMWHITESHIFTS][256];
SDL_Color godshifts[NUMGODSHIFTS][256];
SDL_Color greenshifts[NUMGREENSHIFTS][256];

int damagecount, bonuscount, godinvertcount, biosuitcount;
boolean palshifted;

/*
=====================
=
= InitRedShifts
=
=====================
*/

void InitRedShifts (void)
{
    SDL_Color *workptr, *baseptr;
    int i, j, delta;


//
// fade through intermediate frames
//
    for (i = 1; i <= NUMREDSHIFTS; i++)
    {
        workptr = redshifts[i - 1];
        baseptr = gamepal;

        for (j = 0; j <= 255; j++)
        {
            delta = 256 - baseptr->r;
            workptr->r = baseptr->r + delta * i / REDSTEPS;
            delta = -baseptr->g;
            workptr->g = baseptr->g + delta * i / REDSTEPS;
            delta = -baseptr->b;
            workptr->b = baseptr->b + delta * i / REDSTEPS;
            baseptr++;
            workptr++;
        }
    }

    for (i = 1; i <= NUMWHITESHIFTS; i++)
    {
        workptr = whiteshifts[i - 1];
        baseptr = gamepal;

        for (j = 0; j <= 255; j++)
        {
            delta = 256 - baseptr->r;
            workptr->r = baseptr->r + delta * i / WHITESTEPS;
            delta = 248 - baseptr->g;
            workptr->g = baseptr->g + delta * i / WHITESTEPS;
            delta = 0-baseptr->b;
            workptr->b = baseptr->b + delta * i / WHITESTEPS;
            baseptr++;
            workptr++;
        }
    }

    for (i = 1; i <= NUMGREENSHIFTS; i++)
    {
        workptr = greenshifts[i - 1];
        baseptr = gamepal;

        for (j = 0; j <= 255; j++)
        {
            delta = -baseptr->r;
            workptr->r = baseptr->r + delta * i / GREENSTEPS;
            delta = 256-baseptr->g;
            workptr->g = baseptr->g + delta * i / GREENSTEPS;
            delta = -baseptr->b;
            workptr->b = baseptr->b + delta * i / GREENSTEPS;
            baseptr++;
            workptr++;
        }
    }

    for (i = 1; i <= NUMGODSHIFTS; i++)
    {
        workptr = godshifts[i - 1];
        baseptr = gamepal;

        for (j = 0; j <= 255; j++)
        {
            delta = 256 - baseptr->r;
            workptr->r = baseptr->r + delta * i / GODSTEPS;
            delta = 256 - baseptr->g;
            workptr->g = baseptr->g + delta * i / GODSTEPS;
            delta = 256-baseptr->b;
            workptr->b = baseptr->b + delta * i / GODSTEPS;
            baseptr++;
            workptr++;
        }
    }
}


/*
=====================
=
= ClearPaletteShifts
=
=====================
*/

void ClearPaletteShifts (void)
{
    bonuscount = damagecount = godinvertcount = biosuitcount = 0;
    palshifted = false;
}


/*
=====================
=
= GodInvertFlash
=
=====================
*/

void GodInvertFlash (void)
{
    godinvertcount = NUMGODSHIFTS * 6;    // god shift palette
}

/*
=====================
=
= StartBiosuitFlash
=
=====================
*/

void StartBiosuitFlash (void)
{
    if (!gamestate.godmode || !gamestate.bonusgodmode)
    biosuitcount = NUMGREENSHIFTS * WHITETICS;    // god shift palette
}

/*
=====================
=
= StartBonusFlash
=
=====================
*/

void StartBonusFlash (void)
{
    if (!gamestate.godmode || !gamestate.bonusgodmode) {
        bonuscount = NUMWHITESHIFTS * WHITETICS;    // white shift palette
    }
}


/*
=====================
=
= StartDamageFlash
=
=====================
*/

void StartDamageFlash (int damage)
{
    if (!gamestate.godmode || !gamestate.bonusgodmode) {
        damagecount += damage;
    }
}


/*
=====================
=
= UpdatePaletteShifts
=
=====================
*/

void UpdatePaletteShifts (void)
{
    int red, white, godinvert, green;

    if (bonuscount)
    {
        white = bonuscount / WHITETICS + 1;
        if (white > NUMWHITESHIFTS)
            white = NUMWHITESHIFTS;
        bonuscount -= tics;
        if (bonuscount < 0)
            bonuscount = 0;
    }
    else
        white = 0;

    if (biosuitcount)
    {
        green = biosuitcount / WHITETICS + 1;
        if (green > NUMGREENSHIFTS)
            green = NUMGREENSHIFTS;
        biosuitcount -= tics;
        if (biosuitcount < 0)
            biosuitcount = 0;
    }
    else
        green = 0;

    if (godinvertcount)
    {
        godinvert = godinvertcount / WHITETICS + 1;
        if (godinvert > NUMGODSHIFTS)
            godinvert = NUMGODSHIFTS;
        //godinvertcount -= tics*4;
        //if (godinvertcount < 0)
            godinvertcount = 0;
    }
    else
        godinvert = 0;

    if (damagecount)
    {
        red = damagecount / 10 + 1;
        if (red > NUMREDSHIFTS)
            red = NUMREDSHIFTS;

        damagecount -= tics;
        if (damagecount < 0)
            damagecount = 0;
    }
    else
        red = 0;

    if (red)
    {
        VL_SetPalette (redshifts[red - 1], false);
        palshifted = true;
    }
    else if (white)
    {
        VL_SetPalette (whiteshifts[white - 1], false);
        palshifted = true;
    }
    else if (green)
    {
        VL_SetPalette (greenshifts[green - 1], false);
        palshifted = true;
    }
    else if (godinvert)
    {
        VL_SetPalette (godshifts[godinvert - 1], false);
        palshifted = true;
    }
    else if (palshifted)
    {
        VL_SetPalette (gamepal, false);        // back to normal
        palshifted = false;
    }
}


/*
=====================
=
= FinishPaletteShifts
=
= Resets palette to normal if needed
=
=====================
*/

void FinishPaletteShifts (void)
{
    if (palshifted)
    {
        palshifted = 0;
        VL_SetPalette (gamepal, true);
    }
}


/*
=============================================================================

                                                CORE PLAYLOOP

=============================================================================
*/


/*
=====================
=
= DoActor
=
=====================
*/

void DoActor (objtype * ob)
{
    void (*think) (objtype *);

    if (!ob->active && !areabyplayer[ob->areanumber])
        return;

    if (!(ob->flags & (FL_NONMARK | FL_NEVERMARK)))
        actorat[ob->tilex][ob->tiley] = NULL;

//
// non transitional object
//

    if (!ob->ticcount)
    {
        think = (void (*)(objtype *)) ob->state->think;
        if (think)
        {
            think (ob);
            if (!ob->state)
            {
                RemoveObj (ob);
                return;
            }
        }

        if (ob->flags & FL_NEVERMARK)
            return;

        if ((ob->flags & FL_NONMARK) && actorat[ob->tilex][ob->tiley])
            return;

        actorat[ob->tilex][ob->tiley] = ob;
        return;
    }

//
// transitional object
//
    ob->ticcount -= (short) tics;
    while (ob->ticcount <= 0)
    {
        think = (void (*)(objtype *)) ob->state->action;        // end of state action
        if (think)
        {
            think (ob);
            if (!ob->state)
            {
                RemoveObj (ob);
                return;
            }
        }

        ob->state = ob->state->next;

        if (!ob->state)
        {
            RemoveObj (ob);
            return;
        }

        if (!ob->state->tictime)
        {
            ob->ticcount = 0;
            goto think;
        }

        ob->ticcount += ob->state->tictime;
    }

think:
    //
    // think
    //
    think = (void (*)(objtype *)) ob->state->think;
    if (think)
    {
        think (ob);
        if (!ob->state)
        {
            RemoveObj (ob);
            return;
        }
    }

    if (ob->flags & FL_NEVERMARK)
        return;

    if ((ob->flags & FL_NONMARK) && actorat[ob->tilex][ob->tiley])
        return;

    actorat[ob->tilex][ob->tiley] = ob;
}

//==========================================================================

void ChangeWeaponAnim()
{
    if(gamestate.changingweapon == true)
    {
        if(gamestate.goingdown == false)
        {
            gamestate.weapchange -= 10;
            if(gamestate.weapchange <= 0)
            {
                gamestate.weapchange = 0;
                gamestate.changingweapon = false;
            }
        }
        else
        {
            gamestate.weapchange += 10;
            if(gamestate.weapchange >= 200)
            {
                gamestate.goingdown = false;
                gamestate.chosenweapon = gamestate.weapon = gamestate.nextweapon;
            }
        }
    }
}

/*
===================
=
= PlayLoop
=
===================
*/
int32_t funnyticount;


void PlayLoop (void)
{
    boolean autosave = false;
    if (playstate == ex_completed || playstate == ex_secretlevel) {
        autosave = true;
    }
#if defined(USE_FEATUREFLAGS) && defined(USE_CLOUDSKY)
    if(GetFeatureFlags() & FF_CLOUDSKY)
        InitSky();
#endif

#ifdef USE_SHADING
    InitLevelShadeTable();
#endif

    playstate = ex_stillplaying;
    lasttimecount = GetTimeCount();
    frameon = 0;
    anglefrac = 0;
    facecount = 0;
    funnyticount = 0;
    memset (buttonstate, 0, sizeof (buttonstate));
    ClearPaletteShifts ();

    if (MousePresent && IN_IsInputGrabbed())
        IN_CenterMouse();         // Clear accumulated mouse movement

    if (demoplayback)
        IN_StartAck ();

    do
    {
        PollControls ();

//
// actor thinking
//
        madenoise = false;

        MoveDoors ();
        MovePWalls ();

        for (obj = player; obj; obj = obj->next)
            DoActor (obj);

        UpdatePaletteShifts ();

        ThreeDRefresh ();

        if (optionstate.autosave && autosave) {
            VL_SetSaveGameSlot();
            AutoSave();
            autosave = false;
            GetMessage("Game autosaved.");
        }

        ChangeWeaponAnim();

        //
        // MAKE FUNNY FACE IF BJ DOESN'T MOVE FOR AWHILE
        //
#ifdef SPEAR
        funnyticount += tics;
        if (funnyticount > 30l * 70)
        {
            funnyticount = 0;
            if(viewsize != 21)
                StatusDrawFace(BJWAITING1PIC + (US_RndT () & 1));
            facecount = 0;
        }
#endif

        gamestate.TimeCount += tics;

        UpdateSoundLoc ();      // JAB
        if (screenfaded)
            VW_FadeIn ();

        CheckKeys ();

//
// debug aids
//
        if (singlestep)
        {
            VW_WaitVBL (singlestep);
            lasttimecount = GetTimeCount();
        }
        if (extravbls)
            VW_WaitVBL (extravbls);

        if (demoplayback)
        {
            if (IN_CheckAck ())
            {
                IN_ClearKeysDown ();
                playstate = ex_abort;
            }
        }
    }
    while (!playstate && !startgame);

    if (playstate != ex_died)
        FinishPaletteShifts ();
}
