// WL_STATE.C

#include "wl_def.h"
#pragma hdrstop

/*
=============================================================================

                            LOCAL CONSTANTS

=============================================================================
*/


/*
=============================================================================

                            GLOBAL VARIABLES

=============================================================================
*/


static const dirtype opposite[9] =
    {west,southwest,south,southeast,east,northeast,north,northwest,nodir};

static const dirtype diagonal[9][9] =
{
    /* east */  {nodir,nodir,northeast,nodir,nodir,nodir,southeast,nodir,nodir},
                {nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir},
    /* north */ {northeast,nodir,nodir,nodir,northwest,nodir,nodir,nodir,nodir},
                {nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir},
    /* west */  {nodir,nodir,northwest,nodir,nodir,nodir,southwest,nodir,nodir},
                {nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir},
    /* south */ {southeast,nodir,nodir,nodir,southwest,nodir,nodir,nodir,nodir},
                {nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir},
                {nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir}
};



void    SpawnNewObj (unsigned tilex, unsigned tiley, statetype *state);
void    NewState (objtype *ob, statetype *state);

boolean TryWalk (objtype *ob);
void    MoveObj (objtype *ob, int32_t move);

void    KillActor (objtype *ob);
void    DamageActor (objtype *ob, unsigned damage);

boolean CheckLine (objtype *ob);
void    FirstSighting (objtype *ob);
boolean CheckSight (objtype *ob);

/*
=============================================================================

                                LOCAL VARIABLES

=============================================================================
*/



//===========================================================================


/*
===================
=
= SpawnNewObj
=
= Spaws a new actor at the given TILE coordinates, with the given state, and
= the given size in GLOBAL units.
=
= newobj = a pointer to an initialized new actor
=
===================
*/

void SpawnNewObj (unsigned tilex, unsigned tiley, statetype *state)
{
    GetNewActor ();
    newobj->state = state;
    if (state->tictime)
        newobj->ticcount = DEMOCHOOSE_ORIG_SDL(
                US_RndT () % state->tictime,
                US_RndT () % state->tictime + 1);     // Chris' moonwalk bugfix ;D
    else
        newobj->ticcount = 0;

    newobj->tilex = (short) tilex;
    newobj->tiley = (short) tiley;
    newobj->x = ((int32_t)tilex<<TILESHIFT)+TILEGLOBAL/2;
    newobj->y = ((int32_t)tiley<<TILESHIFT)+TILEGLOBAL/2;
    newobj->dir = nodir;

    actorat[tilex][tiley] = newobj;
    newobj->areanumber =
        *(mapsegs[0] + (newobj->tiley<<mapshift)+newobj->tilex) - AREATILE;
}



/*
===================
=
= NewState
=
= Changes ob to a new state, setting ticcount to the max for that state
=
===================
*/

void NewState (objtype *ob, statetype *state)
{
    ob->state = state;
    ob->ticcount = state->tictime;
}



/*
=============================================================================

                        ENEMY TILE WORLD MOVEMENT CODE

=============================================================================
*/

/*
==================================
=
= TryWalk
=
= Attempts to move ob in its current (ob->dir) direction.
=
= If blocked by either a wall or an actor returns FALSE
=
= If move is either clear or blocked only by a door, returns TRUE and sets
=
= ob->tilex         = new destination
= ob->tiley
= ob->areanumber    = the floor tile number (0-(NUMAREAS-1)) of destination
= ob->distance      = TILEGLOBAl, or -doornumber if a door is blocking the way
=
= If a door is in the way, an OpenDoor call is made to start it opening.
= The actor code should wait until
=       doorobjlist[-ob->distance].action = dr_open, meaning the door has been
=       fully opened
=
==================================
*/
// Enemy AI
char    dx8dir[9]= { 1, 1, 0, -1, -1, -1,  0,  1, 0};  // dx & dy based on direction
char    dy8dir[9]= { 0, 1, 1,  1,  0, -1, -1, -1, 0};  // used mainly in wl_enemy_ai.cpp

boolean CheckMovementOK(int x, int y, int *doornum, objtype *ob)
{
  uintptr_t temp;

    temp =(uintptr_t)actorat[x][y];

    // Blocking Tile Check - Enemy Can't Cross
    switch (MAPSPOT(x,y,1))
    {
        case BLOCKTILE:
            return true;
            break;
    }
    if (temp)
    {
        if (temp<BIT_DOOR && doornum != NULL)
            return true;
        if (temp<BIT_ALLTILES && doornum == NULL)
            return true;
        else if (temp<BIT_ALLTILES && doornum != NULL) {
#ifdef PLAYDEMOLIKEORIGINAL       
            if(DEMOCOND_ORIG) {
                *doornum = (int)temp & LAST_DOORNUM;
            } else {
#endif         
                *doornum = (int)temp & LAST_DOORNUM;
            }
        }
        else if (((objtype *)temp)->flags&FL_SHOOTABLE)
            return true;
    }

    return false;
}

boolean TryWalk (objtype *ob) {
    int     oldx, oldy, newx, newy, doornum = -1;
 
    if(ob->dir == nodir)
        return false;

    oldx=ob->tilex;
    oldy=ob->tiley;
    newx= oldx + dx8dir[ob->dir];
    newy= oldy - dy8dir[ob->dir];   
    if (ob->dir & 1) { // Check Diagonal Movement
      if(CheckMovementOK(newx, oldy, NULL, ob) ||
          CheckMovementOK(oldx, newy, NULL, ob) ||
          CheckMovementOK(newx, newy, NULL, ob))
          return false;
    }
   else {                  // Check Cardinal Movement
        switch (ob->obclass) {
            case demonobj: // cannot go through doors
            case soulobj:
            case cacoobj:
            case painobj:
            case impobj:
            case cyberobj:
            case mancubusobj:
            case revenantobj:
            case bossobj:
            case boss2obj:
            case spiderobj:
                if(CheckMovementOK(newx, newy, NULL, ob)) return false;
                break;
            default:
                if(CheckMovementOK(newx, newy, &doornum, ob)) return false;
                break;   
        }
    }
   
    int lock = doorobjlist[doornum].lock;
    if(doornum != -1 
        && lock >= dr_lock1
        && lock <= dr_lock4
        && (!(gamestate.keys & (1 << (lock-dr_lock1))))) {
        return false;
    } else if (doornum != -1) {
        OpenDoor(doornum);
        ob->distance = -doornum-1;
        return true;
    }
    else
        ob->distance = TILEGLOBAL; // go ahead, its clear

    ob->tilex=newx;
    ob->tiley=newy;
    ob->areanumber=MAPSPOT(newx, newy, 0)-AREATILE;
    return true;
}

/* =============== Swap 2 Direction Values ================== */
dirtype SwapVal (dirtype *value1, dirtype *value2)
{
  dirtype temp = *value1;
  *value1 = *value2;
  *value2 = temp;
  return *value1, *value2;
}

/*
==================================
=
= SelectDodgeDir
=
= Attempts to choose and initiate a movement for ob that sends it towards
= the player while dodging
=
= If there is no possible move (ob is totally surrounded)
=
= ob->dir           =       nodir
=
= Otherwise
=
= ob->dir           = new direction to follow
= ob->distance      = TILEGLOBAL or -doornumber
= ob->tilex         = new destination
= ob->tiley
= ob->areanumber    = the floor tile number (0-(NUMAREAS-1)) of destination
=
==================================
*/
void SelectDodgeDir (objtype *ob)
{
  int     i;
  dirtype dirtry[5], turnaround;
  // turning around is only ok the very first time
  // after noticing the player
  if (ob->flags & FL_FIRSTATTACK) { turnaround = nodir; ob->flags &= ~FL_FIRSTATTACK; }
  else turnaround=opposite[ob->dir];
  // arange 5 direction choices in order of preference
  // the four cardinal directions plus the diagonal
  // straight towards the player
  if (player->tilex - ob->tilex > 0) { dirtry[1]= east;  dirtry[3]= west; }
  else                               { dirtry[1]= west;  dirtry[3]= east; }
  if (player->tiley - ob->tiley > 0) { dirtry[2]= south; dirtry[4]= north; }
  else                               { dirtry[2]= north; dirtry[4]= south; }
  // randomize a bit for dodging
  if (abs(player->tilex - ob->tilex) > abs(player->tiley - ob->tiley))
  { SwapVal (&dirtry[1], &dirtry[2]); SwapVal (&dirtry[3], &dirtry[4]); }
  if (US_RndT() < 128)
  { SwapVal (&dirtry[1], &dirtry[2]); SwapVal (&dirtry[3], &dirtry[4]); }
  dirtry[0] = diagonal[dirtry[1]][dirtry[2]];
  for (i=0;i<5;i++) // try the directions util one works
  {
    if ( dirtry[i] == nodir || dirtry[i] == turnaround) continue;
    ob->dir = dirtry[i];
    if (TryWalk(ob)) return;
  }
  if (turnaround != nodir) // turn around only as a last resort
  { ob->dir = turnaround; if (TryWalk(ob)) return; }
  ob->dir = nodir;
}

/*
============================
=
= SelectChaseDir
=
= As SelectDodgeDir, but doesn't try to dodge
=
============================
*/
void SelectChaseDir (objtype *ob)
{
  dirtype d[3];
  dirtype tdir, olddir, turnaround;

  olddir=ob->dir;
  turnaround=opposite[olddir];
  d[1]=nodir;
  d[2]=nodir;
  if (player->tilex - ob->tilex != 0)
    d[1] = (player->tilex - ob->tilex > 0 ? east:west);
  if (player->tiley - ob->tiley != 0)
    d[2] = (player->tiley - ob->tiley > 0 ? south:north);
  if (abs(player->tiley - ob->tiley) > abs(player->tilex - ob->tilex))
    SwapVal ( &d[1], &d[2] );
  if (d[1] == turnaround) d[1]=nodir;
  if (d[2] == turnaround) d[2]=nodir;
  if (d[1]!=nodir) /*either moved forward or attacked*/
  { ob->dir=d[1]; if (TryWalk(ob)) return; }
  if (d[2]!=nodir)
  { ob->dir=d[2]; if (TryWalk(ob)) return; }
  /* there is no direct path to the player, so pick another direction */
  if (olddir!=nodir)
  { ob->dir=olddir; if (TryWalk(ob)) return; }
  if (US_RndT()>128)      /*randomly determine direction of search*/
  {
#ifdef DARKONEAI
    for (tdir=east; tdir<=south; tdir=(dirtype)(tdir+2))
#else
    for (tdir=north; tdir<=west; tdir=(dirtype)(tdir+1))
#endif
    {
      if (tdir!=turnaround) { ob->dir=tdir; if ( TryWalk(ob) ) return; }
    }
  }
  else
  {
#ifdef DARKONEAI
    for (tdir=south; tdir>=east; tdir=(dirtype)(tdir-2))
#else
    for (tdir=west; tdir>=north; tdir=(dirtype)(tdir-1))
#endif
    {
      if (tdir!=turnaround) { ob->dir=tdir; if ( TryWalk(ob) ) return; }
    }
  }
  if (turnaround != nodir)
  {
    ob->dir=turnaround;
    if (ob->dir != nodir) { if ( TryWalk(ob) ) return; }
  }
  ob->dir = nodir; // can't move
}

/*
============================
=
= SelectRunDir - Retreat
=
= Run Away from player
=
============================
*/
void SelectRunDir (objtype *ob)
{
  dirtype d[3];
  dirtype tdir;
  if (player->tilex - ob->tilex < 0) d[1]= east; else d[1]= west;
  if (player->tiley - ob->tiley < 0) d[2]=south; else d[2]=north;
  if (abs(player->tiley - ob->tiley) > abs(player->tilex - ob->tilex))
  { SwapVal (&d[1], &d[2]); }
  ob->dir=d[1];
  if (TryWalk(ob)) return; /*either moved forward or attacked*/
  ob->dir=d[2];
  if (TryWalk(ob)) return;
  /* there is no direct path to the player, so pick another direction */
  if (US_RndT()>128)      /*randomly determine direction of search*/
  {
#ifdef DARKONEAI
    for (tdir=east; tdir<=south; tdir=(dirtype)(tdir+2))
#else
    for (tdir=north; tdir<=west; tdir=(dirtype)(tdir+1))
#endif
    { ob->dir=tdir; if ( TryWalk(ob) ) return; }
  }
  else
  {
#ifdef DARKONEAI
    for (tdir=south; tdir>=east; tdir=(dirtype)(tdir-2))
#else
    for (tdir=west; tdir>=north; tdir=(dirtype)(tdir-1))
#endif   
    { ob->dir=tdir; if ( TryWalk(ob) ) return; }
  }
  ob->dir = nodir;                // can't move
}

/*
=================
=
= MoveObj
=
= Moves ob be move global units in ob->dir direction
= Actors are not allowed to move inside the player
= Does NOT check to see if the move is tile map valid
=
= ob->x                 = adjusted for new position
= ob->y
=
=================
*/

void MoveObj (objtype *ob, int32_t move)
{
    int32_t    deltax,deltay;

    switch (ob->dir)
    {
        case north:
            ob->y -= move;
            break;
        case northeast:
            ob->x += move;
            ob->y -= move;
            break;
        case east:
            ob->x += move;
            break;
        case southeast:
            ob->x += move;
            ob->y += move;
            break;
        case south:
            ob->y += move;
            break;
        case southwest:
            ob->x -= move;
            ob->y += move;
            break;
        case west:
            ob->x -= move;
            break;
        case northwest:
            ob->x -= move;
            ob->y -= move;
            break;

        case nodir:
            return;

        default:
            Quit ("MoveObj: bad dir!");
    }

    //
    // check to make sure it's not on top of player
    //
    if (areabyplayer[ob->areanumber])
    {
        deltax = ob->x - player->x;
        if (deltax < -MINACTORDIST || deltax > MINACTORDIST)
            goto moveok;
        deltay = ob->y - player->y;
        if (deltay < -MINACTORDIST || deltay > MINACTORDIST)
            goto moveok;

        if (ob->hidden)          // move closer until he meets CheckLine
            goto moveok;

        //
        // back up
        //
        switch (ob->dir)
        {
            case north:
                ob->y += move;
                break;
            case northeast:
                ob->x -= move;
                ob->y += move;
                break;
            case east:
                ob->x -= move;
                break;
            case southeast:
                ob->x -= move;
                ob->y -= move;
                break;
            case south:
                ob->y -= move;
                break;
            case southwest:
                ob->x += move;
                ob->y -= move;
                break;
            case west:
                ob->x += move;
                break;
            case northwest:
                ob->x += move;
                ob->y += move;
                break;

            case nodir:
                return;
        }
        return;
    }
moveok:
    ob->distance -=move;
}

/*
=============================================================================

                                STUFF

=============================================================================
*/

/*
===============
=
= DropItem
=
= Tries to drop a bonus item somewhere in the tiles surrounding the
= given tilex/tiley
=
===============
*/

void DropItem (wl_stat_t itemtype, int tilex, int tiley)
{
    int     x,y,xl,xh,yl,yh;

    //
    // find a free spot to put it in
    //
    if (!actorat[tilex][tiley])
    {
        PlaceItemType (itemtype, tilex,tiley);
        return;
    }

    xl = tilex-1;
    xh = tilex+1;
    yl = tiley-1;
    yh = tiley+1;

    for (x=xl ; x<= xh ; x++)
    {
        for (y=yl ; y<= yh ; y++)
        {
            if (!actorat[x][y])
            {
                PlaceItemType (itemtype, x,y);
                return;
            }
        }
    }
}



/*
===============
=
= KillActor
=
===============
*/

void KillActor (objtype *ob)
{
    int     tilex,tiley;

    tilex = ob->tilex = (word)(ob->x >> TILESHIFT);         // drop item on center
    tiley = ob->tiley = (word)(ob->y >> TILESHIFT);

    switch (ob->obclass)
    {
        case mutantobj:
            GivePoints (100);
            if (ob->hitpoints<=-200)
            NewState (ob,&s_mutantexpl1);
            else
            NewState (ob,&s_mutantdie1);
            PlaceItemType (bo_ammoused,tilex,tiley);
            break;

        case chainguyobj:
            GivePoints (400);
            if (ob->hitpoints<=-280)
            NewState (ob,&s_chainguyexpl1);
            else
            NewState (ob,&s_chainguydie1);
            PlaceItemType (bo_chaingun,tilex,tiley);
            break;

        case impobj:
            GivePoints (200);
            if (ob->hitpoints<=-205)
            NewState (ob,&s_impexpl1);
            else
            NewState (ob,&s_impdie1);
            break;
        case painobj:
            GivePoints (500);
            NewState (ob,&s_paindie1);
            break;
        case soulobj:
            GivePoints (150);
            NewState (ob,&s_souldie1);
            gamestate.lostsoulcount--;
            break;
        case shotguyobj:
            GivePoints (200);
            if (ob->hitpoints<=-310)
            NewState (ob,&s_shotguyexpl1);
            else
            NewState (ob,&s_shotguydie1);
            PlaceItemType (bo_shotgun,tilex,tiley);
            break;

        case cacoobj:
            GivePoints (500);
            NewState (ob,&s_cacodie1);
            break;

        case demonobj:
            GivePoints (100);
            NewState (ob,&s_demondie1);
            break;

        case cyberobj:
            GivePoints (5000);
            NewState (ob,&s_cyberdie1);
            PlaceItemType (bo_key1,tilex,tiley);
            break;

        case mancubusobj:
            GivePoints (5000);
            gamestate.killx = player->x;
            gamestate.killy = player->y;
            NewState (ob,&s_mancubusdie1);
            break;
        case vileobj:
            GivePoints (5000);
            gamestate.killx = player->x;
            gamestate.killy = player->y;
            NewState (ob,&s_viledie1);
            break;

        case spiderobj:
            GivePoints (5000);
            gamestate.killx = player->x;
            gamestate.killy = player->y;
            NewState (ob,&s_spiderdie1);
            break;

        case revenantobj:
            GivePoints (5000);
            NewState (ob,&s_revenantdie1);
            PlaceItemType (bo_key1,tilex,tiley);
            break;
        case bossobj:
            GivePoints (5000);
            NewState (ob,&s_bossdie1);
            if ((gamestate.episode*10 + gamestate.mapon) == 8) { // Episode 1 boss
                PlaceItemType (bo_key1,tilex,tiley);
            }
            break;
        case boss2obj:
            GivePoints (5000);
            NewState (ob,&s_boss2die1);
            break;
        case barrelobj:
            gamestate.killcount--;
            NewState (ob,&s_barreldie1);
            break;
    }

    gamestate.killcount++;
    ob->flags &= ~FL_SHOOTABLE;
    actorat[ob->tilex][ob->tiley] = NULL;
    ob->flags |= FL_NONMARK;
}



/*
===================
=
= DamageActor
=
= Called when the player succesfully hits an enemy.
=
= Does damage points to enemy ob, either putting it into a stun frame or
= killing it.
=
===================
*/

void DamageActor (objtype *ob, unsigned damage)
{
    madenoise = true;

    //
    // do double damage if shooting a non attack mode actor
    //
    if ( !(ob->flags & FL_ATTACKMODE) )
        damage <<= 1;

    ob->hitpoints -= (short)damage;

    if (ob->hitpoints<=0)
        KillActor (ob);
    else
    {
        if (! (ob->flags & FL_ATTACKMODE) )
            FirstSighting (ob);             // put into combat mode

        switch (ob->obclass)                // dogs only have one hit point
        {
            case mutantobj:
                if (ob->hitpoints&1)
                    NewState (ob,&s_mutantpain);
                PlaySoundLocActor (DSPOPAIN,ob, false);
                break;
            case chainguyobj:
                if (ob->hitpoints&1)
                    NewState (ob,&s_chainguypain);
                PlaySoundLocActor (DSPOPAIN,ob, false);
                break;
            case impobj:
                if (ob->hitpoints&1)
                    NewState (ob,&s_imppain);
                PlaySoundLocActor (DSPOPAIN,ob, false);
                break;
            case painobj:
                if (ob->hitpoints&1)
                    NewState (ob,&s_painpain);
                PlaySoundLocActor(DSPEPAIN,ob, false);
                break;
            case soulobj:
                if (ob->hitpoints&1)
                    NewState (ob,&s_soulpain);
                PlaySoundLocActor (DSPOPAIN,ob, false);
                break;
            case demonobj:
                if (ob->hitpoints&1)
                    NewState (ob,&s_demonpain);
                PlaySoundLocActor (DSDMPAIN,ob, false);
                break;
            case shotguyobj:
                if (ob->hitpoints&1)
                    NewState (ob,&s_shotguypain);
                PlaySoundLocActor (DSPOPAIN,ob, false);
                break;
            case cacoobj:
                if (ob->hitpoints&1)
                    NewState (ob,&s_cacopain);
                PlaySoundLocActor (DSDMPAIN,ob, false);
                break;
        }
    }
}

/*
=============================================================================

                                CHECKSIGHT

=============================================================================
*/


/*
=====================
=
= CheckLine
=
= Returns true if a straight line between the player and ob is unobstructed
=
=====================
*/

boolean CheckLine (objtype *ob)
{
    int         x1,y1,xt1,yt1,x2,y2,xt2,yt2;
    int         x,y;
    int         xdist,ydist,xstep,ystep;
    int         partial,delta;
    int32_t     ltemp;
    int         xfrac,yfrac,deltafrac;
    unsigned    value,intercept;

    x1 = ob->x >> UNSIGNEDSHIFT;            // 1/256 tile precision
    y1 = ob->y >> UNSIGNEDSHIFT;
    xt1 = x1 >> 8;
    yt1 = y1 >> 8;

    x2 = plux;
    y2 = pluy;
    xt2 = player->tilex;
    yt2 = player->tiley;

    xdist = abs(xt2-xt1);

    if (xdist > 0)
    {
        if (xt2 > xt1)
        {
            partial = 256-(x1&0xff);
            xstep = 1;
        }
        else
        {
            partial = x1&0xff;
            xstep = -1;
        }

        deltafrac = abs(x2-x1);
        delta = y2-y1;
        ltemp = ((int32_t)delta<<8)/deltafrac;
        if (ltemp > 0x7fffl)
            ystep = 0x7fff;
        else if (ltemp < -0x7fffl)
            ystep = -0x7fff;
        else
            ystep = ltemp;
        yfrac = y1 + (((int32_t)ystep*partial) >>8);

        x = xt1+xstep;
        xt2 += xstep;
        do
        {
            y = yfrac>>8;
            yfrac += ystep;

            value = (unsigned)tilemap[x][y];
            x += xstep;

            if (!value)
                continue;

            if (value<BIT_DOOR || value>BIT_ALLTILES)
                return false;

            //
            // see if the door is open enough
            //
            value &= ~BIT_DOOR;
            intercept = yfrac-ystep/2;

            if (intercept>doorposition[value])
                return false;

        } while (x != xt2);
    }

    ydist = abs(yt2-yt1);

    if (ydist > 0)
    {
        if (yt2 > yt1)
        {
            partial = 256-(y1&0xff);
            ystep = 1;
        }
        else
        {
            partial = y1&0xff;
            ystep = -1;
        }

        deltafrac = abs(y2-y1);
        delta = x2-x1;
        ltemp = ((int32_t)delta<<8)/deltafrac;
        if (ltemp > 0x7fffl)
            xstep = 0x7fff;
        else if (ltemp < -0x7fffl)
            xstep = -0x7fff;
        else
            xstep = ltemp;
        xfrac = x1 + (((int32_t)xstep*partial) >>8);

        y = yt1 + ystep;
        yt2 += ystep;
        do
        {
            x = xfrac>>8;
            xfrac += xstep;

            value = (unsigned)tilemap[x][y];
            y += ystep;

            if (!value)
                continue;

            if (value<BIT_DOOR || value>BIT_ALLTILES)
                return false;

            //
            // see if the door is open enough
            //
            value &= ~BIT_DOOR;
            intercept = xfrac-xstep/2;

            if (intercept>doorposition[value])
                return false;
        } while (y != yt2);
    }

    return true;
}


/*
================
=
= CheckSight
=
= Checks a straight line between player and current object
=
= If the sight is ok, check alertness and angle to see if they notice
=
= returns true if the player has been spoted
=
================
*/

#define MINSIGHT        0x18000l

boolean CheckSight (objtype *ob)
{
    int32_t deltax,deltay;

    //
    // don't bother tracing a line if the area isn't connected to the player's
    //
    if (!areabyplayer[ob->areanumber])
        return false;

    //
    // if the player is real close, sight is automatic
    //
    deltax = player->x - ob->x;
    deltay = player->y - ob->y;

    if (deltax > -MINSIGHT && deltax < MINSIGHT
        && deltay > -MINSIGHT && deltay < MINSIGHT)
        return true;

    //
    // see if they are looking in the right direction
    //
    switch (ob->dir)
    {
        case north:
            if (deltay > 0)
                return false;
            break;

        case east:
            if (deltax < 0)
                return false;
            break;

        case south:
            if (deltay < 0)
                return false;
            break;

        case west:
            if (deltax > 0)
                return false;
            break;

        // check diagonal moving guards fix

        case northwest:
            if (DEMOCOND_SDL && deltay > -deltax)
                return false;
            break;

        case northeast:
            if (DEMOCOND_SDL && deltay > deltax)
                return false;
            break;

        case southwest:
            if (DEMOCOND_SDL && deltax > deltay)
                return false;
            break;

        case southeast:
            if (DEMOCOND_SDL && -deltax > deltay)
                return false;
            break;
    }

    //
    // trace a line to check for blocking tiles (corners)
    //
    return CheckLine (ob);
}


/*
===============
=
= FirstSighting
=
= Puts an actor into attack mode and possibly reverses the direction
= if the player is behind it
=
===============
*/

void FirstSighting (objtype *ob)
{
    int mutantsounds[3]={DSPOSIT1, DSPOSIT2, DSPOSIT3},
        impsounds[2]={DSBGSIT1, DSBGSIT2};
    //
    // react to the player
    //
    switch (ob->obclass)
    {
        case mutantobj:
            PlaySoundLocActor (mutantsounds[US_RndT()%2],ob, true);
            NewState (ob,&s_mutantchase1);
            ob->speed *= 3;
            break;

        case chainguyobj:
            PlaySoundLocActor (mutantsounds[US_RndT()%2],ob, true);
            NewState (ob,&s_chainguychase1);
            ob->speed *= 3;
            break;

        case impobj:
            PlaySoundLocActor (impsounds[US_RndT()%1],ob, true);
            NewState (ob,&s_impchase1);
            ob->speed *= 2;
            break;

        case painobj:
            PlaySoundLocActor(DSPESIT,ob, true);
            NewState (ob,&s_painchase1);
            ob->speed *= 2;
            break;

        case soulobj:
            NewState (ob,&s_soulchase1);
            ob->speed *= 3;
            break;
        case shotguyobj:
            PlaySoundLocActor (mutantsounds[US_RndT()%2],ob, true);
            NewState (ob,&s_shotguychase1);
            ob->speed *= 2;
            break;

        case cacoobj:
            PlaySoundLocActor (DSCACSIT,ob, true);
            NewState (ob,&s_cacochase1);
            ob->speed *= 2;
            break;

        case demonobj:
            PlaySoundLocActor(DSSGTSIT,ob, true);
            NewState (ob,&s_demonchase1);
            ob->speed *= 2;
            break;

        case cyberobj:
            PlaySoundLocActor (DSCYBSIT,ob, true);
            NewState (ob,&s_cyberchase1);
            ob->speed = SPDPATROL*3;
            break;

        case mancubusobj:
            PlaySoundLocActor (DSMANSIT,ob, true);
            NewState (ob,&s_mancubuschase1);
            ob->speed *= 3;
            break;
        case vileobj:
            PlaySoundLocActor (DSVILSIT,ob, true);
            NewState (ob,&s_vilechase1);
            ob->speed *= 3;
            break;
        case spiderobj:
            PlaySoundLocActor (DSBSPSIT,ob, true);
            NewState (ob,&s_spiderchase1);
            ob->speed *= 3;
            break;

        case revenantobj:
            PlaySoundLocActor (DSSKESIT,ob, true);
            NewState (ob,&s_revenantchase1);
            ob->speed *= 3;
            break;

        case bossobj:
            PlaySoundLocActor (DSBRSSIT,ob, true);
            NewState (ob,&s_bosschase1);
            ob->speed *= 3;
            break;
        case boss2obj:
            PlaySoundLocActor (DSKNTSIT,ob, true);
            NewState (ob,&s_boss2chase1);
            ob->speed *= 3;
            break;
    }

    if (ob->distance < 0)
        ob->distance = 0;       // ignore the door opening command

    ob->flags |= FL_ATTACKMODE|FL_FIRSTATTACK;
}



/*
===============
=
= SightPlayer
=
= Called by actors that ARE NOT chasing the player.  If the player
= is detected (by sight, noise, or proximity), the actor is put into
= it's combat frame and true is returned.
=
= Incorporates a random reaction delay
=
===============
*/

boolean SightPlayer (objtype *ob)
{
    if (ob->flags & FL_ATTACKMODE)
        Quit ("An actor in ATTACKMODE called SightPlayer!");

    if (ob->temp2)
    {
        //
        // count down reaction time
        //
        ob->temp2 -= (short) tics;
        if (ob->temp2 > 0)
            return false;
        ob->temp2 = 0;                                  // time to react
    }
    else
    {
        if (!areabyplayer[ob->areanumber])
            return false;

        if (ob->flags & FL_AMBUSH)
        {
            if (!CheckSight (ob))
                return false;
            ob->flags &= ~FL_AMBUSH;
        }
        else
        {
            if (!madenoise && !CheckSight (ob))
                return false;
        }


        switch (ob->obclass)
        {
            case mutantobj:
                ob->temp2 = 1+US_RndT()/4;
                break;
            case chainguyobj:
                ob->temp2 = 2;
                break;
            case impobj:
            case painobj:
            case soulobj:
                ob->temp2 = 1+US_RndT()/4;
                break;
            case shotguyobj:
                ob->temp2 = 1+US_RndT()/6;
                break;
            case cacoobj:
                ob->temp2 = 1+US_RndT()/6;
                break;
            case demonobj:
                ob->temp2 = 1+US_RndT()/8;
                break;
            case cyberobj:
            case spiderobj:
            case revenantobj:
            case bossobj:
            case boss2obj:
            case mancubusobj:
            case vileobj:
                ob->temp2 = 1;
                break;
        }
        return false;
    }

    FirstSighting (ob);

    return true;
}
